########################################################################
#
#  analyze.R
#
#  Analyze results of the Moreton, Pater & Pertsova 2017 simulations.
#
#  E. Moreton  *  UNC-Chapel Hill Linguistics  *  2019 July 31 (W)
#
########################################################################

# Simulation directory names look like this:
# Exp0-S012-Type_I-2000-random-0.05-0.33-1-1-0.025-0-Mut_0.1-Rec.1.00-0-0
# Use the following variables to specify which one to analyze:
eta <- 0.33
mu  <- 0.025
mut <- "Mut_0.1"
replications <- c (100:199) 
name.left  <- "Exp01-"
name.right <- paste (
    "-2000-random-0.05-", eta, "-1-1-", mu, "-0-", mut, "-Rec.1.00-0-",
    sep = "", collapse = ""
)
rep.names.I <- sapply (
    replications,
    function (i) {
        paste (name.left, "S012-Type_I", name.right, i, 
	sep = "", collapse = "")
    }
)
rep.names.II <- sapply (
    replications,
    function (i) {
        paste (name.left, "S131-Type_II", name.right, i, 
	sep = "", collapse = "")
    }
)
rep.names.IV <- sapply (
    replications,
    function (i) {
        paste (name.left, "S013-Type_IV", name.right, i, 
	sep = "", collapse = "")
    }
)


load.report <- function (
    simname,
    dirname = '.'
) {
    report_path = paste (dirname, simname, sep = '/')

    Report <- read.csv (
        paste (report_path, "Report.csv", sep = '/'),
	check.names = FALSE, as.is = TRUE
    )
    correct_line = which (Report$data_type == 'correct')
    Report [4, which (Report [correct_line, ]=='F')] <- FALSE
    Report [4, which (Report [correct_line, ]=='T')] <- TRUE

    Stimuli <- read.csv (
        paste (report_path, "Stimuli.csv", sep = '/'),
	check.names = FALSE
    )
    Constraints <- read.csv (
        paste (report_path, "Constraints.csv", sep = '/'),
	check.names = FALSE
    )
    Lineage <- read.csv (
        paste (report_path, "Lineage.csv", sep = '/'),
	check.names = FALSE
    )

    Harmony <- read.csv (
        paste (report_path, "Harmony.csv", sep = '/'),
	check.names = FALSE, as.is = TRUE
    )

    Weight <- read.csv (
        paste (report_path, "Weight.csv", sep = '/'),
	check.names = FALSE, as.is = TRUE
    )

    # Which column headers are trial numbers?
    n.trials <- ncol (Harmony) - 3
    trials <- 1:n.trials
    trial.cols <- 3 + trials   
    trial.cols.Report <- 1 + trial.cols
    
    list (Stimuli=Stimuli, Constraints = Constraints,
        Lineage=Lineage, Report=Report, Harmony=Harmony, Weight=Weight,
        n.trials = n.trials, trials = trials, trial.cols =trial.cols, 
	trial.cols.Report = trial.cols.Report)
}

#-----------------------------------------------------------------------
# constraint.named
#
# Returns the constraint with the given name.
#-----------------------------------------------------------------------

constraint.named <- function (
    data,
    conname
) {
    i <- which (data$Constraints$conname == conname)
    data$Constraints [i,]
}

#-----------------------------------------------------------------------
#  constraints.which
#
#  Returns a list of names of the constraints that award the specified
#  scores.
#-----------------------------------------------------------------------

constraints.which <- function (
    data,
    does.mark   = NULL,   # Names of stimuli
    doesnt.mark = NULL
) {
    n.stims <- nrow (data$Stimuli)
    marks <- sapply (
        1:n.stims,
        function (i) {
            stimname = data$Stimuli$stimname [i]
            if (any (doesnt.mark == stimname)) {
                "0"
            } else {
                if (any (does.mark == stimname)) {
                    "-?1"
                } else {
                    "(0|-?1)"
                }
            }
        }
    )
    marks.perlre <- paste (marks, collapse = ".")

    data$Constraints$conname [
    			     grep (marks.perlre, data$Constraints$signature, perl=TRUE)]
}

#-----------------------------------------------------------------------
# name.to.weights
#
# Given a constraint name, get the weights that go with it
#-----------------------------------------------------------------------

name.to.weights <- function (
    data,
    name
) {
    i <- which (data$Weight$constraint_name == name)
    # ``as.numeric'' needed to make a vector rather than a 1xt matrix
    as.numeric (data$Weight [i, data$trial.cols])
}

#-----------------------------------------------------------------------
#  peak
# 
#  Returns trial at which a constraint attains its peak weight
#-----------------------------------------------------------------------

peak <- function (
    data,
    con.name,
    value = FALSE   # Return trial number, or peak value?
) {
    w <- name.to.weights (data, con.name)
    if (value) {
            max (w)
    } else {
            as.numeric (which.max (w))
    }
}

#-----------------------------------------------------------------------
#  since
#
#  Returns trial on which a constraint is discovered (when the weight
#  first exceeds zero)
#-----------------------------------------------------------------------

since <- function (
    data,
    con.name
) {
    w <- name.to.weights (data, con.name)
    min (which (w > 0))
}


#-----------------------------------------------------------------------
#  pr.mut
#
#  Probability of Constraint C1 mutating into Constraint C2.
#-----------------------------------------------------------------------

pr.mut <- function (
    data,
    c1.name,
    c2.name
) {
    c1.to.anything <- 0
    c1.to.c2       <- 0
    for (i in 1:nrow (data$Lineage)) {
        count = data$Lineage$count [i]
        if (data$Lineage$mom [i] == c1.name) {
            c1.to.anything <- c1.to.anything + count
            if (data$Lineage$kid [i] == c2.name) {
                c1.to.c2 <- c1.to.c2 + count
            }
        }
    }
    if (c1.to.anything > 0) {
        c1.to.c2/c1.to.anything
    } else {
        NA
    }
}



#-----------------------------------------------------------------------
#  plot.Report
#
#  Plot the pr_corr and right/wrong data from Report.
#-----------------------------------------------------------------------

plot.Report <- function (
    data,
    xlim = NULL,
    ylim = NULL,
    ...
) {
    n.trials <- data$n.trials
    if (is.null (xlim)) {
        xlim = c (1, n.trials)
    }
    trials <- data$trials [1:n.trials]
    trial.cols <- data$trial.cols [1:n.trials]

    if (is.null (ylim)) {
        ylim = c (0,1)
    }

    pr_corr <- data$Report [data$Report$data_type == "pr_corr", trial.cols]    
    correct <- data$Report [data$Report$data_type == "correct", trial.cols]
    col <- sapply (
        correct,
        function (c) {
            if (is.na (c)) {NA} else {if (c==1) {'black'} else {'gray'}}
        }
    )
    plot (
        trials, pr_corr,
        xlim = xlim, ylim = ylim,
	pch = 20, col = col, cex=2/3,
	xlab = 'Trial', ylab = 'Pr (corr)',
        ...
    )
    radius <- 5
    st <- radius+1
    en <- length (trials) - radius -1
    fltr <- rep (1/(2*radius+1), 2*radius+1)
    filtered <- filter (as.numeric (pr_corr [st:en]), fltr, sides=2, method = "convolution")
    lines (trials [st:en], filtered)
}    


#-----------------------------------------------------------------------
#  plot.crit.stims
#
#  Given specified pos and neg stim names, plot their harmonies as a 
#  function of trial number.
#-----------------------------------------------------------------------

plot.crit.stims <- function (
    data,
    crit.pos.stims = NULL,
    crit.neg.stims = NULL,
    lty.pos = 'dashed',
    lty.neg = 'solid',
    add = FALSE,
    xlim = NULL,
    ylim = NULL,
    ...
) {

    if (is.null (crit.pos.stims)) {
        crit.pos.stims <- data$Stimuli$stimname [data$Stimuli$teach == 1]
    }

    if (is.null (crit.neg.stims)) {
        crit.neg.stims <- data$Stimuli$stimname [data$Stimuli$teach == -1]
    }


    n.trials <- data$n.trials
    if (is.null (xlim)) {
        xlim = c (1, n.trials)
    }
    trials <- data$trials [1:n.trials]
    trial.cols <- data$trial.cols [1:n.trials]

    if (is.null (ylim)) {
        ylim <- c (0, max (data$Harmony [, trial.cols]))
    }

    if (!add) {
    plot (
        0,0,
	    xlim = xlim,
	            ylim = ylim,
		    	     xlab = 'Trial',
			     	          ylab = 'Harmony',
					       	     type = 'n',
            ...
	        )
    }

    for (s in crit.pos.stims) {
        h <- data$Harmony [data$Harmony$stimulus_name == s, trial.cols]
	lines (trials, h, lty = lty.pos, ...)
    }
    for (s in crit.neg.stims) {
        h <- data$Harmony [data$Harmony$stimulus_name == s, trial.cols]
	lines (trials, h, lty = lty.neg, ...)
    }
}

#-----------------------------------------------------------------------
#  plot.crit.constraints
#
#  Given a list of critical (macro-)constraints, plot their weights as a
#  function of trial number.
#-----------------------------------------------------------------------

plot.crit.constraints <- function (
    data,
    crit.constraints = NULL,
    add = FALSE,
    xlim = NULL,
    ylim = NULL,
    ...
) {
    n.trials <- data$n.trials
    if (is.null (xlim)) {
        xlim = c (1, n.trials)
    }
    trials <- data$trials [1:n.trials]
    trial.cols <- data$trial.cols [1:n.trials]

    if (is.null (crit.constraints)) {
        crit.constraints <- data$Weight$constraint_name
    }

    if (is.null (ylim)) {
        ylim = c (0, max (data$Weight [, trial.cols]))
    }

    if (!add) {
    plot (
        0,0,
	    xlim = c (0, n.trials),
	            ylim = ylim,
		    	     xlab = 'Trial',
			     	          ylab = 'Weight',
					       	     type = 'n',
            ...
	        )
    }

    max.w <- max (data$Weight [, trial.cols])
    for (c in data$Constraints$conname) {
        c.row <- which (data$Weight$constraint_name == c)
        w <- data$Weight [c.row, trial.cols]

        # Constraints get a hue that depends on when they peaked, and
	# a width that depends on how high the peak was relative to 
	# other peaks at that time:

        peak.w  <- max (w)
        peak.at <- max (which (w == peak.w))
        w.at <- data$Weight [,trial.cols [peak.at]]	
	size <- peak.w / sum (w.at)

	if (any (crit.constraints == c)) {
   	    hue  <- 0.75 * peak.at/n.trials
            sat  <- 1
            val  <- 1
            lwd  <- 1 + 3*size
	        } else {
            hue <- 0
            sat <- 0
            val <- 9/10
            lwd <- 1
        }

        if (!add | any (crit.constraints == c)) {
   	   lines (trials [1:n.trials], w, col = hsv (hue,sat,val), lwd=lwd, ...)
           text (peak.at, peak.w, labels = c, pos = 3, cex = sqrt (sqrt (size)), 
	      	col=hsv (hue, sat, val) )
        }
    }
}    


#-----------------------------------------------------------------------
#  test.prcorr
#
#  Get average Pr (corr) across all test trials for a given simulation.
#-----------------------------------------------------------------------

test.prcorr <- function (
    data
) {
    n.trials <- data$n.trials
    n.test.trials <-32
    first.test.trial <- n.trials - n.test.trials +1
    trial.cols <- data$trial.cols.Report [first.test.trial:n.trials]
    mean (
        as.numeric (
            data$Report [data$Report$data_type == "pr_corr", trial.cols]
        )
    )
}

    