#!/usr/bin/perl
########################################################################
#
#  mksh.pl
#
#  Make the files and commands needed for the MPP 2017 simulations.
#
#  E. Moreton  *  UNC-Chapel Hill Linguistics  *  2019 July 3 (W)
#
########################################################################
use strict;
use warnings;
use Carp;

# For finding the index of an element in an array
# For randomly shuffling a list
use List::Util qw/ first shuffle /;

# For pasting pathnames together
use File::Spec;
use FindBin;
use lib File::Spec->catdir ($FindBin::Bin, "../../..", "Current", 'lib');

use Stimulus;
use StimulusSet;
use GJ2005jrFM;

########################################################################
#
#  Global parameters
#
########################################################################

# Where to find the simulation software
my $Simulate = "/nas/longleaf/home/moreton/Current/Simulate.pl";

# Where to find the .exp files from the human experiment
my $exp_dir = "./exps";

# Where we are now
open (PWD, "pwd |") or confess "Couldn't read output from pwd";
my $pwd = <PWD>;
# E.g., /nas/longleaf/home/moreton/GCL-exps/Current-SHJ/Exp07
$pwd =~ /^.*\/Exp(\d+)$/ or confess "Couldn't parse directory path $pwd";
my $expdigits = $1;
#my $expdigits = "0";

# Read in the stimuli
open (STIMS, "< All.stims") or confess ("Couldn't read data from All.stims");
my ($stims) = &StimulusSet::from_text (
    <STIMS>
    ) or confess ("Couldn't load stimuli from All.stims");
my @stimnames = map {
    $_->get_name
} ($stims->get_all);

# Column headers (last umpteen)
my @Columns = qw/
    subjid paradigm pattern phase block trial N adj zeta eta clutch_size
    parth mu momentum 
    const_mutagen const_recombinator metaconstraints 
    p_memorize mem_polarity 
    replication 
    correct_stim foil
/;

# Columns used for subjid
my @Subjid_Columns = qw/
    pattern N adj zeta eta clutch_size 
    parth mu momentum 
    const_mutagen const_recombinator p_memorize replication
/;

########################################################################
#
#  Experiment-specific parameters
#
########################################################################

my @Ns = (2000);
my @adjs = (
#    'fitness', 
#    'fitnessDraconian', 
#    'fitnessBinary', 
#    'validity', 
#    'validityDraconian', 
#    'validityBinary'
     'random'
);
my @zetas=(0.05);
my @etas = (0.33);
my @parths = (1);
my @mus = (0.025);
my @momenta = (0);
my @clutch_sizes = (1);
my @const_mutagens = qw/
  Mut_0.1
  Mut_0.01
 /;
my @const_recombinators = qw/ Rec.1.00 /;
my @p_mems = (0);

my @replications = (($expdigits * 100 + 1)..($expdigits * 100 + 99));
#my @replications = ($expdigits*100);
my @IDcodes = qw/ S012 S131 S013 /;

my $params;

# When memorizing a trial, memorize winner rather than loser
$params->{mem_polarity} = +1;


########################################################################
# 
#  Subroutines
#
########################################################################

sub load_exp {
    my $IDcode = shift;

    my $exp = File::Spec->catdir ($exp_dir, "$IDcode.exp");
    open (EXP, "<$exp") or confess "Can't open $exp";

    my @Train;
    my @Test;
    while (<EXP>) {
	chomp;
	push @Train, $_ if /^TrialPlayRecord/;
	push @Test, $_  if /^Trial2AFC/;
    }

    (\@Train, \@Test);
}

# Recode filenames as CELEX DISC
sub wavname_to_celex {
    my $name = shift;
    $name =~ tr/xa/{Q/;
    return $name;
}

sub make_subjid {
    my $params_ref = shift;
    my $IDcode = shift;
    my %params = %{$params_ref};
    "Exp$expdigits-$IDcode-" . (join "-", @params {@Subjid_Columns});
}


# Convert seconds to hh:mm:ss (without leading zeros)
sub seconds_to_HMS {
    my $seconds_needed = shift;

    my $hours = int ($seconds_needed/3600);
    $seconds_needed = $seconds_needed % 3600;
    my $minutes = int ($seconds_needed/60);
    my $seconds = $seconds_needed % 60;

    ($hours, $minutes, $seconds)
}

# Because the Math library isn't installed
sub cartesian_product {
    # Given a list of refs to sets, returns list of all possible strings
    # produced by concatenating one element from the first list, one 
    # from the second, etc.
    #
    # If the first argument is a scalar, it will be used as the
    # separation character in concatenation.
    my $sep_char = "";
    unless (ref ($_[0])) {
	$sep_char = shift;
    }
    my @set_refs = @_;

    # Basis case:  Cartesian product of an empty set is null string
    unless (@set_refs) {
	return ('');
    }
    
    # Induction:  
    my @product;
    for my $least_significant (@{shift @set_refs}) {
	for (&cartesian_product ($sep_char, @set_refs)) {
	    unshift @product, "$least_significant$sep_char$_";
	}
    }

    @product;
}

  

########################################################################
#
#  MAIN
#
########################################################################

my @expdefs = &cartesian_product (
    " ",
    \@Ns, \@adjs, \@zetas, \@etas, \@clutch_sizes,
    \@parths, \@mus, \@momenta,
    \@const_mutagens, \@const_recombinators, \@p_mems, \@replications);
for my $expdef (@expdefs) {
    my ($N, $adj, $zeta, $eta, $clutch_size, 
	$parth, $mu, $momentum,
	$const_mutagen, $const_recombinator, $p_mem, $replication) =
	    split /\s+/, $expdef;

    $params->{metaconstraints} = "meta.Const";
    $params->{N} = $N;
    $params->{adj} = $adj;
    if ($zeta=~/total (\d+)/) {
	my $total_harmony = $1;
	$params->{zeta} = sprintf "%4.2f", $total_harmony/$params->{N};
    } else {
	$params->{zeta} = $zeta;
    }
    $params->{eta} = $eta;
    $params->{parth} = $parth;
    $params->{mu} = $mu;
    $params->{clutch_size} = $clutch_size;
    $params->{momentum} = $momentum;
    $params->{const_mutagen} = $const_mutagen;
    $params->{const_recombinator} = $const_recombinator;
    $params->{p_memorize} = $p_mem;
    $params->{replication} = $replication;

    for my $IDcode (@IDcodes) {
	my ($Train, $Test) = &load_exp ($IDcode);
	my @Train = @{$Train}; my @Test = @{$Test};
	
	# TrialPlayRecord,1,dxka.wav,0,Next,dxka1.wav,Shepard.1,S001,14,1,Type_II,.0......|....0...|.....0..|,110,10100111
	my ($Type, $critbits, $true_values) = (split /,/, $Train [0]) [10, 11, 12];
	$params->{pattern} = $Type;

	my $subjid = &make_subjid ($params, $IDcode);
	$params->{subjid} = $subjid;

	my @csv;
	push @csv, (join ",", @Columns);

	my $trial_num = 1;
	$params->{phase} = 'Train';
	$params->{paradigm} = 'Predict';
	for (@Train) {
	    my $wav = (split /,/, $_) [5];

	    $wav =~ /(....)\d\.wav/ or confess "Couldn't parse $wav as wav file name";
	    my $pron = &wavname_to_celex ($1);
	    
	    $params->{correct_stim} = "F:$pron";
	    $params->{foil} = "NA";
	    $params->{trial} = $trial_num++;
	    $params->{block} = int (($params->{trial} -1) / 8) + 1;
	    my %params = %{$params};
	    my $trial_descriptor = join ",", @params {@Columns};
	    push @csv, $trial_descriptor;
	}    

	$params->{phase} = 'Test';
	$params->{paradigm} = '2AFC';
	# Turn off learning:
	my $eta   = $params->{eta};
	my $parth = $params->{parth};
	my $mu    = $params->{mu};
	$params->{eta} = 0;
	$params->{parth} = 1;
	$params->{mu} = 0;
	for (@Test) {
# Trial2AFC,2,dutu.wav,150,gita.wav,0,1,2,1,2,Shepard.1,S001,10,Type_II,.0......|....0...|.....0..|,111,10010001,011,11000011,1
	    /,(....)\.wav,.*,(....)\.wav,.*,([12])$/ or confess "Couldn't parse $_";  #/
  	    my $pron1 = &wavname_to_celex ($1);
	    my $pron2 = &wavname_to_celex ($2);
            my $which_right = $3;
	    
	    $params->{correct_stim} = ($which_right == 1) ? "F:$pron1" : "F:$pron2";
	    $params->{foil}         = ($which_right == 1) ? "F:$pron2" : "F:$pron1";
	    $params->{trial}        = $trial_num++;
	    $params->{block} = int (($params->{trial} -1) / 8) +1;
	    my %params = %{$params};
	    my $trial_descriptor = join ",", @params {@Columns};
	    push @csv, $trial_descriptor;
	}
	$params->{eta} = $eta;
	$params->{parth} = $parth;
	$params->{mu} = $mu;
    
	mkdir File::Spec->catdir (".", $subjid), 0755;
	my $expfile_path = File::Spec->catdir (".", $subjid, "Experiment.csv");
	open (EXPFILE, "> $expfile_path") or 
	    confess "Couldn't open $expfile_path for writing";
	print  EXPFILE (join "\n", @csv);
	print EXPFILE "\n";
	close EXPFILE;
    
	my ($hours, $minutes, $seconds) = (
	    23, 59, 59
	    );
    
	# Make the sh file
	my $sh_path = File::Spec->catdir (".", $subjid, "Simulate.sh");
	open (SHFILE, "> $sh_path") or
	    confess "Couldn't open $sh_path for writing";
	select SHFILE;
	print <<"MKSH";
#!/bin/bash
#SBATCH --ntasks=1
#SBATCH --time=$hours:$minutes:$seconds
#SBATCH --mem=1g
MKSH
	select STDOUT;
	print SHFILE join " ", ("perl $Simulate", "-repscheme GJ2005jrFM",  
				"-initcon 0.const", 
				"-stims All.stims", 
				"-exppath .", 
				"-subjid $subjid",
				"-mutagens Const.mut", "-recombinators Const.rec",
				"-interval 1 " ,
				"-log_format flat ",
				" -meta " . $params->{metaconstraints} . "\n");
	close SHFILE;				

	print "sbatch $sh_path\n";
    }
}
