########################################################################
#
#  analyze.R
#
#  Analyze results of the Reuse simulations.
#
#  E. Moreton  *  UNC-Chapel Hill Linguistics  *  2018 December 16 ($)
#
########################################################################

library (coin)    # For Wilcoxon rank-sum test



etas <- c (0.10, 0.15, 0.25, 0.30)
mus  <- c (0.025, 0.05, 0.1, 0.15)
muts <- c ("Mut_0.005");
replications <- c (800:899) 

make.simname <- function (eta, mu, mut, condition, replication) {
    name.left  <- "Exp08-2AFC-"
    name.right <- paste (
	"-200-random-0.05-", eta, "-1-1-", mu, "-0-", mut, "-Rec.1.00-0-",
	sep = "", collapse = ""
    )
    paste (name.left, condition, name.right, replication, sep = "", collapse = "")
}    

set.params <- function (eta, mu, mut) {

    name.left  <- "Exp08-2AFC-"
    name.right <- paste (
	"-200-random-0.05-", eta, "-1-1-", mu, "-0-", mut, "-Rec.1.00-0-",
	sep = "", collapse = ""
    )
    rep.names.A <- sapply (
	replications,
	function (i) {
	    paste (name.left, "A", name.right, i, sep = "", collapse = "")
	}
    )
    rep.names.B <- sapply (
	replications,
	function (i) {
	    paste (name.left, "B", name.right, i, sep = "", collapse = "")
	}
    )

    return (list (rep.names.A = rep.names.A, rep.names.B = rep.names.B))
}

load.report <- function (
    simname,
    dirname = '.'
) {
    report_path <- paste (dirname, simname, sep = '/')
    report_filename <- paste (report_path, "Report.csv", sep = '/')

    if (!file.exists (report_filename)) {
        return (NA)
    }

    Report <- tryCatch (read.csv (
        paste (report_path, "Report.csv", sep = '/'),
	check.names = FALSE, as.is = TRUE
    ), error = function (cond) {
          message (paste ("Couldn't read Report.csv from", report_path, sep = ' '))
          return (NA)
       },
       warning = function (cond) {
           message ("Warning from load.report")
           return (NA)
       }
    )
    correct_line = which (Report$data_type == 'correct')
    Report [4, which (Report [correct_line, ]=='F')] <- FALSE
    Report [4, which (Report [correct_line, ]=='T')] <- TRUE


    if (!file.exists (paste (report_path, "Stimuli.csv", sep = '/'))) {
        return (NA)
    }
    Stimuli <- read.csv (
        paste (report_path, "Stimuli.csv", sep = '/'),
	check.names = FALSE
    )
    if (!file.exists (paste (report_path, "Constraints.csv", sep = '/'))) {
        return (NA)
    }
    Constraints <- tryCatch (read.csv (
        paste (report_path, "Constraints.csv", sep = '/'),
	check.names = FALSE
    ), error = function (cond) {message (paste ("Couldn't read Constraints.csv from", report_path)); return (NA)},
       error = function (cond) {message (paste ("Warning reading Constraints.csv from", report_path)); return (NA)}
    )
    if (!file.exists (paste (report_path, "Lineage.csv", sep = '/'))) {
        return (NA)
    }
    Lineage <- read.csv (
        paste (report_path, "Lineage.csv", sep = '/'),
	check.names = FALSE
    )
    if (!file.exists (paste (report_path, "Harmony.csv", sep = '/'))) {
        return (NA)
    }
    Harmony <- read.csv (
        paste (report_path, "Harmony.csv", sep = '/'),
	check.names = FALSE, as.is = TRUE
    )
    if (!file.exists (paste (report_path, "Weight.csv", sep = '/'))) {
        return (NA)
    }
    Weight <- read.csv (
        paste (report_path, "Weight.csv", sep = '/'),
	check.names = FALSE, as.is = TRUE
    )

    # Which column headers are trial numbers?
    n.trials <- ncol (Harmony) - 3
    trials <- 1:n.trials
    trial.cols <- 3 + trials   
    trial.cols.Report <- 1 + trial.cols
    
    list (Stimuli=Stimuli, Constraints = Constraints,
        Lineage=Lineage, Report=Report, Harmony=Harmony, Weight=Weight,
        n.trials = n.trials, trials = trials, trial.cols =trial.cols, 
	trial.cols.Report = trial.cols.Report)
}

#-----------------------------------------------------------------------
# constraint.named
#
# Returns the constraint with the given name.
#-----------------------------------------------------------------------

constraint.named <- function (
    data,
    conname
) {
    i <- which (data$Constraints$conname == conname)
    data$Constraints [i,]
}

#-----------------------------------------------------------------------
#  constraints.which
#
#  Returns a list of names of the constraints that award the specified
#  scores.
#-----------------------------------------------------------------------

constraints.which <- function (
    data,
    does.mark   = NULL,   # Names of stimuli
    doesnt.mark = NULL
) {
    n.stims <- nrow (data$Stimuli)
    marks <- sapply (
        1:n.stims,
        function (i) {
            stimname = data$Stimuli$stimname [i]
            if (any (doesnt.mark == stimname)) {
                "0"
            } else {
                if (any (does.mark == stimname)) {
                    "-?1"
                } else {
                    "(0|-?1)"
                }
            }
        }
    )
    marks.perlre <- paste (marks, collapse = ".")

    data$Constraints$conname [
    			     grep (marks.perlre, data$Constraints$signature, perl=TRUE)]
}

#-----------------------------------------------------------------------
# name.to.weights
#
# Given a constraint name, get the weights that go with it
#-----------------------------------------------------------------------

name.to.weights <- function (
    data,
    name
) {
    i <- which (data$Weight$constraint_name == name)
    # ``as.numeric'' needed to make a vector rather than a 1xt matrix
    as.numeric (data$Weight [i, data$trial.cols])
}

#-----------------------------------------------------------------------
#  peak
# 
#  Returns trial at which a constraint attains its peak weight
#-----------------------------------------------------------------------

peak <- function (
    data,
    con.name,
    value = FALSE   # Return trial number, or peak value?
) {
    w <- name.to.weights (data, con.name)
    if (value) {
            max (w)
    } else {
            as.numeric (which.max (w))
    }
}

near.peak <- function (
    data,
    con.name,
    value = FALSE,
    tol.prop = 0.05  
) {
    w <- name.to.weights (data, con.name)
    criterion <- max (w) * (1-tol.prop)
    if (value) {
        criterion
    } else {
        w.purged <- w
        w.purged [w > criterion] <- NA
        as.numeric (which.max (w.purged))
    }
}

#-----------------------------------------------------------------------
#  since
#
#  Returns trial on which a constraint is discovered (when the weight
#  first exceeds zero)
#-----------------------------------------------------------------------

since <- function (
    data,
    con.name
) {
    w <- name.to.weights (data, con.name)
    min (which (w > 0))
}


#-----------------------------------------------------------------------
#  pr.mut.to
#
#  Probability of Constraint C1 mutating into Constraint C2.
#-----------------------------------------------------------------------

pr.mut.to <- function (
    data,
    c1.name,
    c2.name,
    include.self = TRUE
) {
    c1.to.anything <- 0
    c1.to.c2       <- 0
    c1.to.c1       <- 0
    for (i in 1:nrow (data$Lineage)) {
        count = data$Lineage$count [i]
        if (data$Lineage$mom [i] == c1.name) {
            c1.to.anything <- c1.to.anything + count
            if (data$Lineage$kid [i] == c2.name) {
                c1.to.c2 <- c1.to.c2 + count
            }
            if (data$Lineage$kid [i] == c1.name) {
                c1.to.c1 <- c1.to.c1 + count
            }
        }
    }
    if (include.self) {  
	if (c1.to.anything > 0) {
	    c1.to.c2/c1.to.anything
	} else {
	    NA
	}
    } else {
	if (c1.to.anything > c1.to.c1) {
	    c1.to.c2/(c1.to.anything - c1.to.c1 &  c1.name != c2.name)
	} else {
	    NA
	}
    }
}

#-----------------------------------------------------------------------
#  pr.mut.from
#
#  Probability that a given C2 has C1 for a mom
#-----------------------------------------------------------------------

pr.mut.from <- function (
    data,
    c1.name,
    c2.name,
    include.self = TRUE
) {
    c2.from.anything <- 0
    c2.from.c1       <- 0
    c2.from.c2       <- 0
    for (i in 1:nrow (data$Lineage)) {
        count = data$Lineage$count [i]
        if (data$Lineage$kid [i] == c2.name) {
            c2.from.anything <- c2.from.anything + count
            if (data$Lineage$mom [i] == c1.name) {
                c2.from.c1 <- c2.from.c1 + count
            }
            if (data$Lineage$mom [i] == c2.name) {
                c2.from.c2 <- c2.from.c2 + count
            }     
        }
    }
    
    if (include.self) {
	if (c2.from.anything > 0) {
	    c2.from.c1/c2.from.anything
	} else {
	    NA
	}
    } else {
	if (c2.from.anything > c2.from.c2 & c1.name != c2.name) {
	    c2.from.c1/(c2.from.anything - c2.from.c2)
	} else {
	    NA
	}
    }
}


#-----------------------------------------------------------------------
# NoDorsalCoda.from.NoCoda
#
# Find out what proportion of NoDorsalCodas are mutants from NoCoda.
#-----------------------------------------------------------------------

NoDorsalCoda.from.NoCoda <- function (
    data
) {
    NoCoda.name       <- positional.ban (data, '[ptk]', 'coda', 'F')
    NoDorsalCoda.name <- positional.ban (data, '[k]'  , 'coda', 'F')
    if (length (NoCoda.name) == 0 | length (NoDorsalCoda.name) == 0) {
        return (NA)
    } else {
        pr.mut.from (data, NoCoda.name, NoDorsalCoda.name, FALSE)
    }
}



#-----------------------------------------------------------------------
# seg.by.position
#
# Find stimuli which have (lack) a particular set of segments in a 
# particular position.
#-----------------------------------------------------------------------

seg.by.position <- function (
    data,
    segments,         # as regexp, e.g. ``[ptk]''
    position,         # 'onset', 'coda'
    polarity = '[FM]' # or '[F]' or '[M]' to exclude pos/neg stims
) {
    re <- paste (
        '^',
        polarity,
        ':',
        sep = "", collapse = ""
    )
    if (position == "onset" ) {
       re <- paste (
         re,
         segments,
         sep = "", collapse = ""
       )
    } else {
       if (position == "coda") {
           re <- paste (
               re,
               ".*",
               segments,
               "$",
               sep = "", collapse = ""
           )
       } else {
           stop (paste ("Unknown position ", position, collapse=""))
       }
    }

    grep (re, data$Stimuli$stimname, value = TRUE)
}

#-----------------------------------------------------------------------
# positional.ban
#
# Find micro-constraints corresponding to the macro-constraint ``no 
# [segments] in [position]'', and return a list of their names.
#-----------------------------------------------------------------------

positional.ban <- function (
    data,
    segments,         # as regexp, e.g. ``[ptk]''
    position,         # 'onset', 'coda'
    polarity = '[FM]' # Restrict to only pos/neg stims
) {
    not.segments <- sub ('\\[', '\\[^', segments)
    constraints.which (
        data,
        does.mark   = seg.by.position (data,     segments, position, polarity),
        doesnt.mark = seg.by.position (data, not.segments, position, polarity)
    )
}

#-----------------------------------------------------------------------
#  plot.Report
#
#  Plot the pr_corr and right/wrong data from Report.
#
#-----------------------------------------------------------------------

plot.Report <- function (
    data,
    xlim = NULL,
    ylim = NULL,
    ...
) {
    n.trials <- data$n.trials
    if (is.null (xlim)) {
        xlim = c (1, n.trials)
    }
    trials <- data$trials [1:n.trials]
    trial.cols <- data$trial.cols [1:n.trials]

    if (is.null (ylim)) {
        ylim = c (0,1)
    }

    pr_corr <- data$Report [data$Report$data_type == "pr_corr", trial.cols]    
    correct <- data$Report [data$Report$data_type == "correct", trial.cols]
    col <- sapply (
        correct,
        function (c) {
            if (is.na (c)) {NA} else {if (c==1) {'black'} else {'gray'}}
        }
    )
    plot (
        trials, pr_corr,
        xlim = xlim, ylim = ylim,
	pch = 20, col = col, cex=2/3,
	xlab = 'Trial', ylab = 'Pr (corr)',
        ...
    )
    radius <- 5
    st <- radius+1
    en <- length (trials) - radius -1
    fltr <- rep (1/(2*radius+1), 2*radius+1)
    filtered <- filter (as.numeric (pr_corr [st:en]), fltr, sides=2, method = "convolution")
    lines (trials [st:en], filtered)
    

}    


#-----------------------------------------------------------------------
#  plot.crit.stims
#
#  Given specified pos and neg stim names, plot their harmonies as a 
#  function of trial number.
#-----------------------------------------------------------------------

plot.crit.stims <- function (
    data,
    crit.pos.stims = NULL,
    crit.neg.stims = NULL,
    lty.pos = 'dashed',
    lty.neg = 'solid',
    add = FALSE,
    xlim = NULL,
    ylim = NULL,
    ...
) {

    if (is.null (crit.pos.stims)) {
        crit.pos.stims <- data$Stimuli$stimname [data$Stimuli$teach == 1]
    }

    if (is.null (crit.neg.stims)) {
        crit.neg.stims <- data$Stimuli$stimname [data$Stimuli$teach == -1]
    }


    n.trials <- data$n.trials
    if (is.null (xlim)) {
        xlim = c (1, n.trials)
    }
    trials <- data$trials [1:n.trials]
    trial.cols <- data$trial.cols [1:n.trials]

    if (is.null (ylim)) {
        ylim <- c (0, max (data$Harmony [, trial.cols]))
    }

    if (!add) {
    plot (
        0,0,
	    xlim = xlim,
	            ylim = ylim,
		    	     xlab = 'Trial',
			     	          ylab = 'Harmony',
					       	     type = 'n',
            ...
	        )
    }

    for (s in crit.pos.stims) {
        h <- data$Harmony [data$Harmony$stimulus_name == s, trial.cols]
	lines (trials, h, lty = lty.pos, ...)
    }
    for (s in crit.neg.stims) {
        h <- data$Harmony [data$Harmony$stimulus_name == s, trial.cols]
	lines (trials, h, lty = lty.neg, ...)
    }
}

#-----------------------------------------------------------------------
#  plot.crit.constraints
#
#  Given a list of critical (macro-)constraints, plot their weights as a
#  function of trial number.
#-----------------------------------------------------------------------

plot.crit.constraints <- function (
    data,
    crit.constraints = NULL,
    add = FALSE,
    xlim = NULL,
    ylim = NULL,
    ...
) {
    n.trials <- data$n.trials
    if (is.null (xlim)) {
        xlim = c (1, n.trials)
    }
    trials <- data$trials [1:n.trials]
    trial.cols <- data$trial.cols [1:n.trials]

    if (is.null (crit.constraints)) {
        crit.constraints <- data$Weight$constraint_name
    }

    if (is.null (ylim)) {
        ylim = c (0, max (data$Weight [, trial.cols]))
    }

    if (!add) {
    plot (
        0,0,
	    xlim = c (0, n.trials),
	            ylim = ylim,
		    	     xlab = 'Trial',
			     	          ylab = 'Weight',
					       	     type = 'n',
            ...
	        )
    }

    max.w <- max (data$Weight [, trial.cols])
    for (c in data$Constraints$conname) {
        c.row <- which (data$Weight$constraint_name == c)
        w <- data$Weight [c.row, trial.cols]

        # Constraints get a hue that depends on when they peaked, and
	# a width that depends on how high the peak was relative to 
	# other peaks at that time:

        peak.w  <- max (w)
        peak.at <- max (which (w == peak.w))
        w.at <- data$Weight [,trial.cols [peak.at]]	
	size <- peak.w / sum (w.at)

	if (any (crit.constraints == c)) {
   	    hue  <- 0.75 * peak.at/n.trials
            sat  <- 1
            val  <- 1
            lwd  <- 1 + 3*size
	        } else {
            hue <- 0
            sat <- 0
            val <- 9/10
            lwd <- 1
        }

        if (!add | any (crit.constraints == c)) {
   	   lines (trials [1:n.trials], w, col = hsv (hue,sat,val), lwd=lwd, ...)
           text (peak.at, peak.w, labels = c, pos = 3, cex = sqrt (sqrt (size)), 
	      	col=hsv (hue, sat, val) )
        }
    }
}    


#-----------------------------------------------------------------------
# plot.AB.crit
#
# Compare a yoked pair of simulations (A vs. B).
#-----------------------------------------------------------------------

plot.AB.crit <- function (
    either.simname,
    dirname = '.',
    plot.weights = TRUE,
    plot.harmony = TRUE,
    plot.indiv = FALSE,
    garnish = TRUE,
    ...
) {
    simname.pieces <- unlist (strsplit (either.simname, "-(A|B)-"))
    simname.A <- paste (simname.pieces [1], "A", simname.pieces [2], sep = "-")
    simname.B <- paste (simname.pieces [1], "B", simname.pieces [2], sep = "-")

    A <- tryCatch (load.report (simname.A, dirname), error=function (e){NA})
    B <- tryCatch (load.report (simname.B, dirname), error=function (e){NA})
    if (is.na (A) || is.na (B)) {
        return (NA)
    }

    A.no.coda.k <- positional.ban (A, '[k]', 'coda', 'F')
    A.no.coda.p <- positional.ban (A, '[p]', 'coda', 'F')
    A.no.coda   <- positional.ban (A, '[ptk]', 'coda', 'F')
    B.no.coda.k <- positional.ban (B, '[k]', 'coda', 'F')
    B.no.ons.p  <- positional.ban (B, '[p]', 'onset', 'F')
    B.no.coda   <- positional.ban (B, '[ptk]', 'coda', 'F')
    B.no.ons    <- positional.ban (B, '[ptk]', 'onset', 'F')

    w.A.no.coda.k  <- name.to.weights (A, A.no.coda.k)
    w.A.no.coda.p  <- name.to.weights (A, A.no.coda.p)
    w.A.no.coda    <- name.to.weights (A, A.no.coda)
    w.B.no.coda.k  <- name.to.weights (B, B.no.coda.k)
    w.B.no.ons.p   <- name.to.weights (B, B.no.ons.p)
    w.B.no.coda    <- name.to.weights (B, B.no.coda)

    if (plot.weights & plot.harmony) {
        oldpar <- par (mfrow = c(1,2))  # rows, columns
    }

    # Correlation between weights of *[Dor]]_sigma and *[Lab]]_sigma in A, vs.
    # *[Dor]]_sigma and *_sigma[[Lab] in B.

    if (plot.weights) {
    max.w <- max (unlist (c (w.A.no.coda.k, w.A.no.coda.p, w.B.no.coda.k, w.B.no.ons.p)))
    if (!is.finite (max.w)) {
        warning ("Non-finite max.w, setting to 1")
        max.w <- 1
    }

    if (garnish) {
           plot (0,0, xlim = c(0, max.w), ylim = c(0, max.w), type = 'n',
	            xlab = 'No dorsal coda', ylab = 'No labial ons/coda', main = 'Constraint weights')
        } else {
	    par (mar = c (0,0,0,0)+0.1)
            plot (0,0, xlim = c(0, max.w), ylim = c(0, max.w), type = 'n',
                 xaxt = 'n', yaxt = 'n')
        }
	lines (w.B.no.coda.k, w.B.no.ons.p, col = 'gray', lwd=2.5)
	lines (w.A.no.coda.k, w.A.no.coda.p)
    }

    if (length (w.A.no.coda.k > 1) && length (w.A.no.coda.p) == length (w.A.no.coda.k)) {
        corr.w.A <- cor (w.A.no.coda.k, w.A.no.coda.p)
    } else {
        corr.w.A <- NA
    }
    if (length (w.B.no.coda.k > 1) && length (w.B.no.ons.p) == length (w.B.no.coda.k)) {
        corr.w.B <- cor (w.B.no.coda.k, w.B.no.ons.p)
    } else {
        corr.w.B <- NA
    }
    

    # Correlation between harmonies of [uk] and [up] in A, [uk] and [pu] in B
    if (plot.harmony) {
	h.A.uk <- A$Harmony [A$Harmony$stimulus_name == 'F:uk', A$trial.cols]
	h.A.up <- A$Harmony [A$Harmony$stimulus_name == 'F:up', A$trial.cols]
	h.A.ut <- A$Harmony [A$Harmony$stimulus_name == 'F:ut', A$trial.cols]
	h.B.uk <- B$Harmony [B$Harmony$stimulus_name == 'F:uk', B$trial.cols]
	h.B.ut <- B$Harmony [B$Harmony$stimulus_name == 'F:ut', B$trial.cols]
	h.B.pu <- B$Harmony [B$Harmony$stimulus_name == 'F:pu', B$trial.cols]
	h.B.tu <- B$Harmony [B$Harmony$stimulus_name == 'F:tu', B$trial.cols]
	h.A.ut.uk <- h.A.ut-h.A.uk
	h.A.ut.up <- h.A.ut-h.A.up
	h.B.ut.uk <- h.B.ut-h.B.uk
	h.B.tu.pu <- h.B.tu-h.B.pu
	range.h <- range (-(unlist (c (h.A.ut.uk, h.A.ut.up, h.B.ut.uk, h.B.tu.pu))))

	if (garnish) {
	    plot (0,0, xlim = range.h, range.h, type = 'n',
		    xlab = 'ut-uk', ylab = 'ut-up or tu-pu', main = 'Candidate harmony difference')
	    } else {
		par (mar = c (0,0,0,0)+0.1)
		plot (0,0, xlim = range.h,  ylim = range.h, type = 'n',
		     xaxt = 'n', yaxt = 'n')
	    }


	    lines (-h.B.ut.uk, -h.B.tu.pu, col = 'gray', lwd=2.5)
	    lines (-h.A.ut.uk, -h.A.ut.up)
    }

    #  Individual stimulus and constraint plots

    if (plot.indiv) {
    quartz ()

    oldpar <- par (mfrow = c(2,2))  # rows, columns
    cex.main <- 0.75
    
#   plot.crit.stims (A, pos.A.coda, neg.A.coda.k,
#       main = simname.A, cex.main = cex.main, ...)
#	plot.crit.stims (A, pos.A.coda, neg.A.coda.p,
#			    add=TRUE, col = 'gray', ...)

#			    plot.crit.stims (B, pos.B.coda, neg.B.coda.k,
#			        main = simname.B, cex.main = cex.main, ...)
#				plot.crit.stims (B, pos.B.ons,  neg.B.ons.p,
#						    add=TRUE, col = 'gray', ...)

						            
    plot.crit.constraints (A, A.no.coda.k, ...)
    plot.crit.constraints (A, A.no.coda.p, add=TRUE,...)
    plot.crit.constraints (A, A.no.coda, add=TRUE,...)
    plot.crit.constraints (B, B.no.coda.k, ...)
    plot.crit.constraints (B, B.no.ons.p, add=TRUE, ...)
    plot.crit.constraints (B, B.no.coda, add=TRUE, ...)
    plot.crit.constraints (B, B.no.ons, add=TRUE, ...)

    par (oldpar)
    
    }    

    # Discovery of each constraint
    t.A.no.coda.k <- min (which (c (w.A.no.coda.k) > 0))
    t.A.no.coda.p <- min (which (c (w.A.no.coda.p) > 0))
    t.A.diff      <- abs (t.A.no.coda.k - t.A.no.coda.p)
    t.A.no.coda   <- min (which (c (w.A.no.coda) > 0))
    t.B.no.coda.k <- min (which (c (w.B.no.coda.k) > 0))
    t.B.no.ons.p  <- min (which (c (w.B.no.ons.p) > 0))
    t.B.diff <- abs (t.B.no.coda.k - t.B.no.ons.p)
    t.B.no.coda   <- min (which (c (w.B.no.coda) > 0))

    list (
      A=A, B=B, 
      discovery = list (
          A = c (
	        t.A.no.coda.k = t.A.no.coda.k, 
              t.A.no.coda.p = t.A.no.coda.p,
              t.A.diff = t.A.diff,
              t.A.no.coda = t.A.no.coda
          ),
          B = c (
	        t.B.no.coda.k = t.B.no.coda.k,
		      t.B.no.ons.p  = t.B.no.ons.p,
		            t.B.diff = t.B.diff,
                t.B.no.coda=t.B.no.coda
          )
       ),
      corr.w.A = corr.w.A, corr.w.B = corr.w.B
    )
}

#-----------------------------------------------------------------------
#  plot.many.AB
#
#  Make small-multiples plot of A/B critical-constraint weights or of
#  the critical-candidate harmonies
#-----------------------------------------------------------------------

plot.many.AB <- function (
    eta, mu, mut,
    type = 'weights',  # or 'harmonies'
    range = NULL
) {
    if (type == 'weights') {
        plot.weights = TRUE; plot.harmony = FALSE
    } else {
        if (type == 'harmonies') {
            plot.weights = FALSE; plot.harmony = TRUE
        } else {
            stop (paste ("Unknown type ", type))
        }
    }
    par (mfrow=c (4,3)) # (rows, columns)
    if (is.null (range)) {
        range <- 1:length (replications)
    }
    sapply (
        replications [range],
        function (i) {
	    name = make.simname (eta, mu, mut, "A", i)
#            name = paste (name.left, "A", name.right, i, sep = "", collapse = "")
            AB <- plot.AB.crit (name, ".", plot.weights, plot.harmony, FALSE, FALSE)
            if (length (AB)> 0) {
                c (AB$discovery$A, AB$discovery$B, corr.w.A = AB$corr.w.A, corr.w.B = AB$corr.w.B)
            } else {
                c (NA, NA, NA, NA, NA, NA, NA, NA)
            }
        }
    )
}

#-----------------------------------------------------------------------
#  since.peak.many
#
#  Given a specification of a positional ban, returns the trial of
#  discovery, the trial at peak weight, and the peak weight itself
#  for each of the given replication names.
#-----------------------------------------------------------------------

since.peak.many <- function (
    segments,   # regexp, e.g. ``[ptk]''
    position,   # 'onset' or 'coda'
    rep.names,  # Names of the replications
    rep.dir     # Directory with the replications in them (as subdirs)
) {
    sapply (
       rep.names,
       function (n) {
           data <- load.report (n, rep.dir)
           con <- positional.ban (data, segments, position, 'F')
           c (since=since (data, con), peak=peak (data, con), max=peak (data, con, TRUE))
       }
    )
}

since.near.peak.many <- function (
    segments,   # regexp, e.g. ``[ptk]''
    position,   # 'onset' or 'coda'
    rep.names,  # Names of the replications
    rep.dir,    # Directory with the replications in them (as subdirs)
    tol.prop = 0.05
) {
    sapply (
       rep.names,
       function (n) {
           data <- load.report (n, rep.dir)
           con <- positional.ban (data, segments, position, 'F')
           c (since=since (data, con), 
              peak=near.peak (data, con, value=FALSE, tol.prop = tol.prop), 
              max=peak (data, con, TRUE))
       }
    )
}


########################################################################
#
#  Analyses for the 2019/20 SCiL paper
#
########################################################################

# Analysis of all replications of a single simulation

single <- function (eta, mu, mut) {

    sp <- set.params (eta, mu, mut)
    rep.names.A <- sp$rep.names.A
    rep.names.B <- sp$rep.names.B

    print ("Simulation parameters:")

    # Pick a name apart (*** make sure that this matches mksh.pl ***)
    # Assumes that parameters don't change during experiment
    a.name <- rep.names.A [1]
    fields = unlist (strsplit (a.name, '-', fixed = TRUE))

    print (paste (
	c ('Exp', 'paradigm', 'ignore_this', 'N', 'adj', 
	   'zeta', 'eta', 'clutch_size', 'parth', 'mu', 'momentum',
	   'const_mutagen', 'const_recombinator', 'p_memorize', 
	   'ignore-this'),
	fields,
	sep = ' = ' 
    ))

    crit <- eta * (1-mu)/mu
    print (paste (c("n.crit = ", crit)))

    print ("*** Mean performance **************************************")

    get.prcorr <- function (
	name
    ) {
	data <- load.report (name, ".")
	rept <- data$Report
	rownum <- which (rept$data_type == "pr_corr")
	colnum <- which (names (rept) == "0")
	as.numeric (rept [rownum, colnum:ncol (rept)])
    }

    get.mean.prcorr <- function (
	name,
	last.n.trials=NULL
    ) {
	if (is.null (last.n.trials)) {
	    mean (get.prcorr (name), na.rm = TRUE)
	} else {
	    mean (rev (get.prcorr (name)) [1:last.n.trials])
	}
    }

    window <- 128
    if (TRUE) {
	mean.prcorr.A <- sapply (
	    rep.names.A,
	    function (name) {
		get.mean.prcorr (name, window)
	    }
	)
	mean.prcorr.B <- sapply (
	    rep.names.B,
	    function (name) {
		get.mean.prcorr (name, window)
	    }
	)
    }
    print (paste (
       c ("Average proportion correct over last", window, "trials"),
       collapse = " "
    ))
    print ("Condition A:")
    print (summary (mean.prcorr.A))
    print ("Condition B:")
    print (summary (mean.prcorr.B))
    print (wilcox.test (x=mean.prcorr.A, y=mean.prcorr.B))


    print ("*** Trials to discovery ************************************")
    if (TRUE) {
	pmAB <- plot.many.AB (eta, mu, mut)
    }
    print ("Condition A, NoDorsalCoda:")
    print (summary (pmAB[1,]))
    print ("Condition A, NoLabialCoda:")
    print (summary (pmAB[2,]))
    print ("Condition A, absolute difference:")
    print (summary (pmAB[3,]))
    print ("Condition B, NoDorsalCoda:")
    print (summary (pmAB[5,]))
    print ("Condition B, NoLabialOnset:")
    print (summary (pmAB[6,]))
    print ("Condition B, absolute difference:")
    print (summary (pmAB[7,]))
    print ("Absolute difference B-A:")
    print (summary (pmAB[7,]-pmAB[3,]))
    print ("Wilcoxon test of absolute difference in A vs absolute difference in B")
    print (wilcox.test (x=pmAB[3,], y=pmAB[7,], conf.int=TRUE))

    print ("*** Correlation of weights *********************************")
    if (TRUE) {
	corr.w <- sapply (
	    rep.names.A, 
	    function (name) {
		pABc <- plot.AB.crit (name)
		c (pABc$corr.w.A, pABc$corr.w.B)
	    }
	)
    }
    print ("Condition A, within-participant correlations between w(NoDorsalCoda) and w(NoLabialCoda)")
    print (summary (as.numeric (corr.w [1,])))
    print ("Condition B, within-participant correlations between w(NoDorsalCoda) and w(NoLabialOnset)")
    print (summary (as.numeric (corr.w [2,])))
    print (
	wilcox.test (
	    x = as.numeric (corr.w [1,]),
	    y = as.numeric (corr.w [2,]),
	)
    )


    print ("*** Discovery and growth of NoCoda *******************")
    if (TRUE) {
	spmA <- since.peak.many ('[ptk]', 'coda', rep.names.A, '.')
	spmB <- since.peak.many ('[ptk]', 'coda', rep.names.B, '.')
    }
    print ("Condition A:")
    print ("    Time to discover NoCoda:")
    print (summary (spmA[1,]))
    print ("    NoCoda attains peak population on trial:")
    print (summary (spmA[2,]))
    print ("    Peak weight (pop. times weight quantum) is:")
    print (summary (spmA[3,]))
    print ("Condition B:")
    print ("    Time to discover NoCoda:")
    print (summary (spmB[1,]))
    print ("    NoCoda attains peak population on trial:")
    print (summary (spmB[2,]))
    print ("    Peak weight (pop. times weight quantum) is:")
    print (summary (spmB[3,]))




    print ("*** NoCoda as parent of NoDorsalCoda *********************")
    print ("Difference in time to discover NoDorsalCoda:")
    print ("Condition A, NoDorsalCoda:")
    print (summary (pmAB[1,]))
    print ("Condition B, NoDorsalCoda:")
    print (summary (pmAB[5,]))
    print (wilcox.test (x=pmAB[1,], y=pmAB[5,]))

    if (TRUE) {
	nc.A <- sapply (
	    1:length (rep.names.A), 
	    function (i) {
		NoDorsalCoda.from.NoCoda (load.report (rep.names.A[i]))
	    }
	)
	nc.B <- sapply (
	    1:length (rep.names.B), 
	    function (i) {
		NoDorsalCoda.from.NoCoda (load.report (rep.names.B[i]))
	    }
	)
    }
    print ("Proportion of NoDorsalCodas that have NoCoda as a parent:")
    print ("Condition A:")
    print (summary (nc.A))
    print ("Condition B:")
    print (summary (nc.B))
    print (wilcox.test (nc.A, nc.B, conf.int=TRUE))
}

# Analysis of final performance as a function of critical number

n.crit <- function (
    eta, 
    mu
) {
    eta * (1-mu)/mu
}

final.prcorr <- function (
    simname,
    final.n = 1
) {
    data <- load.report (simname)
    if (is.na (data)) {
        return (NA)
    }
    n.trials <- data$n.trials
    trial.cols <- data$trial.cols [1:n.trials]
    pr_corr <- data$Report [data$Report$data_type == "pr_corr", trial.cols]    
    mean (as.numeric (rev (pr_corr) [1:final.n]))
}

prcorr.by.n.crit <- function (
    etas,
    mus,
    replications,
    mut,
    condition,
    final.n = 16
) {
    df <- expand.grid (
        eta = etas,
        mu  = mus,
        replications = replications
    )
    df$n.crit <- sapply (
        1:nrow (df),
        function (i) {
            n.crit (df$eta[i], df$mu[i])
        }
    )
    df$final.prcorr <- sapply (
        1:nrow (df),
        function (i) {
            final.prcorr (
                make.simname (
                    df$eta [i], df$mu [i], mut, condition, df$replication [i]
                ), final.n
            )
         }
    )
    df    
}

# Plot data frame returned by prcorr.by.n.crit
plot.pbnc <- function (
    pbnc
) {
    bp <- boxplot (
        final.prcorr~n.crit, data=pbnc, log='x', boxwex=1/12, notch=TRUE, pch=20, cex=1/2, 
	xlab = 'n.crit (log scale)', ylab = 'Pr (corr), last 16 trials', xaxt='n'
    )
    axis (1, at=1:length (bp$names), labels = sprintf ("%3.2f", as.numeric (bp$names)))
    abline (v=7.5)  # Vertical line at n.crit = 2
}

