#!/usr/bin/perl
########################################################################
#
#  mksh.pl
#
#  Make the files and commands needed for the Reuse9 simulations.
#
#  E. Moreton  *  UNC-Chapel Hill Linguistics  *  2019 June 24 (M)
#
########################################################################
use strict;
use warnings;
use Carp;

# For finding the index of an element in an array
# For randomly shuffling a list
use List::Util qw/ first shuffle /;

# For pasting pathnames together
use File::Spec;
use FindBin;
use lib File::Spec->catdir ($FindBin::Bin, "../../..", "Current", 'lib');

use Stimulus;
use StimulusSet;
use GJ2005jrFM;

########################################################################
#
#  Global parameters
#
########################################################################

# Where to find the simulation software
my $Simulate = "../../Current/Simulate.pl";

# Where we are now
open (PWD, "pwd |") or confess "Couldn't read output from pwd";
my $pwd = <PWD>;
# E.g., /nas/longleaf/home/moreton/GCL-exps/Current-SHJ/Exp07
$pwd =~ /^.*\/Exp(\d+)$/ or confess "Couldn't parse directory path $pwd";
my $expdigits = $1;
# my $expdigits = "0";

# Read in the stimuli
open (STIMS, "< All.stims") or confess ("Couldn't read data from All.stims");
my ($stims) = &StimulusSet::from_text (
    <STIMS>
    ) or confess ("Couldn't load stimuli from All.stims");
my @stimnames = map {
    $_->get_name
} ($stims->get_all);

# The correct category labels of the words for each pattern
my $Label;
for (@stimnames) {
    /^(.):(.+)$/ or die "Couldn't parse stimulus name $_: $!";
    my ($gender, $word) = ($1, $2);
    # Pattern A:  No labial codas, no dorsal codas
    if ($word =~ /p$/ or $word =~ /k$/) {
	$Label->{$word}->{A} = 'M';
    } else {
	$Label->{$word}->{A} = 'F';
    }
    # Pattern B:  No labial onsets, no dorsal codas
    if ($word =~ /^p/ or $word =~ /k$/) {
	$Label->{$word}->{B} = 'M';
    } else {
	$Label->{$word}->{B} = 'F';
    }
}
my @Words = keys %{$Label};


# Column headers (last umpteen)
my @Columns = qw/
    subjid paradigm pattern phase block trial N adj zeta eta clutch_size
    parth mu momentum 
    const_mutagen const_recombinator metaconstraints 
    p_memorize mem_polarity 
    replication 
    correct_stim foil
/;

# Columns used for subjid
my @Subjid_Columns = qw/
    paradigm pattern N adj zeta eta clutch_size 
    parth mu momentum 
    const_mutagen const_recombinator p_memorize replication
/;

########################################################################
#
#  Experiment-specific parameters
#
########################################################################

#my @paradigms = qw/ Label 2AFC Predict /;
my @paradigms = qw/ 2AFC /;
# Number of training blocks (SHJ1961 and Nosofsky et al. 1994 used 400
# trials as a maximum)
my $N_trials = 2048;
my %n_blocks = (
    'Label'    =>   $N_trials/16,  # Label each of 8 pos, 8 neg
    '2AFC'     =>   $N_trials/64,  # Choose pos from 8 x 8 pos-neg pairs
    'Predict'  =>   $N_trials/8    # Predict next stim; is it 1 of 8 pos?
    );
my @patterns = qw/ A B /;
my @Ns = (200);
my @adjs = (
#    'fitness', 
#    'fitnessDraconian', 
#    'fitnessBinary', 
#    'validity', 
#    'validityDraconian', 
#    'validityBinary'
     'random'
);
my @zetas=(0.05);
my @etas = (0.10, 0.15, 0.25, 0.30);
my @parths = (1);
my @mus = (0.025, 0.05, 0.1, 0.15);
my @momenta = (0);
my @clutch_sizes = (1);
my @const_mutagens = qw/
  Mut_0.005
 /;
my @const_recombinators = qw/ Rec.1.00 /;
my @p_mems = (0);

my @replications = (($expdigits * 100 + 00)..($expdigits * 100 + 99));
#my @replications = ($expdigits*100);   # SCiL

my $params;

# When memorizing a trial, memorize winner rather than loser
$params->{mem_polarity} = +1;


########################################################################
# 
#  Subroutines
#
########################################################################

# Given a pronunciation and a pattern, get the label
sub label_word {
    my (
	$word,
	$pattern
	) = @_;

    $Label->{$word}->{$pattern};
}

# Given "+", return "-", and vice versa
sub invert {
    my $label = shift;
    if ($label eq '+') {
	return '-'
    }
    if ($label eq '-') {
	return '+'
    } 
    if ($label eq 'pos') {
	return 'neg' 
    }
    if ($label eq 'neg') {
        return 'pos'
    }
    if ($label eq 'M') {
	return 'F' 
    }
    if ($label eq 'F') {
        return 'M'
    }
    confess "Argument to &invert must be + or -, pos or neg: $!";
}

# Given a pattern, get positive or negative words
sub get_words {
    my (
	$pattern, 
	$label
	) = @_;
    
    unless ($label eq '+' or $label eq '-'
	or $label eq 'F' or $label eq 'M') {
	confess "$label should be + or -, F or M: $!";
    }

    grep {
	$label eq &label_word ($_, $pattern);
    } @Words;
}

# Return pos-neg stimulus pairs, pairing each positive
# with each negative
sub make_2AFC_pairs {
    my $pattern = shift;

    my @pos = &get_words ($pattern, 'F');
    my @neg = &get_words ($pattern, 'M');

    my @pairs;
    for my $p (@pos) {
	for my $n (@neg) {
	    push @pairs, "F:$p,F:$n";
	}
    }

    @pairs;
}

# Return pos-neg stimulus pairs, one for each word
sub make_labelling_pairs {
    my $pattern = shift;
    map {
        my $label = &label_word ($_, $pattern);
        "$label:$_," . &invert ($label) . ":$_";
    } @Words;
}


sub make_subjid {
    my $params_ref = shift;
    my %params = %{$params_ref};
    "Exp$expdigits-" . (join "-", @params {@Subjid_Columns});
}


# Generate one entire Experiment.csv file for a classification
# (labelling) experiment
sub LabellingExperiment {
    my (
	$n_blocks,    # How many blocks?
	$params_ref   # Other params
	) = @_;
    
    my %params = %{$params_ref};
    my @exp;

    # Headers
    push @exp, (join ",", @Columns);

    # Trials
    my $trial_num = 1;
    for my $block_num (1..$n_blocks) {
	$params {block} = $block_num;
	for my $pair (&shuffle (&make_labelling_pairs ($params {pattern}))) {
	    my ($correct_stim, $foil) = split ",", $pair;
	    $params {correct_stim} = $correct_stim;
	    $params {foil} = $foil;
	    $params {trial} = $trial_num++;
	    my $trial_descriptor = join ",", @params {@Columns};
	    push @exp, $trial_descriptor;
	}
    }

    @exp;
}

# Generate whole Experiment.csv file for a 2AFC experiment
sub TwoAFCExperiment {
    my (
	$n_blocks,    # How many blocks?
	$params_ref   # Other params
	) = @_;
    
    my %params = %{$params_ref};
    my @exp;

    # Headers
    push @exp, (join ",", @Columns);

    # Trials
    my $trial_num = 1;
    for my $block_num (1..$n_blocks) {
	$params {block} = $block_num;
	for my $pair (&shuffle (&make_2AFC_pairs ($params {pattern}))) {
	    my ($correct_stim, $foil) = split ",", $pair;
	    $params {correct_stim} = $correct_stim;
	    $params {foil} = $foil;
	    $params {trial} = $trial_num++;
	    my $trial_descriptor = join ",", @params {@Columns};
	    push @exp, $trial_descriptor;
	}
    }

    @exp;
}


# Generate whole Experiment.csv file for a prediction (unsupervised) experiment
sub PredictionExperiment {
    my (
	$n_blocks,    # How many blocks?
	$params_ref   # Other params
	) = @_;
    
    my %params = %{$params_ref};
    my @exp;

    # Headers
    push @exp, (join ",", @Columns);

    # Trials
    my $trial_num = 1;
    for my $block_num (1..$n_blocks) {
	$params {block} = $block_num;
	for my $correct_stim (&shuffle (&get_words ($params{pattern}, 'F'))) {
	    $params {correct_stim} = "+:$correct_stim";
	    $params {foil} = "NA";
	    $params {trial} = $trial_num++;
	    my $trial_descriptor = join ",", @params {@Columns};
	    push @exp, $trial_descriptor;
	}
    }

    @exp;
}

# Convert seconds to hh:mm:ss (without leading zeros)
sub seconds_to_HMS {
    my $seconds_needed = shift;

    my $hours = int ($seconds_needed/3600);
    $seconds_needed = $seconds_needed % 3600;
    my $minutes = int ($seconds_needed/60);
    my $seconds = $seconds_needed % 60;

    ($hours, $minutes, $seconds)
}

# Because the Math library isn't installed
sub cartesian_product {
    # Given a list of refs to sets, returns list of all possible strings
    # produced by concatenating one element from the first list, one 
    # from the second, etc.
    #
    # If the first argument is a scalar, it will be used as the
    # separation character in concatenation.
    my $sep_char = "";
    unless (ref ($_[0])) {
	$sep_char = shift;
    }
    my @set_refs = @_;

    # Basis case:  Cartesian product of an empty set is null string
    unless (@set_refs) {
	return ('');
    }
    
    # Induction:  
    my @product;
    for my $least_significant (@{shift @set_refs}) {
	for (&cartesian_product ($sep_char, @set_refs)) {
	    unshift @product, "$least_significant$sep_char$_";
	}
    }

    @product;
}

  

########################################################################
#
#  MAIN
#
########################################################################

my @expdefs = &cartesian_product (
    " ",
    \@paradigms, \@patterns, \@Ns, \@adjs, \@zetas, \@etas, \@clutch_sizes,
    \@parths, \@mus, \@momenta,
    \@const_mutagens, \@const_recombinators, \@p_mems, \@replications);
for my $expdef (@expdefs) {
    my ($paradigm, $pattern, $N, $adj, $zeta, $eta, $clutch_size, 
	$parth, $mu, $momentum,
	$const_mutagen, $const_recombinator, $p_mem, $replication) =
	    split /\s+/, $expdef;

    $params->{paradigm} = $paradigm;
    $params->{metaconstraints} = "meta.$paradigm";
    $params->{pattern} = $pattern;
    $params->{N} = $N;
    $params->{adj} = $adj;
    if ($zeta=~/total (\d+)/) {
	my $total_harmony = $1;
	$params->{zeta} = sprintf "%4.2f", $total_harmony/$params->{N};
    } else {
	$params->{zeta} = $zeta;
    }
    $params->{eta} = $eta;
    $params->{parth} = $parth;
    $params->{mu} = $mu;
    $params->{clutch_size} = $clutch_size;
    $params->{momentum} = $momentum;
    $params->{const_mutagen} = $const_mutagen;
    $params->{const_recombinator} = $const_recombinator;
    $params->{p_memorize} = $p_mem;
    $params->{replication} = $replication;
    
    my $subjid = &make_subjid ($params);
    $params->{subjid} = $subjid;

    # Make the training part of the experiment
    my $n_blocks = $n_blocks {$params->{paradigm}};
    my @training;
    $params->{phase} = 'Train';
    if ($params->{paradigm} eq 'Label') {
	@training= &LabellingExperiment ($n_blocks, $params);
    } elsif ($params->{paradigm} eq '2AFC') {
	@training= &TwoAFCExperiment ($n_blocks, $params);
    } elsif ($params->{paradigm} eq 'Predict') {
	@training = &PredictionExperiment ($n_blocks, $params);
    } else {
	confess "Unrecognized experimental paradigm " . 
	    $params->{paradigm};
    }
    
    ### MAKE TEST PHASE HERE ###
    
    my @experiment = (@training);
    
#    $params->{eta} = $eta;
#    $params->{const_mutagen} = $const_mutagen;
#    $params->{p_memorize} = $p_mem;
#    $params->{parth} = $parth;
#    $params->{mu} = $mu;
#    $params->{momentum} = $momentum;
    
    mkdir File::Spec->catdir (".", $subjid), 0755;
    my $expfile_path = File::Spec->catdir (".", $subjid, "Experiment.csv");

    if (-e $expfile_path) {
	warn "$expfile_path exists:  skipping ...";
	next;
    }
    unless (open (EXPFILE, "> $expfile_path")) {
	confess "Couldn't open $expfile_path for writing";
    }

    print EXPFILE (join "\n", @experiment);
    print EXPFILE "\n";
    close EXPFILE;
    
    my ($hours, $minutes, $seconds) = (
	23, 59, 59
	);
    
    # Make the sh file
    my $sh_path = File::Spec->catdir (".", $subjid, "Simulate.sh");
    open (SHFILE, "> $sh_path") or
	confess "Couldn't open $sh_path for writing";
    select SHFILE;
    print <<"MKSH";
#!/bin/bash
#SBATCH --ntasks=1
#SBATCH --time=$hours:$minutes:$seconds
#SBATCH --mem=1g
MKSH
    select STDOUT;
    print SHFILE join " ", ("perl $Simulate", "-repscheme GJ2005jrFM",  
			    "-initcon 0.const", 
			    "-stims All.stims", 
			    "-exppath .", 
			    "-subjid $subjid",
			    "-mutagens Const.mut", "-recombinators Const.rec",
			    "-interval 1 " ,
			    "-log_format flat ",
			    " -meta " . $params->{metaconstraints} . "\n");
    close SHFILE;				
    
    print "sbatch $sh_path\n";
}
