#!/usr/bin/perl
########################################################################
#
#  mksh.pl
#
#  Make the files and commands needed for the SHJ1961wug simulations.
#
#  E. Moreton  *  UNC-Chapel Hill Linguistics  *  2019 March 4 (T)
#
########################################################################
use strict;
use warnings;
use Carp;

# For finding the index of an element in an array
# For randomly shuffling a list
use List::Util qw/ first shuffle /;

# For pasting pathnames together
use File::Spec;

########################################################################
#
#  Global parameters
#
########################################################################

# Where to find the simulation software
my $Simulate = "/nas/longleaf/home/moreton/Current/Simulate.pl";

# Where we are now
open (PWD, "pwd |") or confess "Couldn't read output from pwd";
my $pwd = <PWD>;
# E.g., /nas/longleaf/home/moreton/GCL-exps/Current-SHJ/Exp07
$pwd =~ /^.*\/Exp(\d+)$/ or confess "Couldn't parse directory path $pwd";
my $expdigits = $1;
#my $expdigits = "0";

# The eight SHJ shapes
my @Shape = qw /
    +++ ++- +-+ +-- -++ -+- --+ ---
/;

# The category labels of those shapes for each SHJ Type
my %Label = (
#           + + + + - - - -
#           + + - - + + - -
#           + - + - + - + -
    I   => '+ + + + - - - -',
    II  => '+ + - - - - + +',
    III => '+ + - + - - - +',
    IV  => '+ + + - + - - -',
    V   => '+ + + - - - - +',
    VI  => '+ - - + - + + -'
);

# Column headers (last umpteen)
my @Columns = qw/
    subjid paradigm type phase block trial N adj zeta eta clutch_size
    parth mu momentum 
    const_mutagen const_recombinator metaconstraints 
    p_memorize mem_polarity 
    replication 
    correct_stim foil
/;

# Columns used for subjid
my @Subjid_Columns = qw/
    paradigm type N adj zeta eta clutch_size 
    parth mu momentum 
    const_mutagen const_recombinator p_memorize replication
/;

########################################################################
#
#  Experiment-specific parameters
#
########################################################################

#my @paradigms = qw/ Label 2AFC Predict /;
my @paradigms = qw/ Label /;
# Number of training blocks (SHJ1961 and Nosofsky et al. 1994 used 400
# trials as a maximum)
my %n_blocks = (
    'Label' =>   50,  # Each block is  8 trials long
    '2AFC'  =>   25,  #               16
    'Predict' =>100,  #                4
    );
# SHJ pattern types:
my @types = qw/ I II IV /;
# Micro-constraint population size:
my @Ns = (
   7
);
# Population-adjustment procedures:
my @adjs = (
    'fitness', 
#    'fitnessDraconian', 
#    'fitnessBinary', 
#    'validity', 
#    'validityDraconian', 
#    'validityBinary',
#    'random'
);
# Weight quanta:
my @zetas=(12);
# Learning rates:
my @etas = (1);
# Probabilities of parthenogenesis:
my @parths = (1);
# Mutation rates:
my @mus = (1);
# Momenta (mixing previous fitness with current fitness):
my @momenta = (0);
# Clutch sizes ("s" in paper)
my @clutch_sizes = (12);
# Names of constraint mutagens:
my @const_mutagens = qw/
  Mut_0.5
 /;
# Names of constraint recombinators:
my @const_recombinators = qw/ 
  Rec.1.00
 /;
# Probabilities of memorizing winnter/loser
my @p_mems = (
  0
);

my @replications = (($expdigits * 100 + 0)..($expdigits * 100 + 4));
#my @replications = ($expdigits*100);

my $params;

# When memorizing a trial, memorize winner rather than loser
$params->{mem_polarity} = +1;


########################################################################
# 
#  Subroutines
#
########################################################################

# Given a shape and a type, get the label
sub label_shape {
    my (
	$shape,
	$type
	) = @_;

    my $i = first {$Shape [$_] eq $shape} 0..7;
    my @labels = split /\s+/, $Label {$type};
    $labels [$i];
}

# Given "+", return "-", and vice versa
sub invert {
    my $label = shift;
    if ($label eq '+') {
	return '-'
    }
    if ($label eq '-') {
	return '+'
    } 
    confess "Argument to &invert must be + or -: $!";
}

# Given a type, get positive or negative shapes
sub get_shapes {
    my (
	$type, 
	$label
	) = @_;
    
    unless ($label eq '+' or $label eq '-') {
	confess "$label should be + or -: $!";
    }

    grep {
	$label eq &label_shape ($_, $type);
    } @Shape;
}

# Return 16 pos-neg stimulus pairs, pairing each positive
# with each negative
sub make_2AFC_pairs {
    my $type = shift;

    my @pos = &get_shapes ($type, '+');
    my @neg = &get_shapes ($type, '-');

    my @pairs;
    for my $p (@pos) {
	for my $n (@neg) {
	    # Add useless + label
	    push @pairs, "+:$p,+:$n";
	}
    }

    @pairs;
}

# Return 8 pos-neg stimulus pairs, one for each shape
sub make_labelling_pairs {
    my $type = shift;
    map {
	my $label = &label_shape ($_, $type);
        "$label:$_," . &invert ($label) . ":$_";
    } @Shape;
}

sub make_subjid {
    my $params_ref = shift;
    my %params = %{$params_ref};
    "Exp$expdigits-" . (join "-", @params {@Subjid_Columns});

}


# Generate one entire Experiment.csv file for a classification
# (labelling) experiment
sub LabellingExperiment {
    my (
	$n_blocks,    # How many 8-trial blocks?
	$params_ref   # Other params
	) = @_;
    
    my %params = %{$params_ref};
    my @exp;

    # Headers
    push @exp, (join ",", @Columns);

    # Trials
    my $trial_num = 1;
    for my $block_num (1..$n_blocks) {
	$params {block} = $block_num;
	for my $pair (&shuffle (&make_labelling_pairs ($params {type}))) {
	    my ($correct_stim, $foil) = split ",", $pair;
	    $params {correct_stim} = $correct_stim;
	    $params {foil} = $foil;
	    $params {trial} = $trial_num++;
	    my $trial_descriptor = join ",", @params {@Columns};
	    push @exp, $trial_descriptor;
	}
    }

    @exp;
}

# Generate whole Experiment.csv file for a 2AFC experiment
sub TwoAFCExperiment {
    my (
	$n_blocks,    # How many 16-trial blocks?
	$params_ref   # Other params
	) = @_;
    
    my %params = %{$params_ref};
    my @exp;

    # Headers
    push @exp, (join ",", @Columns);

    # Trials
    my $trial_num = 1;
    for my $block_num (1..$n_blocks) {
	$params {block} = $block_num;
	for my $pair (&shuffle (&make_2AFC_pairs ($params {type}))) {
	    my ($correct_stim, $foil) = split ",", $pair;
	    $params {correct_stim} = $correct_stim;
	    $params {foil} = $foil;
	    $params {trial} = $trial_num++;
	    my $trial_descriptor = join ",", @params {@Columns};
	    push @exp, $trial_descriptor;
	}
    }

    @exp;
}


# Generate whole Experiment.csv file for a prediction (unsupervised) experiment
sub PredictionExperiment {
    my (
	$n_blocks,    # How many 16-trial blocks?
	$params_ref   # Other params
	) = @_;
    
    my %params = %{$params_ref};
    my @exp;

    # Headers
    push @exp, (join ",", @Columns);

    # Trials
    my $trial_num = 1;
    for my $block_num (1..$n_blocks) {
	$params {block} = $block_num;
	for my $correct_stim (&shuffle (&get_shapes ($params{type}, '+'))) {
	    $params {correct_stim} = "+:$correct_stim";
	    $params {foil} = "NA";
	    $params {trial} = $trial_num++;
	    my $trial_descriptor = join ",", @params {@Columns};
	    push @exp, $trial_descriptor;
	}
    }

    @exp;
}

# Convert seconds to hh:mm:ss (without leading zeros)
sub seconds_to_HMS {
    my $seconds_needed = shift;

    my $hours = int ($seconds_needed/3600);
    $seconds_needed = $seconds_needed % 3600;
    my $minutes = int ($seconds_needed/60);
    my $seconds = $seconds_needed % 60;

    ($hours, $minutes, $seconds)
}

# Because the Math library isn't installed
sub cartesian_product {
    # Given a list of refs to sets, returns list of all possible strings
    # produced by concatenating one element from the first list, one 
    # from the second, etc.
    #
    # If the first argument is a scalar, it will be used as the
    # separation character in concatenation.
    my $sep_char = "";
    unless (ref ($_[0])) {
	$sep_char = shift;
    }
    my @set_refs = @_;

    # Basis case:  Cartesian product of an empty set is null string
    unless (@set_refs) {
	return ('');
    }
    
    # Induction:  
    my @product;
    for my $least_significant (@{shift @set_refs}) {
	for (&cartesian_product ($sep_char, @set_refs)) {
	    unshift @product, "$least_significant$sep_char$_";
	}
    }

    @product;
}

  

########################################################################
#
#  MAIN
#
########################################################################

my @expdefs = &cartesian_product (
    " ",
    \@paradigms, \@types, \@Ns, \@adjs, \@zetas, \@etas, \@clutch_sizes,
    \@parths, \@mus, \@momenta,
    \@const_mutagens, \@const_recombinators, \@p_mems, \@replications);
for my $expdef (@expdefs) {
    my ($paradigm, $type, $N, $adj, $zeta, $eta, $clutch_size, 
	$parth, $mu, $momentum,
	$const_mutagen, $const_recombinator, $p_mem, $replication) =
	    split /\s+/, $expdef;

    $params->{paradigm} = $paradigm;
    $params->{metaconstraints} = "meta.$paradigm";
    $params->{type} = $type;
    $params->{N} = $N;
    $params->{adj} = $adj;
    if ($zeta=~/total (\d+)/) {
	my $total_harmony = $1;
	$params->{zeta} = sprintf "%4.2f", $total_harmony/$params->{N};
    } else {
	$params->{zeta} = $zeta;
    }
    $params->{eta} = $eta;
    $params->{parth} = $parth;
    $params->{mu} = $mu;
    $params->{clutch_size} = $clutch_size;
    $params->{momentum} = $momentum;
    $params->{const_mutagen} = $const_mutagen;
    $params->{const_recombinator} = $const_recombinator;
    $params->{p_memorize} = $p_mem;
    $params->{replication} = $replication;
    
    my $subjid = &make_subjid ($params);
    $params->{subjid} = $subjid;

    # Make the training part of the experiment
    my $n_blocks = $n_blocks {$params->{paradigm}};
    my @training;
    $params->{phase} = 'Train';
    if ($params->{paradigm} eq 'Label') {
	@training= &LabellingExperiment ($n_blocks, $params);
    } elsif ($params->{paradigm} eq '2AFC') {
	@training= &TwoAFCExperiment ($n_blocks, $params);
    } elsif ($params->{paradigm} eq 'Predict') {
	@training = &PredictionExperiment ($n_blocks, $params);
    } else {
	confess "Unrecognized experimental paradigm " . 
	    $params->{paradigm};
    }
    
    ### MAKE TEST PHASE HERE ###
    
    my @experiment = (@training);
    
#    $params->{eta} = $eta;
#    $params->{const_mutagen} = $const_mutagen;
#    $params->{p_memorize} = $p_mem;
#    $params->{parth} = $parth;
#    $params->{mu} = $mu;
#    $params->{momentum} = $momentum;
    
    mkdir File::Spec->catdir (".", $subjid), 0755;
    my $expfile_path = File::Spec->catdir (".", $subjid, "Experiment.csv");
    open (EXPFILE, "> $expfile_path") or 
	confess "Couldn't open $expfile_path for writing";
    print  EXPFILE (join "\n", @experiment);
    print EXPFILE "\n";
    close EXPFILE;
    
    my ($hours, $minutes, $seconds) = 
	&seconds_to_HMS (
	60*60-1
	);
    
    # Make the sh file
    my $sh_path = File::Spec->catdir (".", $subjid, "Simulate.sh");
    open (SHFILE, "> $sh_path") or
	confess "Couldn't open $sh_path for writing";
    select SHFILE;
    print <<"MKSH";
#!/bin/bash
#SBATCH --ntasks=1
#SBATCH --time=$hours:$minutes:$seconds
#SBATCH --mem=1g
MKSH
    select STDOUT;
    print SHFILE join " ", ("perl $Simulate", "-repscheme SHJ1961wug",  
			    "-initcon 0.const", 
			    "-stims SHJ1961wug.stims", 
			    "-exppath .", 
			    "-subjid $subjid",
			    "-mutagens Const.mut", "-recombinators Const.rec",
			    "-interval 1 " ,
			    "-log_format flat ",
			    " -meta " . $params->{metaconstraints} . "\n");
    close SHFILE;				
    
    print "sbatch $sh_path\n";
}
