#!/usr/bin/perl
########################################################################
#
#  whereperf.pl
#
#  Find the first block of trials in each simulation run for which
#  PrCorr on every stimulus exceeds a given criterion (0.99 by default).
#
#  Also find the first n (= 4 by default) consecutive blocks in which
#  the simulated participant gave the right answer every time.
#
#  E. Moreton  *  UNC-Chapel Hill Linguistics  *  2019 March 30 (S)
#
########################################################################

use strict;
use warnings;
use Carp;

my $perfection = 0.99;    # Criterion for first "perfect" block
my $criterion  = 4;

# Has this one even run yet?
sub has_log {
    my $simname = shift;

    return 1 if (open (LOG, "< $simname/Log.txt"));
}

# Has it finished running?
sub has_data {
    my $simname = shift;

    return 1 if (open (DATA, "< $simname/Data.csv"));
}
	

# Find first block where all prcorrs exceed $crit
sub first_perfect_block {
    my ($simname, $crit) = @_;

    unless (open (DATA, "< $simname/Data.csv")) {
	 carp "Couldn't open $simname/Data.csv --- skipping";
#	confess "Couldn't open $simname/Data.csv";
	 return
    }
    
    my $header_line = <DATA>;
    chomp $header_line;
    my @headers = split (/,/, $header_line);
    unless (@headers) {
	carp "Couldn't find header line in $simname/Data.csv --- skipping";
	return;
    }
    
    my %data;
    my $current_block = 0;
    my $is_perfect = 0;
    while (<DATA>) {
	next unless /Train/;
	@data {@headers} = split /,/;

	unless (defined $data{pr_corr}) {
	    carp "Corrupt data line in $simname/Data.csv:\$_---skipping";
	    return;
	}

	if ($data{'block'} == $current_block) {
	    if ($data {'pr_corr'} < $crit) {
		$is_perfect = 0;
	    }
	} else {
	    if ($is_perfect) {
		last;
	    }
	    $current_block = $data{'block'};
	    if ($data{'pr_corr'} < $crit) {
		$is_perfect = 0;
	    } else {
		$is_perfect = 1;
	    }
	}
    }
    if ($is_perfect) {
	return $current_block;
    } else {
	return "NA";
    }
}

# Find first $crit consecutive blocks where the responses were all
# correct.
sub first_consecutive {
    my ($simname, $crit) = @_;

    unless (open (DATA, "< $simname/Data.csv")) {
	 carp "Couldn't open $simname/Data.csv --- skipping";
#	confess "Couldn't open $simname/Data.csv";
	 return
    }
    
    my $header_line = <DATA>;
    chomp $header_line;
    my @headers = split (/,/, $header_line);
    unless (@headers) {
	carp "Couldn't find header line in $simname/Data.csv --- skipping";
	return;
    }
    
    my %data;
    my $current_block = 0;
    my $is_perfect = 0;
    my $consecutive_perfect_blocks = 0;
    while (<DATA>) {
	next unless /Train/;
	chomp;
	@data {@headers} = split /,/;

	unless (defined $data{pr_corr}) {
	    carp "Corrupt data line in $simname/Data.csv: $_ --- skipping";
	}

	if ($data{'block'} == $current_block) {
	    $is_perfect *= $data {'corr'};
	} else {
	    if ($is_perfect) {
		$consecutive_perfect_blocks++;
		last if ($consecutive_perfect_blocks >= $crit);
	    } else {
		$consecutive_perfect_blocks = 0;
	    }
	    $current_block = $data{'block'};
	    $is_perfect = $data{'corr'};
	}
    }
    if ($consecutive_perfect_blocks >= $crit) {
	return $current_block - $consecutive_perfect_blocks + 1;
    } else {
	return "NA";
    }
}


if (@ARGV) {
    $criterion = shift @ARGV;
}
print "Exp,Task,Type,N,adj,zeta,eta,clutch_size,parth,mu,momentum,Mut,Rec,mem,rep,consec,perf\n";

while (<STDIN>) {
    chomp;
    my $simname = $_;
    $simname =~ /^(.*)-(\d+)$/;

    my ($cond, $rep) = ($1, $2);
    $cond =~ s/\-/,/g;

    next unless (&has_data ($simname));

    my $first = &first_perfect_block ($simname, $perfection);
    my $consec = &first_consecutive ($simname, $criterion);
    
#    next unless (defined $first and defined $consec);

    print "$cond,$rep,$consec,$first\n";
}

