#!/usr/bin/perl
use strict;
use warnings;
use Carp;

# Enables finding other modules in the same directory, even when invoked 
# from elsewhere:
# https://perlmaven.com/how-to-add-a-relative-directory-to-inc
use FindBin;
use File::Spec;
use lib File::Spec->catdir($FindBin::Bin, '.', 'lib');

# CSV reader
# use Text::CSV;
# except that longleaf.its.unc.edu doesn't have it

use Constraint;
use Learner;
use MutagenSet;
use RecombinatorSet;
use Report;

########################################################################
#
#  Simulate.pl
#
#  Run the Winnow-MaxEnt-Subtree learner in an experiment.
#
#  E. Moreton  *  UNC-Chapel Hill Linguistics  *  2019 February 1 (F)
#
########################################################################
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################

# Separator character for reading the experiment file
my $sepchar = ",";

my %options = @ARGV;

########################################################################
sub process_options {
    unless (defined $options{-exppath}) {
	$options{-exppath} = ".";
    }
    unless (defined $options{-subjid}) {
	$options{-subjid} = ".";
    }
    unless (defined $options{-simname}) {
	$options{-simname} = time () . "_" . rand ();
    }
    unless (defined $options{-repscheme}) {
	confess "No representation scheme specified (-repscheme)";
    }
    unless (defined $options{-initcon}) {
	confess "No initial constraint population specified (-initcon)";
    }
    unless (defined $options{-stims}) {
	confess "No stimulus space specified (-stims)";
    }
    unless (defined $options{-mutagens}) {
	confess "No mutagens specified (-mutagens)";
    }
    unless (defined $options{-recombinators}) {
	confess "No recombinators specified (-recombinators)";
    }
    # Report interval defaults to 1
    unless (defined $options{-interval}) {
	$options{-interval} = 1;
    }
}

########################################################################
sub parse_trial {
    my (
	$headers,
	$fields
	) = @_;

    # Structure of a trial descriptor is:
    #
    # <indep vars 1..m>, <learner vars 1..n>, correct_stim, <other stims 1..o>
    # 
    # The learner vars can be in any order.

    my @headers = @{$headers};
    my @fields  = @{$fields};

    my $correct_stim_index = 0;
    while ($headers [$correct_stim_index] ne "correct_stim") {
	$correct_stim_index++;
	if ($correct_stim_index > @headers) {
	    confess "Didn't find correct_stim column: $!";
	}
    }

    my $trial;
    for (0..$correct_stim_index) {
	$trial->{$headers [$_]} = $fields [$_];
    }

    if ((not defined $fields [$correct_stim_index+1]) or
	!$fields [$correct_stim_index+1] or
	($fields [$correct_stim_index+1] eq 'NA')) {
	# it stays undefined
    } else {
	my @other_candidates;
	for (($correct_stim_index+1)..$#headers) {
	    push @other_candidates, $fields [$_]
	}
	$trial->{other_candidates} = \@other_candidates;
    }
    return $trial;
}

########################################################################
#
#  MAIN
#
########################################################################

# Parse the invocation
&process_options;

# ----------------------------------------------------------------------
# Load the information that lives in the overall experiment directory
# (exppath) and is common to all simulated participants:

# ... the representation scheme
my $repscheme = $options{-repscheme};
require "$repscheme.pm";
$repscheme->import;

my $learner = new Learner;

# ... the stimulus space
my $stimfile = File::Spec->catfile ($options{-exppath}, $options{-stims});
open (STIMS, "< $stimfile") or confess "Couldn't read from $stimfile (-stims): $!";
my @stims = <STIMS>;
my ($stims, undef) = &StimulusSet::from_text (@stims);
$learner->set_stimulus_space ($stims);

# ... the initial constraint population
my ($initcon, undef) =  &ConstraintSet::from_text (
    &Globals::read_text (
	 File::Spec->catfile ($options{-exppath}, $options{-initcon})
    ));
$learner->set_initial_population ($initcon);
#for ($initcon->get_all) {
#    $learner->add_constraint ($_);
#}

# ... the metaconstraints
if ($options{-meta}) {
    my ($meta, undef) =  &ConstraintSet::from_text (
	&Globals::read_text (
	     File::Spec->catfile ($options{-exppath}, $options{-meta})
	));
    $learner->set_meta ($meta);
} else {
    $learner->set_meta (new ConstraintSet);
}

# ... the mutagens
my $mutagens = &MutagenSet::from_text (
    &Globals::read_text (
	 File::Spec->catfile ($options{-exppath}, $options{-mutagens})
    ));
$learner->set_mutagens ($mutagens);    

# ... the recombinators
my $recombinators = &RecombinatorSet::from_text (
    &Globals::read_text (
	 File::Spec->catfile ($options{-exppath}, $options{-recombinators})
    ));
$learner->set_recombinators ($recombinators);    

 
# DEBUG:  Check that the learner can be saved and restored
#print (join "\n", ($learner->to_text));
#print "\n=====\n";
#my $duplicate = (new Learner)->from_text ($learner->to_text);
#print (join "\n", ($duplicate->to_text));
#print "\n=====\n";

# ----------------------------------------------------------------------
# Open files that live in the participant's individual directory 
# (exppath/subjid):

# ... the experiment file (the one with the trial descriptions)
if (defined $options{-exppath} and defined $options{-subjid}) {
    my $expfile_path = File::Spec->catfile ($options{-exppath}, $options{-subjid}, "Experiment.csv");
    open (EXPFILE, "< $expfile_path") or confess "Couldn't open $expfile_path (-expfile): $!";
} else {
    warn "Reading expfile from STDIN";
    *EXPFILE = *STDIN;
}

# ... the data file (trial descriptions plus ``participant'' responses)
my $datfile_path = File::Spec->catfile ($options{-exppath}, $options{-subjid}, "Data.csv");
open (DATFILE, "> $datfile_path") or confess "Couldn't open $datfile_path: $!";

# ... the various performance reports
my $report = new Report ($learner, $options{-subjid}, $options{-exppath}, 
    $options{-log_format});
$report->start;

# Get column headers from the experiment file and get set up to parse 
# subsequent lines
my $exp_headers = <EXPFILE>;
chomp $exp_headers;
my @exp_headers = split /$sepchar/, $exp_headers;
my @exp_headers_up_to_correct_stim;
for my $i (0..$#exp_headers) {
    push @exp_headers_up_to_correct_stim, $exp_headers [$i];
    last if ($exp_headers [$i] eq 'correct_stim');
} 

# Extra column headers that extend the experiment file to the data file
my @dat_headers = qw/ chosen_stim pr_corr corr / ;
print DATFILE ((join $sepchar, (@exp_headers, @dat_headers)) . "\n");

# ----------------------------------------------------------------------
# Read and process each trial description from the experiment file 
# (main loop)
my $err_num = 0;
while (<EXPFILE>) {
    chomp;
    my @fields = split /$sepchar/;

    my $parse = &parse_trial (\@exp_headers, \@fields);
    my ($chosen_stim, $pr_corr, $corr) = 
	$learner->learn (&parse_trial (\@exp_headers, \@fields));
    $report->add_trial ($parse->{correct_stim}, $chosen_stim, $pr_corr, $corr);

    print DATFILE (join $sepchar, (@{$parse}{@exp_headers_up_to_correct_stim}));
    
    print DATFILE $sepchar;
    if (defined ($parse->{other_candidates})) {
	print DATFILE (join $sepchar, (@{$parse->{other_candidates}}, $chosen_stim));
    } else {
	print DATFILE $chosen_stim;
    }

#    print DATFILE (join $sepchar, (@{$parse}{@exp_headers_up_to_correct_stim}, 
#				   @{$parse->{other_candidates}}, $chosen_stim));


    printf DATFILE "%s%6.4f%s%d\n", $sepchar, $pr_corr, $sepchar, $corr;

    # Update reports on every footh error
    if ($err_num % $options{-interval} == 0) {
	$report->update;
    }
    # Update logfile on every error
    if (!$corr) {
        $report->add_to_log ($parse->{correct_stim}, $chosen_stim, $pr_corr);
	$err_num++;
    }
}

close EXPFILE;
close DATFILE;

$report->finish;
