package BinaryFeature;
use Globals;
use Environment;
use BinaryFeatureGenus;
use BinaryFeatureVariable;
use warnings;
use strict;
use Carp;

########################################################################
#
# BinaryFeature.pm
#
# Object implementing a binary feature node.
#
# BinaryFeatures have + or - values and no dependents of any sort.
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2009 March 8 ($)
#
# 2010 January 23 (S) -- added breed method
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


########################################################################
# 
# Constructors
#
########################################################################

# plus, minus -- create [+Foo] or [-Foo] node.

sub plus {
    my (
	$class,
	$BFT
	) = @_;
    new BinaryFeature ($BFT, 1);
}

sub minus {
    my (
	$class,
	$BFT
	) = @_;
    new BinaryFeature ($BFT, -1);
}

# new -- internal use only
sub new {
    my $class = shift;
    my $BFT = shift;
    my $coefficient = shift;
    unless (ref ($BFT) and $BFT->isa ('BinaryFeatureGenus')) {
	carp "Found ", ref (\$BFT), " while looking for a BinaryFeatureGenus: $!";
	confess;
	die;
    }
    unless (abs ($coefficient) == 1) {
	carp "BinaryFeature coefficient must be +1 or -1, not $coefficient:$!";
	confess;
	die;
    }

    my $this = {};
    $this->{genus} = $BFT;
    $this->{genus_name} = $BFT->get_name;
    $this->{coefficient} = $coefficient;
    bless ($this, $class);
    $this;
}

sub clone {
    my $this = shift;
    my $copy = new BinaryFeature (
	$this->get_genus,
	$this->get_coefficient
	);
    $copy;
}

# pseudo_clone -- Faster than real clone, and safe if neither original
#     nor copy is modified.
sub pseudo_clone {
    my $this = shift;
    $this;
}

# inverted_clone -- Same feature, opposite sign.
sub inverted_clone {
    my $this = shift;
    my $copy = new BinaryFeature (
	$this->get_genus,
	-$this->get_coefficient
	);
    $copy;
}

sub inverted_pseudo_clone {
    my $this = shift;
    $this->inverted_clone;
}

sub mutant {
    my $this = shift;
    my $mutagen = (shift or '(no argument)');
    unless (ref ($mutagen) and $mutagen->isa ('Mutagen')) {
	carp "Argument to mutant must be of type Mutagen, not $mutagen:$!";
	confess;
	die;
    }

    return $this->clone unless ($mutagen->mutate);

    # There are only two things you can do to a BinaryFeature, namely,
    # invert the sign, and turn it into a variable.

    if ($mutagen->binary_feature_to_variable) {

	# This BinaryFeature becomes a BinaryFeatureVariable
	if ($mutagen->invert_binary_coefficient) {
	    if ($this->get_coefficient < 0) {
		minus BinaryFeatureVariable (
		    $this->get_genus,
		    'alpha'
		    );
	    } else {
		plus BinaryFeatureVariable (
		    $this->get_genus,
		    'alpha'
		    );
	    }
	} else {
	    if ($this->get_coefficient < 0) {
		plus BinaryFeatureVariable (
		    $this->get_genus,
		    'alpha'
		    );
	    } else {
		minus BinaryFeatureVariable (
		    $this->get_genus,
		    'alpha'
		    );
	    }
	}
    } else {

	# It stays a BinaryFeature
	if ($mutagen->invert_binary_coefficient) {
	    $this->inverted_clone;
	} else {
	    $this->clone;
	}
    }
}

sub breed {
    my (
	$this, $that,       # Parents
        $recombinator
	) = @_;
    
    if ((ref $that) and ($that->isa ("Recombinator"))) {
	$recombinator = $that; $that = undef
    }
    unless (defined $recombinator) {
	confess "No Recombinator in call to breed (): $!";
    }

    my $kid;

  SWITCH: {
      if (!defined ($that)) {
	  $kid = (!$recombinator->null_wins) ? $this->clone : undef;
	  last SWITCH;
      }

      unless ($this->get_genus_name eq $that->get_genus_name) {
	  carp "Genera don't match: " . $this->get_genus_name . " != " . 
	      $that->get_genus_name . ":$!";
	  confess;
	  die;
      }
    
      if ($that->isa ("BinaryFeature")) {
	  $kid = ($recombinator->this_wins) ? $this->clone : $that->clone;
	  last SWITCH;
      }
    	
      if ($that->isa ("BinaryFeatureVariable")) {
	  $kid = ($recombinator->this_wins) ? $this->clone : $that->clone;
	  last SWITCH;
      }

      carp "Can't breed a BinaryFeature with a " . ref ($that) . ": $!";
      confess;
      die;
    }
    
    return $kid;
}

########################################################################
#
# Access methods for permanent attributes
#
########################################################################

sub get_genus {
    my $this = shift;
    $this->{genus};
}

sub get_genus_name {
    my $this = shift;
#    $this->get_genus->get_name;
    $this->{genus_name};
}

sub get_genus_abbrev {
    my $this = shift;
    $this->get_genus->get_abbrev;
}

sub get_coefficient {
    my $this = shift;
    $this->{coefficient};
}

sub get_sign {
    my $this = shift;
    ($this->get_coefficient == 1) ? '+' : '-';
}

sub has_dependents {
    return;
}

sub get_depth {
    0;             # Binary features have no dependents.
}

sub get_subtree_size {
    1;             # Again, binary features are terminal.
}

########################################################################
#
# Matching methods
#
########################################################################

# A BinaryFeature $this matches in another BinaryFeature $that iff 
# $this and $that have the same genus and same sign.

sub matches_in {
    my (
	$this, 
	$that,
	$env
	) = @_;
    defined ($env) or ($env = new Environment);

    unless (ref ($that) and $that->can ('matches_in')) {
	carp "Match target must be a matchable object, not $that: $!";
        confess;
        die;
    }
    unless (ref ($env) and $env->isa ('Environment')) {
        carp "Second argument to matches_in must be Environment, not $that: $!";
        confess;
        die;
    }

    if ($that->isa ('BinaryFeature')) {
	return unless ($this->get_genus_name eq $that->get_genus_name);
	return unless ($this->get_coefficient eq $that->get_coefficient);
	return (my @there_must_be_a_better_way_to_return_a_singleton_array 
		= ($env));
    } else {
	return;
    }
}

sub equals {
    my ($this, $that, $env) = @_;
    $this->matches_in ($that, $env) and $that->matches_in ($this, $env);
}
	
########################################################################
#
# I/O methods
#
########################################################################

sub to_flat_string {
    my (
	$this,
	$is_head
	) = (@_, 0);
    my $left  = ($is_head ? '[*' : '[');
    my $right = ($is_head ? '*]' : ']');
    $left . $this->get_sign . $this->get_genus_abbrev . $right;
}
	

sub to_text {
    my (
	$this,
	$depth,       # Depth of nesting
	$is_head    # Head of parent?
	) = (@_, 0, 0);

    if ($is_head) {
	($Globals::indent x $depth) .
	    '*' . 
	    $this-> get_sign . 
	    $this-> get_genus_name .
	    '*'
    } else {
	($Globals::indent x $depth) .
	    '[' . 
	    $this-> get_sign .
	    $this-> get_genus_name .
	    ']';
    }
}

sub from_text {
    # Pass text in as ref to array, or array itself
    my $text = (ref $_[0] eq 'ARRAY') ? $_[0] : \@_;

    # Return empty-handed unless the text begins by describing a BinFeat
    unless (
	$text->[0] =~ /^((?:$Globals::indent_re)*)
                    ([\[\*])    \s*
                    ([+\-])     \s*
                    ([\w\s]+\w) \s*
                    [\]\*]      \s*$/x) {
	return;
    }
    my ($indent_string, 
	$head_string,
        $sign,
        $genus_name) = ($1, $2, $3, $4);
    
    my $depth = length ($indent_string)/length ($Globals::indent);
    my $is_head = ($head_string eq '*') ? 1 : 0;
    my $coefficient = ($sign eq '+') ? 1 : -1;
    my $genus = &Globals::get_representations->get ($genus_name);

    unless (ref ($genus) and $genus->isa ('BinaryFeatureGenus')) {
	return;
    }
 
    my $new = new BinaryFeature ($genus, $coefficient);
    shift @{$text};
    ($new, $depth, $is_head);
}
   
sub print {
    my (
	$this,
	$depth,     # Depth of nesting
	$is_head
	) = (@_, 0, 0);
    print $this->to_text ($depth, $is_head), "\n";
}

sub to_tree {
    my $this = shift;
    my $width = (shift or '1in');
   
    my $indent_level = (shift or 0);
    my $indent_quantum = 2;
    my $indent = ' ' x ($indent_quantum * (1+$indent_level));

    my $leader_level = (shift or 0);
    my $leader_quantum = 1;
    my $leader = '.' x ($leader_quantum * $leader_level);

    my $style = (shift or '\ttfamily');

    my $string = $indent . "$style" . "{$leader}" . '{[}';
    $string .= (($this->get_coefficient == 1) ? '+' : '-');
    $string .= $this->get_genus_abbrev . "{]}\\\\\n";
    $string;
}

1;
