package BinaryFeatureVariable;
use Environment;
use warnings;
use strict;
use Carp;

our @ISA = qw( Variable );

########################################################################
#
# BinaryFeatureVariable.pm
#
# A Variable, together with a mutant () method that makes it compatible
# with BinaryFeatures.
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2009 March 8 ($)
#
# 2010 January 23 (S) -- added breed method
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


sub plus {
    my (
	$class, 
	$BFT,
	$species
	) = @_;
    unless (ref ($BFT) and $BFT->isa ('BinaryFeatureGenus')) {
	carp "Found ", ref (\$BFT), " while looking for a BinaryFeatureGenus: $!";
	confess;
        die;
    }

    my $this = plus Variable ($BFT, $species);
    bless $this, $class;
    $this;
}

sub minus {
    my (
	$class, 
	$BFT,
	$species
	) = @_;
    unless (ref ($BFT) and $BFT->isa ('BinaryFeatureGenus')) {
	carp "Found ", ref (\$BFT), " while looking for a BinaryFeatureGenus: $!";
	confess;
        die;
    }

    my $this = minus Variable ($BFT, $species);
    bless $this, $class;
    $this;
}

sub clone {
    my $this = shift;
    my $copy;
    if ($this->is_negated) {
        $copy = minus BinaryFeatureVariable (
            $this->get_genus,
            $this->get_species
            );
    } else {
        $copy = plus BinaryFeatureVariable (
            $this->get_genus,
            $this->get_species
            );
    }
    $copy;
}

sub inverted_clone {
    my $this = shift;
    my $copy;
    unless ($this->is_negated) {
        $copy = minus BinaryFeatureVariable (
            $this->get_genus,
            $this->get_species
            );
    } else {
        $copy = plus BinaryFeatureVariable (
            $this->get_genus,
            $this->get_species
            );
    }
    $copy;
}	

sub pseudo_clone {
    my $this = shift;
    $this;
}

sub mutant {
    my $this = shift;
    my $mutagen = (shift or '(no argument)');
    unless (ref ($mutagen) and $mutagen->isa ('Mutagen')) {
        carp "Argument to mutant must be of type Mutagen, not $mutagen:$!";
        confess;
        die;
    }

    return $this->clone unless ($mutagen->mutate);
    if ($mutagen->binary_variable_to_feature) {
	# This BinaryFeatureVariable becomes a BinaryFeature
	if ($mutagen->invert_binary_coefficient) {
	    if ($this->is_negated) {
		plus BinaryFeature ($this->get_genus);
	    } else {
		minus BinaryFeature ($this->get_genus);
	    }
	} else {
	    unless ($this->is_negated) {
	        plus BinaryFeature ($this->get_genus);
	    } else {
		minus BinaryFeature ($this->get_genus);
	    }
	}
    } else {
        # It stays a BinaryFeatureVariable
        if ($mutagen->invert_binary_coefficient) {
            $this->inverted_clone;
        } else {
            $this->clone;
        }
    }
}

sub breed {
    my (
        $this, $that,       # Parents
	$recombinator
        ) = @_;

    if ((ref $that) and ($that->isa ("Recombinator"))) {
        $recombinator = $that; $that = undef
    }
    unless (defined $recombinator) {
        confess "No Recombinator in call to breed (): $!";
    }

    my $kid;

  SWITCH: {
      if (!defined ($that)) {
          $kid = (!$recombinator->null_wins) ? $this->clone : undef;
          last SWITCH;
      }

      unless ($this->get_genus_name eq $that->get_genus_name) {
          carp "Genera don't match: " . $this->get_genus_name . " != " .
              $that->get_genus_name . ":$!";
          confess;
          die;
      }

      if ($that->isa ("BinaryFeature") or
	  $that->isa ("BinaryFeatureVariable")) {
          $kid = ($recombinator->this_wins) ? $this->clone : $that->clone;
          last SWITCH;
      }

      carp "Can't breed a BinaryFeatureVariable with a " . ref ($that) . ": $!";
      confess;
    }

    return $kid;
}

########################################################################
#
# I/O methods
#
########################################################################

sub to_flat_string {
    my (
	$this,
	$is_head
	) = (@_, 0);
    
    if ($is_head) {
	'[*' . $this->get_sign . '<' . $this->get_species . '>' . 
	    $this->get_genus_abbrev . '*]'
    } else {
	'[' . $this->get_sign . '<' . $this->get_species . '>' . 
	    $this->get_genus_abbrev . ']'
    }
}	    

sub to_text {
        my (
	    $this,
	    $depth,    # Depth of nesting
	    $is_head   # Head of parent?
	    ) = (@_, 0, 0);

	if ($is_head) {
	    ($Globals::indent x $depth) .
            '*' .
            $this-> get_sign .
	    '<' . $this->get_species . '> ' .
            $this-> get_genus_name .
            '*'
	} else {
        ($Globals::indent x $depth) .
            '[' .
            $this-> get_sign .
	    '<' . $this->get_species . '> ' .
            $this-> get_genus_name .
            ']';
    }
}

sub from_text {
    # Pass text in as ref to array, or as array itself
    my $text = (ref $_[0] eq 'ARRAY') ? $_[0] : \@_;

    # Return empty-handed unless the text begins by describing a BinFeatVar
    # e.g., [+ <gamma> spread glottis]
    unless (
        $text->[0] =~ /^((?:$Globals::indent_re)*)
                    ([\[\*])      \s*
                    ([+\-])       \s*
                    <([\w\s]+\w)> \s*
                    ([\w\s]+\w)   \s*
                    [\]\*]        \s*$/x) {
        return;
    }
    my ($indent_string,
        $head_string,
        $sign,
        $species,
        $genus_name) = ($1, $2, $3, $4, $5);

    my $depth = length ($indent_string)/length ($Globals::indent);
    my $is_head = ($head_string eq '*') ? 1 : 0;
    my $coefficient = ($sign eq '+') ? 1 : -1;
    my $genus = &Globals::get_representations->get ($genus_name);

    unless (ref ($genus) and $genus->isa ('BinaryFeatureGenus')) {
        return;
    }

    my $new;
    if ($coefficient == 1) {
	$new = $genus->plus ($species);
    } else {
	$new = $genus->minus ($species);
    }
    shift @{$text};
    ($new, $depth, $is_head);
}

sub print {
    my (
        $this,
        $depth,     # Depth of nesting
        $is_head
        ) = (@_, 0, 0);
    print $this->to_text ($depth, $is_head), "\n";
}


sub to_tree {
    my $this = shift;
    my $width =(shift or '1in');

    my $indent_level = (shift or 0);
    my $indent_quantum = 2;
    my $indent = ' ' x ($indent_quantum * (1+$indent_level));

    my $leader_level = (shift or 0);
    my $leader_quantum = 1;
    my $leader = '.' x ($leader_quantum * $leader_level);

    my $style = (shift or '\ttfamily');

    my $string = $indent . $style . "{$leader}" . '{[}';
    $string .= (($this->get_coefficient == 1) ? '+' : '-') . 
	$this->get_species . ' ' .
        $this->get_genus_abbrev . '{]}\\\\' . "\n";;

    $string;
}


1;
