package CELEX;
use GJ2005jr;
use Carp; 
use strict;
use warnings;

########################################################################
#
# CELEX.pm
#
# For converting CELEX representations to PrWds, via the GJ2005jr package.
# 
# Tense vowels are represented as diphthongs ending in -i, -u, or -r.
# Exception:  Open [O] is represented with no offglide.
#
# Pronunciations are somewhat Americanized by restoring coda [r].
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


our @IPA_Cs   = qw/
  ph     th  tS  kh 
  b      d   dZ  g 
  f  T   s   S   x    h
  v  D   z   Z   
  m      n       N
         l
         r
         r
  w          j
/;
our @CELEX_Cs = qw/  
  p      t   J   k
  b      d   _   g
  f   T  s   S   x    h
  v   D  z   Z
  m      n       N
         l
         r
         R
  w          j
/;

our @IPA_Vs  = qw/
  i        u
  e  ^  ^  O   
  ae       a  

  ae a a a
/;

our @CELEX_Vs = qw/
  I        U 
  E  @  V  $  
  {        Q

  c  q  0  ~
/;

our @IPA_syllabic_C = qw/
  m, n, N, 
     l,
/;
our @CELEX_syllabic_C = qw/
  F  H  C
     P
/;

our @IPA_diphthongs = qw/
  ij       uj
        ^r
        ar
  ej aj Oj ow aw ir er Or
/;
our @CELEX_diphthongs = qw/
  i        u
        3
        #
  1  2  4  5  6  7  8  9
/;

our %equiv;
for (1) {
    my @IPA = (@IPA_Cs, @IPA_Vs,   @IPA_syllabic_C,   @IPA_diphthongs);
    my @CELEX = (@CELEX_Cs, @CELEX_Vs, @CELEX_syllabic_C, @CELEX_diphthongs);
    for my $i (0..$#IPA) {
	$equiv {$CELEX [$i]} = $IPA[$i];
    }
}

sub to_representation {
    my @CELEX = split '', (shift);
    my $rep = &GJ2005jr::get ('PrWd')->double_anchored;
    $rep->put (&GJ2005jr::get ('Syll')->double_anchored);    
    for my $celex (@CELEX) {

	if ($celex eq "'") {
	    $rep->set_head;
	    next;
	}
	if ($celex eq '"') {
	    next;          # Ignore secondary stress
	}
	if ($celex eq '-') {
	    $rep->put (&GJ2005jr::get ('Syll')->double_anchored);
	    next;
	}
	if ($celex eq "\n") {
	    next;
	}
	

	# It's a segmental symbol.
	if (grep {/^$celex$/} @CELEX_diphthongs) {
	    my ($V, $G) = (split '', $equiv {$celex});
	    $rep->put_head (&GJ2005jr::to_feature_tree ($V));
	    $rep->put (&GJ2005jr::to_feature_tree ($G));
	    next;
	}
	my $target = quotemeta ($celex);
	if (grep {/^$target$/} (@CELEX_Vs, @CELEX_syllabic_C)) {
	    $rep->put_head (&GJ2005jr::to_feature_tree ($equiv {$celex}));
	    next;
	}
	
	unless ($equiv {$celex}) {
	    carp "Unknown CELEX symbol [$celex]: $!";
	    confess;
	    die;
	}

	$rep->put (&GJ2005jr::to_feature_tree ($equiv {$celex}));
    }
    $rep;
}


 
1;

