package Constraint;
use MCat;
use warnings;
use strict; 
use Carp;
########################################################################
#
# Constraint.pm
#
# Object implementing a constraint.
#
# Essentially just a wrapper for a ``target'' (a PrCat or feature tree)
# which adds a ``polarity'', so that we can specify whether matches of
# the target count as +1 or as -1.
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2010 February 5 (F)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


our $max_unique_id;

########################################################################
#
# Constructors
#
########################################################################

sub new {
    my $class = shift;
    my $target = shift;
    my $polarity = 0+shift;
    my $name = shift;

    unless (ref ($target) and $target->can ('matches_in')) {
	carp "Target must be of a class that can match, not " .
	    ref (\$target) . ": $!";
	confess;
	die;
    }
    unless (ref ($target) and $target->can ('matches_in')) {
	carp "Target must be of a class that can match, not " .
	    ref (\$target) . ": $!";
	confess;
	die;
    }
    unless ($polarity) {
	confess "Invalid polarity $polarity: $!";
    }

    my $this = {};
    $this->{target} = $target;
    $this->{polarity} = $polarity;
    $this->{name} = ($name or "Constraint");
    $this->{signature_class_name} = "";
    $this->{unique_id} = $max_unique_id++;

    $this->{right_total}   = 0;
    $this->{neither_total} = 0;
    $this->{wrong_total}   = 0;
    $this->{fitness}       = 1;

    # Memoize scoring
    $this->{constraint_score} = {};


    # Memoize conversion to a string

    bless ($this, $class);
    $this->{string} = to_string ($this);
    $this;
}

sub plus {
    my $class = shift;
    my $target = shift;
    new Constraint ($target, 1);
}

sub minus {
    my $class = shift;
    my $target = shift;
    new Constraint ($target, -1);
}

sub clone {
    my $this = shift;
    my $constraint = new Constraint (
	$this->get_target->clone,
	$this->get_polarity,
	$this->get_name
	);
    $constraint->set_fitness ($this->get_fitness);
    $constraint->set_mean_harmony_gap ($this->get_mean_harmony_gap);
    $constraint->set_mutation_propensity ($this->get_mutation_propensity);
    $constraint;
}

sub pseudo_clone {
    my $this = shift;
    my $constraint = Constraint (
	$this->get_target->pseudo_clone,
	$this->get_polarity,
	$this->get_name
	);
    $constraint->set_fitness ($this->get_fitness);
    $constraint->set_mean_harmony_gap ($this->get_mean_harmony_gap);
    $constraint->set_mutation_propensity ($this->get_mutation_propensity);
    $constraint;
}

sub mutant {
    my $this = shift;
    my $mutagen = (shift or '(no argument)');
    my $noclone = shift;

    unless (ref $mutagen and $mutagen->isa ('Mutagen')) {
	confess "First argument to mutant must be Mutagen, not " . (ref $mutagen);
    }

    if ($noclone) {
	return $this unless ($mutagen->mutate);
    } else {
	return $this->clone unless ($mutagen->mutate);
    }

    if ($mutagen->invert_polarity) {
	new Constraint (
	    $this->get_target->mutant ($mutagen),
	    -$this->get_polarity
	    );
    } else {
	new Constraint (
	    $this->get_target->mutant ($mutagen),
	    $this->get_polarity
	    );
    }
}

sub breed {
    my (
	$this, $that,   # Parents
	$recombinator
	) = @_;

    if ((ref $that) and ($that->isa ("Recombinator"))) {
        $recombinator = $that; $that = undef
    }
    unless (defined $recombinator) {
        confess "No Recombinator in call to breed (): $!";
    }

    new Constraint (
	$this->get_target->breed ($that->get_target, $recombinator),
	($recombinator->this_wins) ? $this->get_polarity : $that->get_polarity
	);
}

########################################################################
#
# Access methods for intrinsic constraint properties
#
########################################################################

sub get_target {
    my $this = shift;
    $this->{target};
}

sub get_polarity {
    my $this = shift;
    $this->{polarity};
}

sub set_polarity {
    my $this = shift;
    my $polarity = shift;

    $this->{polarity} = $polarity;
    $this->reset_string;
}

sub invert_polarity {
    my $this= shift;
    $this->set_polarity (-$this->get_polarity);
    $this->reset_string;
}

sub is_positive {
    my $this = shift;
    return ($this->get_polarity > 0)
}

sub is_negative {
    my $this = shift;
    return ($this->get_polarity < 0)
}

sub set_name {
    my $this = shift;
    my $new_name = shift;
    $this->{name} = $new_name;
    $this->reset_string;
    return ($this->{name});
}

sub get_name {
    my $this = shift;
    return ($this->{name});
}

sub get_unique_id {
    my $this = shift;
    return ($this->{unique_id});
}


########################################################################
#
# Access methods for extrinsic constraint properties (the ones used by
# Learner.pm).
#
########################################################################

sub set_outcome {
    my (
	$this,
	$pair,
	$outcome
	) = @_;
    
    unless (ref ($pair) and $pair->isa ('StimulusPair')) {
	carp "$pair is not a StimulusPair: $!";
    }

    if (exists $this->{outcome}->{$pair}) {
	$this->delete_outcome ($pair);
    }

    $this->{outcome}->{$pair} = $outcome;

    $this->{right_total}++   if ($outcome >  0);
    $this->{neither_total}++ if ($outcome == 0);
    $this->{wrong_total}++   if ($outcome <  0);

    $this->get_grand_total;
}

sub get_outcome {
    my (
        $this,
        $pair
        ) = @_;

    $this->{outcome}->{$pair};
}

sub delete_outcome {
    my (
        $this,
        $pair
        ) = @_;

    unless (exists ($this->{outcome}->{$pair})) {
        carp "Unknown constraint $pair: $!";
        confess;
        die;
    }

    my $outcome = $this->get_outcome ($pair);
    $this->{right_total}--   if ($outcome > 0);
    $this->{neither_total}-- if ($outcome == 0);
    $this->{wrong_total}--   if ($outcome <  0);

    delete ($this->{outcome}->{$pair});

    $this->get_grand_total;

}

sub get_right_total {
    my $this = shift;
    $this->{right_total};
}
sub get_neither_total {
    my $this = shift;
    $this->{neither_total};
}
sub get_wrong_total {
    my $this = shift;
    $this->{wrong_total};
}
sub get_grand_total {
    my $this = shift;
    $this->{right_total} + $this->{neither_total} + $this->{wrong_total};
}

sub set_fitness {
    my ($this, $new_fitness) = @_;

    $this->{fitness} = $new_fitness;
}

sub add_fitness {
    my ($this, $fitness_increment) = @_;

    $this->{fitness} += $fitness_increment;
}

sub get_fitness {
    my ($this) = @_;
    if (exists $this->{fitness}) {
	$this->{fitness};
    } else {
	0
    }
}

sub set_mean_harmony_gap {
    my ($this, $new_mean_harmony_gap) = @_;
    $this->{mean_harmony_gap} = $new_mean_harmony_gap;
}

sub get_mean_harmony_gap {
    my $this = shift;

    if (exists $this->{mean_harmony_gap}) {
        $this->{mean_harmony_gap}
    } else {
        0
    }
}

sub set_mutation_propensity {
    my (
	$this,
	$new_mutation_propensity
	) = @_;
    $this->{mutation_propensity} = $new_mutation_propensity;
}

sub get_mutation_propensity {
    my $this = shift;

    if (exists $this->{mutation_propensity}) {
        $this->{mutation_propensity}
    } else {
        1
    }
}


########################################################################
#
# Matching methods
#
########################################################################

sub matches_in {
    my (
	$this,  # Thing to match
	$that,  # Thing to match it in
	$env    # Envionment with variable assignments
	) = @_;

    if ($that->isa ('Constraint')) {
	$this->get_target->matches_in ($that->get_target, $env);
    } else {
	$this->get_target->matches_in ($that, $env);
    }
}

sub score {
    my (
	$this,  # Thing to match
	$that  # Thing to match it in
	) = @_;

    my $memo = $this->{constraint_score}->{$this}->{$that};
    if (defined ($memo)) {
	$memo;
    } else {
	my @matches = $this->matches_in ($that);
	$this->{constraint_score}->{$this}->{$that} = (0+@matches) * $this->get_polarity;
    }
}

sub scores {
    my (
	$this,   # Thing to match
	@thats   # Things to match them in
	) = @_;
    map {
	$this->score ($_)
    } @thats;
}


########################################################################
#
# I/O methods
#
########################################################################

sub to_flat_string {
    my $this = shift;
    
    '[' . 
	$this->get_polarity  . " " .
	($this->get_name or 'Constraint') .
	$this->get_target->to_flat_string .
	']';
}

sub to_string {
    my $this = shift;
    
    # Memoize, since the string form is used as a hash key.
    unless ($this->{string}) {
	# Omit constraint name, so that all constraints with same
	# content have same string form
	$this->{string} = (join "\n", $this->to_text (0, 1))
    }
    $this->{string}
}

sub to_target_string {
    my $this = shift;
    join ("\n", $this->to_text (1,1));
}

sub reset_string {
    my $this = shift;
    
    delete ($this->{string});
    $this->to_string;
}
    

sub to_text {
    my (
	$this,
	$depth,     # Depth of nesting
	$noname     # Omit constraint name
	) = (@_, 0, 0);
    
    my @text;
    my $name = $noname ? 'Constraint' : ($this->get_name or 'Constraint');
    if ($this->get_polarity == +"Inf") {
	push @text, 
	    ($Globals::indent x $depth) . '+' . $this->get_polarity . " " .
	    $name
    } else {
	push @text, ($Globals::indent x $depth) . $this->get_polarity . " " .
	    $name
    }
    push @text, $this->get_target->to_text ($depth + 1);
    
    @text;
}

sub from_text {
    # Pass text in as ref to array, or array itself
    my $text = (ref $_[0] eq 'ARRAY') ? $_[0] : \@_;

    # Return empty-handed unless the text begins with a constraint identifier
    unless (
        ($text->[0] =~ /^((?:$Globals::indent_re)*)
                    ([\+\-]?[Ii][Nn][Ff]) \s+
                    (.+\S)       \s*$/x)
	or
        ($text->[0] =~ /^((?:$Globals::indent_re)*)
                    ([\+\-]?\d*\.?\d*) \s*
                    (.+\S)       \s*$/x)
	) {
        return;
    }

    my ($indent_string, 
        $polarity_string,
        $name) = ($1, $2, $3);

    my $depth = length ($indent_string)/length ($Globals::indent);
    my $polarity;

#    print "polarity_string is $polarity_string and name is $name, so ... ";

    if ($polarity_string eq '+') {
	$polarity = +1;
    } elsif ($polarity_string eq '-') {
	$polarity = -1;
    } elsif ($polarity_string =~ /^\+[Ii][Nn][Ff]$/) {
	$polarity = +"Inf";
    } elsif ($polarity_string =~ /^\-[Ii][Nn][Ff]$/) {
	$polarity = -"Inf";
    } else {
	$polarity = 0 + $polarity_string;
    }

#    print "polarity is $polarity\n";

    unless ($polarity) {
	confess "Unrecognized polarity $polarity_string: $!";
    }

    shift @{$text};

    my $next_line = $text->[0];
    $next_line =~ /^((?:$Globals::indent_re)*)/;
    my $next_depth = length ($1)/length ($Globals::indent);
    if ($next_depth <= $depth) {
	carp "Expected target for constraint $polarity_string$name, found $next_line: $!";
	confess;
    }

    my $base = &Globals::get_representations->get_base;
    my $target;
    if ($base->isa ('MCatGenus')) {
	($target) = &MCat::from_text ($text);
    } elsif ($base->isa ('PrCatGenus')) {
	($target) = &PrCat::from_text ($text);
    }
    unless ($target) {
	carp "Couldn't find target for constraint $polarity_string$name in $next_line: $!";
	confess;
    }

    my $new;
    unless ($new = new Constraint ($target, $polarity, $name)) {
	carp "Error in $next_line: $!";
	carp "Couldn't add target for constraint $polarity_string$name.";
	confess;
    }

    $new;
}


sub print {
    my $this = shift;
    print join ("\n", $this->to_text), "\n";
}

sub to_tree {
    my $this = shift;
    my $width = (shift or '1in');

    my $string = '[.{';
    $string .=  (sprintf "\\textbf{%+2d}", $this->get_polarity);
    $string .= '}' . "\n";
    $string .= $this->get_target->to_tree ($width);
    $string .= ' ]' . "\n";

    $string;
}
    

1;
