package ConstraintSet;
use Constraint;
use warnings;
use strict;
use Carp;
use List::Util;

########################################################################
#
# ConstraintSet.pm
#
# Just that --- a set of constraints.
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


########################################################################
#
# Constructors
#
########################################################################

sub new {
    my $class = shift;

    my $this = {};
    bless $this, $class;

    $this->put (@_);

    $this;
}

########################################################################
#
# Access methods
#
########################################################################

sub put {
    my $this = shift;
    grep {
	unless (ref ($_) and $_->isa ('Constraint')) {
	    carp "ConstraintSet can't contain $_: $!";
	    confess;
	    die;
	}

	push @{$this->{contents}}, $_;
    } @_;
}

sub delete {
    my $this = shift;
    my @purged = ();
    for my $current_con ($this->get_all) {
	next if (grep {$current_con eq $_} (@_));
	push @purged, $current_con;
    }
    $this->{contents} = \@purged;
}

sub delete_random {
    my $this = shift;
    my $how_many = (shift or 1);

    my @shuffled = &List::Util::shuffle (@{$this->{contents}});
    my @deleted = ();
    if ($how_many >= 1) {
	for (1..$how_many) {
	    push @deleted, (pop @shuffled);
	}
	$this->{contents} = \@shuffled;
    }

    \@deleted;
}

sub delete_least_fit {
    my $this = shift;
    my $how_many = (shift or 1);

    my @shuffled = &List::Util::shuffle (@{$this->{contents}});
    my @sorted_by_fitness = sort {
        $b->get_fitness <=> $a->get_fitness
    } @shuffled;
    my @deleted = ();
    if ($how_many >= 1) {
	for (1..$how_many) {
	    push @deleted, (pop @sorted_by_fitness);
	}
	$this->{contents} = \@sorted_by_fitness;
    }

    \@deleted;
}


sub duplicate_random {
    my $this = shift;
    my $needed = (shift or 1);

    my @duplicates;
    while (@duplicates < $needed) {
	my @shuffled = &List::Util::shuffle (@{$this->{contents}});
	while (@duplicates < $needed) {
	    last unless (@shuffled);
	    my $con = pop @shuffled;
	    push @duplicates, $con->clone;
	}	    
    }
    $this->put (@duplicates);
    \@duplicates;
}


sub get_one {
    my $this = shift;

    ${$this->{contents}}[rand ($this->how_many)];
}


sub get_all {
    my $this = shift;
    if (exists ($this->{contents})) {
	@{$this->{contents}};
    } else {
	undef;
    }
}

sub how_many {
    my $this = shift;
    exists ($this->{contents}) ? 
	0+@{$this->{contents}} : 0;
}

sub all_positive {
    my $this = shift;

    !grep {
	$_->get_polarity < 0
    } ($this->get_all)
}

sub all_negative {
    my $this = shift;

    !grep {
	$_->get_polarity > 0
    } ($this->get_all)
}
	
	    

########################################################################
#
# I/O methods
#
########################################################################

sub to_flat_string {
    my $this = shift;
    
    my $flat_string = "[ConstraintSet";
    my @all = $this->get_all;
    for (@all) {
        $flat_string .= $_->to_flat_string;
    }
    $flat_string .= "]";
    $flat_string;
}

sub to_text {
    my (
        $this,
        $depth      # Depth of nesting
        ) = (@_, 0);

    my @text;
    push @text, (
        ($Globals::indent x $depth) .
        "ConstraintSet"
    );

    my @all = $this->get_all;
    for (@all) {
	next unless $_;
        push @text, $_->to_text ($depth + 1);
    }

    @text;
}

sub from_text {
    # Pass text in as ref to array, or array itself
    my $text = (ref $_[0] eq 'ARRAY') ? $_[0] : \@_;

    # Return empty-handed unless the text begins by describing a ConstSet
    unless (
        $text->[0] =~ /^((?:$Globals::indent_re)*)
                    ConstraintSet\s*$/x) {
        return;
    }
    my $depth = length ($1)/length ($Globals::indent);

    my $new = new ConstraintSet;
    shift @{$text};

    while (@{$text}) {
        # Have we run out of dependents?
        my $next_line = $text->[0];

        $next_line =~ /^((?:$Globals::indent_re)*)/;
        my $next_depth = length ($1)/length ($Globals::indent);
        last if ($next_depth <= $depth);

        my ($child, $child_depth, $child_is_head) =
            &Constraint::from_text ($text);
        unless ($child) {
            carp "Couldn't parse $next_line: $!";
            confess;
        }
        
        unless ($new->put ($child)) {
            carp "Error in $next_line: $!";
            carp "Can't add $child of genus " . $child->get_genus_name;
            confess;
        }
    }

    ($new, $depth);
}

sub print {
    my $this = shift;
    print (join "\n", $this->to_text);
    print "\n";
}

1;
    
	   


