package Environment;
use EnvironmentVariable;
use warnings;
use strict;
use Carp;

########################################################################
#
# Environment.pm
#
# A memory device for backreferences in matching.  E.g., in a constraint
# that contains the nodes [+alpha Place] [-alpha Place], when matches_in
# is called on [-alpha Place], it checks the Environment to find out
# what was matched by [+alpha Place].  The job of remembering the 
# positive and negative cases of [alpha Place] is done by an 
# EnvironmentVariable; the job of remembering all the EnvironmentVariables
# is done by the Environment.
#
# All of the matches_in routines return one (possibly empty) Environment
# for every successful match.
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2009 March 8 ($)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


sub new {
    my $class = shift;
    my $this = {};
    bless $this, $class;
    %{$this->{contents}} = ();
    $this;
}

sub clone {
    my $this = shift;
    my $copy = new Environment;
    for ($this->get_all_environment_variables) {
	$copy->add_environment_variable ($_->clone);
    }
    $copy;
}

########################################################################
#
# Access methods
#
########################################################################

sub add_environment_variable {
    my (
	$this,
	$new_environment_variable
	) = @_;
    unless (ref ($new_environment_variable) and
	    $new_environment_variable->isa ('EnvironmentVariable')) {
	carp "Found ", ref (\$new_environment_variable), " while looking for EnvironmentVariable: $!";
	confess;
	die;
    }
    if ($this->get_environment_variable (
	    $new_environment_variable->get_variable
	)) {
	carp "Tried to add pre-existing EnvironmentVariable of genus ", 
	$new_environment_variable->get_variable->get_genus_name, ": $!";
	confess;
	die;
    }

    $this->{contents}->{$new_environment_variable->get_name} = 
	$new_environment_variable;
}

sub get_environment_variable {
    my (
	$this,
	$variable
	) = @_;
    unless (ref($variable) and
            $variable->isa ('Variable')) {
	carp "Found ", ref (\$variable), " while looking for Variable: $!";
	confess;
	die;
    }

    $this->{contents}->{$variable->get_name};
}

sub get_all_environment_variables {
    my $this = shift;
    values (%{$this->{contents}});
}

sub get_all_positive_cases_of {
    my (
	$this,
	$variable
	) = @_;
    my $environment_variable = $this->get_environment_variable ($variable);
    if ($environment_variable) {
	$environment_variable->get_all_positive_cases;
    } else {
	();
    }
}

sub get_all_negative_cases_of {
    my (
	$this,
	$variable
	) = @_;
    my $environment_variable = $this->get_environment_variable ($variable);
    if ($environment_variable) {
	$environment_variable->get_all_negative_cases;
    } else {
	();
    }
}

sub add_positive_case_of {
    my (
	$this,
	$variable,
	$case
	) = @_;
    my $environment_variable = $this->get_environment_variable ($variable);
    if ($environment_variable) {
	$environment_variable->add_positive_case ($case);
    } else {
	$environment_variable = new EnvironmentVariable ($variable);
	$environment_variable->add_positive_case ($case);
	$this->add_environment_variable ($environment_variable);
    }
}

sub add_negative_case_of {
    my (
	$this,
	$variable,
	$case
	) = @_;
    my $environment_variable = $this->get_environment_variable ($variable);
    if ($environment_variable) {
	$environment_variable->add_negative_case ($case);
    } else {
	$environment_variable = new EnvironmentVariable ($variable);
	$environment_variable->add_negative_case ($case);
	$this->add_environment_variable ($environment_variable);
    }
}

########################################################################
#
# I/O methods
#
########################################################################

sub print {
    my $this = shift;
    for ($this->get_all_environment_variables) {
	if ($_->get_all_positive_cases) {
	    print "+ ", $_->get_name, ":\n";
	    for ($_->get_all_positive_cases) {
		$_->print;
	    }
	}
	if ($_->get_all_negative_cases) {
	    print "- ", $_->get_name, ":\n";
	    for ($_->get_all_negative_cases) {
		$_->print;
	    }
	}
    }
}

1;
