package FeatureGenus;
use UnaryFeature;
use strict;
use warnings;
use Carp;

########################################################################
#
# FeatureGenus.pm
#
# Used for defining a feature system and feature geometry.  Specifies
# a given feature's name, range (+? -? other?), immediate dependents,
# and parent.
#
# Features differ from PrCats in having names, ranges, and *named*
# (rather than ordered) dependents.
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2009 March 8 ($)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


########################################################################
#
# Constructors
#
########################################################################

sub new {
    my $class = shift;
    my (
	$name,    # Name of feature
	$abbrev,  # Short name of feature
	@range    # List of permissible values
	) = @_;

    my $this = {};
    $this->{name} = $name;
    $this->{abbrev} = $abbrev;
    @{$this->{range}} = @range;
    $this->{parent} = ();
    %{$this->{dependents}} = ();
    @{$this->{dependent_features}} = ();

    # Memoize lookup
    $this->{is_dependent_of_MEMO} = {};

    bless ($this, $class);
}

########################################################################
#
# Access methods
#
########################################################################

sub get_name {
    my $this = shift;
    $this->{name};
}

sub get_abbrev {
    my $this = shift;
    $this->{abbrev};
}

sub get_range {
    my $this = shift;
    @{$this->{range}};
}

sub get_arity {
    my $this = shift;
    0 + $this->get_range;
}

# range_includes -- Is the given $x one of this feature's possible values?
sub range_includes {
    my ($this, $x) = @_;
    grep {$_ eq $x} $this->get_range;
}

# add_dependent_features -- Specify other FeatureGenuses as dependents of
#     this one.
sub add_dependent_features {
    my $this = shift;
    for (@_) {
	unless ($_->isa ('FeatureGenus')) {
	    die "Dependent feature must be a FeatureGenus: $!";
	}
	$_->set_parent ($this);

	# Dependents are stored in a hash, keyed by their names:
	$this->{dependents}->{$_->get_name} = $_;
	
	# Keep track of our dependent features, for easy access:
	@{$this->{dependent_features}} = values %{$this->{dependents}};
    }
}

sub get_dependent_features {
    my $this = shift;
    @{$this->{dependent_features}};
}

sub get_dependent_features_ref {
    my $this = shift;
    $this->{dependent_features};
}

# get_dependent_feature_names -- Get list of names of all immediate dependents
sub get_dependent_feature_names {
    my $this = shift;
    keys %{$this->{dependents}};
}

sub get_dependent_feature_named {
    my ($this, $name) = @_;
    $this->{dependents}->{$name}
}

sub is_dependent_of {
    my ($this, $that) = @_;

    my $is = $this->{is_dependent_of_MEMO}->{$this}->{$that};
    defined ($is) ? $is : ($this->get_parent eq $that);
}

# set_parent -- Create upward link from dependent feature.
sub set_parent {
    my $this = shift;
    my $new_parent = shift;
    unless (ref ($new_parent) and (
		$new_parent->isa ('FeatureGenus') or
		$new_parent->isa ('PrCatGenus')
	    )) {
	carp "Parent feature ($new_parent) must be a FeatureGenus or PrCatGenus: $!";
	confess;
	die;
    }
    $this->{parent} = $new_parent;
}

sub get_parent {
    my $this = shift;
    $this->{parent};
}

########################################################################
#
# I/O methods
#
########################################################################

sub print {
    my $this = shift;
    my $indent = (shift or "");

    print $indent;
    printf "%s (%s) {", $this->get_name, $this->get_abbrev;
    for ($this->get_range) {
	printf "%s", $_;
    }
    print "}";
    if ($this->get_parent) {
	print ", dependent of ", $this->get_parent->get_name;
    }
    print "\n";
    for ($this->get_dependent_features) {
	$_->print ("$indent  ");
    }
}

1;
