package GJ2005jrFM;
use Representations;
use UnaryFeatureGenus;
use BinaryFeatureGenus;
use PrCatGenus;
use MCatGenus;
use Carp;
use strict;
use warnings;
use Exporter;
our @ISA = 'Exporter';
our @EXPORT = qw($IPA, $text);

########################################################################
#
# IPA.pm
#
# IPA-to-tree conversion.
#
# This package isn't part of the Subtree Constraint Schema; none of those
# packages depend on it, and scripts using the SCS packages are free to
# substitute something else for this. 
#
# This particular feature system is based on that in the Gussenhoven and
# Jacobs 2005 textbook.  I've commented out constricted glottis to give
# Laryngeal the same number of dependents as Dorsal (for evening the
# height/height-height/voice-voice/voice playing field), and Radical 
# as irrelevant in English except perhaps for [r].  I didn't implement
# a tense/lax or long/short vowel distinction in the feature system;
# you'd have to do that by putting one vowel segment in twice.  I've
# simplified the prosody to PrWd--Syllable--Segment (= [Root]).
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2009 March 8 ($)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


our $IPA;    # ASCII IPA symbols => feature, value, feature, value ...
our $text;   # Feature tree (as text) => ASCII IPA symbol

my $RepSys = new Representations;
my @text = &Globals::read_text ($Globals::home_dir . "/GJ2005jrFM.rep");
&Globals::set_representations ($RepSys);
$RepSys->from_text (@text);

#_______________________________________________________________________
my @Root = ($RepSys->get ('Root'), '+');

my @voiced = (
    $RepSys->get ('Lar'), '+', 
    $RepSys->get ('voi'), '+', 
    $RepSys->get ('spgl'), '-'
    );
my @voiceless = (
    $RepSys->get ('Lar'), '+', 
    $RepSys->get ('voi'), '-', 
    $RepSys->get ('spgl'), '-'
    );
my @voiceless_aspirated = (
    $RepSys->get ('Lar'), '+', 
    $RepSys->get ('voi'), '-', 
    $RepSys->get ('spgl'), '+'
    );

#_______________________________________________________________________
my @bilabial = (
    $RepSys->get ('Place'), '+', $RepSys->get ('Lab'), '+'
    );
my @labiodental = (
    $RepSys->get ('Place'), '+', $RepSys->get ('Lab'), '+'
    );
my @interdental = (
    $RepSys->get ('Place'), '+', $RepSys->get ('Cor'), '+', 
    $RepSys->get ('ant'), '+', $RepSys->get ('dist'), '+'
    );
my @alveolar = (
    $RepSys->get ('Place'), '+', $RepSys->get ('Cor'), '+', 
    $RepSys->get ('ant'), '+', $RepSys->get ('dist'), '-'
    );
my @palatoalveolar = (
    $RepSys->get ('Place'), '+', $RepSys->get ('Cor'), '+', 
    $RepSys->get ('ant'), '-', $RepSys->get ('dist'), '+', 
    );
my @velar = (
    $RepSys->get ('Place'), '+', $RepSys->get ('Dor'), '+', 
    $RepSys->get ('high'), '+', $RepSys->get ('low'), '-', $RepSys->get ('back'), '-'
    );

#_______________________________________________________________________
my @oral_stop = (
    $RepSys->get ('syll'), '-',
    $RepSys->get ('cons'), '+', 
    $RepSys->get ('son'), '-', 
    $RepSys->get ('approx'), '-',
    $RepSys->get ('cont'), '-', 
    $RepSys->get ('nas'), '-', 
    $RepSys->get ('lat'), '-'
    );
my @affricate = (
    $RepSys->get ('syll'), '-',
    $RepSys->get ('cons'), '+', 
    $RepSys->get ('son'), '-', 
    $RepSys->get ('approx'), '-',
    $RepSys->get ('cont'), '-', 
    $RepSys->get ('nas'), '-', 
    $RepSys->get ('lat'), '-'
    );
my @fricative = (
    $RepSys->get ('syll'), '-',
    $RepSys->get ('cons'), '+', 
    $RepSys->get ('son'), '-', 
    $RepSys->get ('approx'), '-',
    $RepSys->get ('cont'), '+', 
    $RepSys->get ('nas'), '-', 
    $RepSys->get ('lat'), '-'
    );
my @nas_stop = (
    $RepSys->get ('syll'), '-',
    $RepSys->get ('cons'), '+', 
    $RepSys->get ('son'), '+', $RepSys->get ('approx'), '-',
    $RepSys->get ('voi'), '+', $RepSys->get ('spgl'), '-', 
    $RepSys->get ('cont'), '-', 
    $RepSys->get ('nas'), '+', $RepSys->get ('lat'), '-'
    );
my @syllabic_nasal = (
    $RepSys->get ('syll'), '+',
    $RepSys->get ('cons'), '+', 
    $RepSys->get ('son'), '+', $RepSys->get ('approx'), '-',
    $RepSys->get ('voi'), '+', $RepSys->get ('spgl'), '-', 
    $RepSys->get ('cont'), '-', 
    $RepSys->get ('nas'), '+', $RepSys->get ('lat'), '-'
    );
my @liquid = (
    $RepSys->get ('syll'), '-',
    $RepSys->get ('cons'), '+', 
    $RepSys->get ('son'), '+', $RepSys->get ('approx'), '+',
    $RepSys->get ('voi'), '+', $RepSys->get ('spgl'), '-', 
    $RepSys->get ('cont'), '+', $RepSys->get ('nas'), '-'
    );
my @syllabic_liquid = (
    $RepSys->get ('syll'), '+',
    $RepSys->get ('cons'), '+', 
    $RepSys->get ('son'), '+', $RepSys->get ('approx'), '+',
    $RepSys->get ('voi'), '+', $RepSys->get ('spgl'), '-', 
    $RepSys->get ('cont'), '+', $RepSys->get ('nas'), '-'
    );

#_______________________________________________________________________
my @vowel = (
    $RepSys->get ('syll'), '+',
    $RepSys->get ('cons'), '-', 
    $RepSys->get ('son'), '+', 
    $RepSys->get ('approx'), '+', 
    $RepSys->get ('cont'), '+', 
    $RepSys->get ('nas'), '-', 
    $RepSys->get ('lat'), '-',
    @voiced,
    $RepSys->get ('Place'), '+', 
    $RepSys->get ('Dor'), '+'
    );

my @glide = (
    $RepSys->get ('syll'), '-',
    $RepSys->get ('cons'), '-', 
    $RepSys->get ('son'), '+', 
    $RepSys->get ('approx'), '+', 
    $RepSys->get ('cont'), '+', 
    $RepSys->get ('nas'), '-', 
    $RepSys->get ('lat'), '-',
    @voiced,
    $RepSys->get ('Place'), '+', 
    $RepSys->get ('Dor'), '+'
    );

#_______________________________________________________________________

$IPA->{ph} = [(@Root, @voiceless_aspirated, @bilabial, @oral_stop)];
$IPA->{th} = [(@Root, @voiceless_aspirated, @alveolar, @oral_stop)];
$IPA->{kh} = [(@Root, @voiceless_aspirated, @velar,    @oral_stop)];

$IPA->{p} = [(@Root, @voiceless, @bilabial, @oral_stop)];
$IPA->{t} = [(@Root, @voiceless, @alveolar, @oral_stop)];
$IPA->{k} = [(@Root, @voiceless, @velar,    @oral_stop)];

$IPA->{tS} = [(@Root, @voiceless, @palatoalveolar, @affricate, 
	       $RepSys->get ('strid'), '+'
	      )];

$IPA->{b} = [(@Root, @voiced, @bilabial, @oral_stop)];
$IPA->{d} = [(@Root, @voiced, @alveolar, @oral_stop)];
$IPA->{g} = [(@Root, @voiced, @velar,    @oral_stop)];

$IPA->{dZ} = [(@Root, @voiced, @palatoalveolar, @affricate, 
	       $RepSys->get ('strid'), '+'
	      )];

$IPA->{f} = [(@Root, @voiceless, @labiodental,    @fricative)];
$IPA->{T} = [(@Root, @voiceless, @interdental,    @fricative, 
	      $RepSys->get ('strid'), '-'
	     )];
$IPA->{s} = [(@Root, @voiceless, @alveolar,       @fricative, 
	      $RepSys->get ('strid'), '+'
	     )];
$IPA->{S} = [(@Root, @voiceless, @palatoalveolar, @fricative, 
	      $RepSys->get ('strid'), '+'
	     )];
$IPA->{x} = [(@Root, @voiceless, @velar,          @fricative, 
	      $RepSys->get ('strid'), '-'
	     )];
$IPA->{G} = [(@Root, @voiced,    @velar,          @fricative, 
	      $RepSys->get ('strid'), '-'
	     )];
$IPA->{h} = [(@Root, @voiceless_aspirated, @fricative)];

$IPA->{v} = [(@Root, @voiced, @labiodental,    @fricative)];
$IPA->{D} = [(@Root, @voiced, @interdental,    @fricative, 
	      $RepSys->get ('strid'), '-'
	     )];
$IPA->{z} = [(@Root, @voiced, @alveolar,       @fricative, 
	      $RepSys->get ('strid'), '+'
	     )];
$IPA->{Z} = [(@Root, @voiced, @palatoalveolar, @fricative, 
	      $RepSys->get ('strid'), '+'
	     )];

$IPA->{m} = [(@Root, @voiced, @bilabial, @nas_stop)];
$IPA->{n} = [(@Root, @voiced, @alveolar, @nas_stop)];
$IPA->{N} = [(@Root, @voiced, @velar,    @nas_stop)];

$IPA->{"m,"} = [(@Root, @voiced, @bilabial, @syllabic_nasal)];
$IPA->{"n,"} = [(@Root, @voiced, @alveolar, @syllabic_nasal)];
$IPA->{"N,"} = [(@Root, @voiced, @velar,    @syllabic_nasal)];

$IPA->{l} = [(@Root, @voiced, @alveolar, $RepSys->get ('lat'), '+', @liquid)];
$IPA->{r} = [(@Root, @voiced, @alveolar, 
	      $RepSys->get ('lat'), '-', @liquid)];
$IPA->{"l,"} = [(@Root, @voiced, @alveolar, 
		 $RepSys->get ('lat'), '+', @syllabic_liquid)];
$IPA->{"r,"} = [(@Root, @voiced, @alveolar, 
		 $RepSys->get ('lat'), '-', @syllabic_liquid)];
# The following is probably more accurate:  Treat /r/ as a glide/vowel
# pair:
#$IPA->{'r,'} = [(@Root, @vowel, $RepSys->get ('high'), '-',
#		  $RepSys->get ('low'), '-', $RepSys->get ('back'),
#		  '+', $RepSys->get ('Lab'), '+', $RepSys->get
#		  ('rnd'), '+', )];


$IPA->{i} = [(@Root, @vowel, $RepSys->get ('high'), '+', $RepSys->get
('low'), '-', $RepSys->get ('back'), '-')];
$IPA->{j} = [(@Root, @glide, $RepSys->get ('high'), '+', $RepSys->get
('low'), '-', $RepSys->get ('back'), '-')];

$IPA->{e}  = [(@Root, @vowel, $RepSys->get ('high'), '-', $RepSys->get 
('low'), '-', $RepSys->get ('back'), '-')];

$IPA->{ae} = [(@Root, @vowel, $RepSys->get ('high'), '-', $RepSys->get
('low'), '+', $RepSys->get ('back'), '-')];

$IPA->{y} = [(@Root, @vowel, $RepSys->get ('high'), '+', $RepSys->get
	      ('low'), '-', $RepSys->get ('back'), '-', $RepSys->get
	      ('Lab'), '+', $RepSys->get ('rnd'), '+')];

$IPA->{0} = [(@Root, @vowel, $RepSys->get ('high'), '-', $RepSys->get
	      ('low'), '-', $RepSys->get ('back'), '-', $RepSys->get
	      ('Lab'), '+', $RepSys->get ('rnd'), '+')];

$IPA->{oe} = [(@Root, @vowel, $RepSys->get ('high'), '-', $RepSys->get
	      ('low'), '-', $RepSys->get ('back'), '-', $RepSys->get
	      ('Lab'), '+', $RepSys->get ('rnd'), '+')];

$IPA->{'^'} = [(@Root, @vowel, $RepSys->get ('high'), '-',
$RepSys->get ('low'), '-', $RepSys->get ('back'), '+')];

$IPA->{a} = [(@Root, @vowel, $RepSys->get ('high'), '-', $RepSys->get
('low'), '+', $RepSys->get ('back'), '+')];

$IPA->{u} = [(@Root, @vowel, $RepSys->get ('high'), '+', $RepSys->get
	      ('low'), '-', $RepSys->get ('back'), '+', $RepSys->get
	      ('Lab'), '+', $RepSys->get ('rnd'), '+')];
$IPA->{w} = [(@Root, @glide, $RepSys->get ('high'), '+', $RepSys->get
	      ('low'), '-', $RepSys->get ('back'), '+', $RepSys->get
	      ('Lab'), '+', $RepSys->get ('rnd'), '+')];
$IPA->{W} = [(@Root, @vowel, $RepSys->get ('high'), '+', $RepSys->get
	      ('low'), '-', $RepSys->get ('back'), '+')];

$IPA->{o} = [(@Root, @vowel, $RepSys->get ('high'), '-', $RepSys->get
	      ('low'), '-', $RepSys->get ('back'), '+', $RepSys->get
	      ('Lab'), '+', $RepSys->get ('rnd'), '+')];
$IPA->{7} = [(@Root, @vowel, $RepSys->get ('high'), '-', $RepSys->get
	      ('low'), '-', $RepSys->get ('back'), '+')];

$IPA->{O} = [(@Root, @vowel, $RepSys->get ('high'), '-', $RepSys->get
	      ('low'), '+', $RepSys->get ('back'), '+', $RepSys->get
	      ('Lab'), '+', $RepSys->get ('rnd'), '+')];



# Make hash for reverse lookup
for my $ipa_symbol (keys %{$IPA}) {
    my $tree = join ("; ", &to_feature_tree ($ipa_symbol)->to_text);
    if (exists ($text->{$tree})) {
        $text->{$tree} .= " $ipa_symbol";
    } else {
        $text->{$tree} = "$ipa_symbol";
    }
}

#_______________________________________________________________________

sub get {
    my $target = shift;
    $RepSys->get ("$target");
}


sub to_features {
    my $symbol = shift;
    my $ipa = $IPA->{$symbol};
    if ($ipa) {
	@{$ipa};
    } else {
	carp "Unknown symbol [$symbol]: $!";
	confess;
	die;
    }
}

sub to_representation {
    my $string = shift;
    my $rep = $RepSys->get ('MWd')->plus;

    my @symbols = (split /\s/, $string);
    my $s;
    
    $s = shift (@symbols);
    if ($s eq 'F') {
	$rep->put ($RepSys->get ('feminine')->plus);
    } else {
	if ($s eq 'M') {
	    $rep->put ($RepSys->get ('feminine')->minus);
	} else {
	    confess "Unrecognized gender $s";
	}
    }

    return $rep unless @symbols;

    $rep->put ($RepSys->get ('PrWd')->double_anchored);
    $rep->put ($RepSys->get ('Syllable')->double_anchored);
    while (@symbols) {
	$s = shift @symbols;
	if ($s eq '.') {
	    $rep->put ($RepSys->get ('Syllable')->double_anchored);
	    next;
	}
	if ($s eq "\'") {
#	    $s = shift @symbols;
#	    $rep->put_head (to_feature_tree ($s));
	    $rep->set_head;
	    next;
	}
	if (is_symbol ($s)) {
	    $rep->put (to_feature_tree ($s));
	    next;
	}
	confess "Unrecognized symbol $s: $!";
    }
    $rep;
}


sub to_feature_tree {
    my $symbol = shift;
    my $tree;
    my @feature_rep = to_features ($symbol);

    while (@feature_rep) {
	my $feature = shift @feature_rep ;
	my $value = shift @feature_rep;

	if (defined $tree) {
	    $tree->put ($feature->plus) if ($value eq '+');
	    $tree->put ($feature->minus) if ($value eq '-');
	} else {
	    $tree = $feature->plus if ($value eq '+');
	    $tree = $feature->minus if ($value eq '-');
	}
    }
    $tree;
}

# Convert a representation back to ASCII IPA
sub to_IPA {
    my $rep = shift;
    unless ($rep->get_genus_name eq 'Prosodic Word') {
        carp ("Missing PrWd level in representation\n" . $rep->to_text);
    }
    my @syllable_reps = $rep->get_constituent->get_all_constituents;
    my @syllable_IPAs;
    for my $syllable_rep (@syllable_reps) {
        my @segment_reps = $syllable_rep->get_all_constituents;
        my @segment_IPAs;
        if ($rep->is_head_constituent ($syllable_rep)) {
            push @segment_IPAs, "\'";
        }
        for my $segment_rep (@segment_reps) {
            my $segment_rep_text = join ("; ", $segment_rep->to_text);
            my $ipa_symbol = $text->{$segment_rep_text};
	    unless ($ipa_symbol) {
		$ipa_symbol = '?';
	    }
            push @segment_IPAs, $ipa_symbol;
        }
        push @syllable_IPAs, (join " ", @segment_IPAs);
    }
    join " . ", @syllable_IPAs;
}

sub all_features {
    &Representations::get_all_features
}

sub all_symbols {
    keys %{$IPA};
}

sub is_symbol {
    my $symbol = shift;
    grep {$_ eq $symbol} (all_symbols ());
}

1;
