package Learner;
use Globals;
use StimulusSet;
use ConstraintSet;
use warnings;
use strict;
use Carp;
use List::Util qw/ first /;
use Scalar::Util qw/ looks_like_number /;

########################################################################
#
# Learner.pm
#
# Mutation and selection algorithm.
#
# Learner is given training stimuli, each with the desired response,
# and an initial set of constraints
#
# The rule for assigning breeding and survival probabilities implements
# the perceptron learning rule, with sets of extensionally-equivalent
# constraints playing the role of input units, and the sizes of those
# sets playing the role of weights.
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2010 March 21 ($)
#
# Modifications for SCiL 2019 paper:
#
# * Uses Winnow-2AFC update
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


########################################################################
#
# Constructors
#
########################################################################


sub new {
    my $class = shift;
    my ($stimulus_space,
	$initial_constraint_set,
        $meta,          # metaconstraints
        $mutagens,      # MutagenSet
        $recombinators # RecombinatorSet
	) = @_;

    my $this = {};
    bless $this, $class;

    $this->{stimulus_space} = new StimulusSet;
    if (ref ($stimulus_space) and 
	$stimulus_space->isa ('StimulusSet')) {
	for ($stimulus_space->get_all) {
	    $this->add_stimulus ($_);
	}
    }

    if (ref ($initial_constraint_set) and 
	$initial_constraint_set->isa ('ConstraintSet')) {
	$this->{constraints} = $initial_constraint_set;
    } else {
	$this->{constraints} = new ConstraintSet;
    }

    if (ref ($meta) and 
	$meta->isa ('ConstraintSet')) {
	$this->{meta} = $meta;
    } else {
	$this->{meta} = new ConstraintSet;
    }

    $this->{score_matrix} = {};         # What constraint gives what score to what candidate
    $this->{current_harmony} = {};      # Current harmony of each stimulus

    if (ref $mutagens) {
	unless ($mutagens->isa ('MutagenSet')) {
	    confess "Expected MutagenSet, got $mutagens: $!";
	}
	$this->{mutagens} = $mutagens;       # Pointer to hash of mutagens
    }
    if (ref $recombinators) {
	unless ($recombinators->isa ('RecombinatorSet')) {
	    confess "Expected RecombinatorSet, got $recombinators: $!";
	}
	$this->{recombinators} = $recombinators;       # Pointer to hash of recombinators
    }

    @{$this->{tries}} = ();

    $this->{last_trial} = 0;
    $this->{signature_class_names} = {};
    $this->{signature_class_since} = {};
    $this->{total_signature_classes_seen} = 0;
    $this->{signature_class_examples} = {};
    $this->{how_many_constraints} = ($this->get_constraints->how_many or 0);
    $this->{max_signature_class} = 0;

    $this->{lineage} = [];  

    bless $this, $class;

    $this;
}

########################################################################
#
# Access methods
#
########################################################################

sub add_stimulus {
    my $this = shift;
    my $new_stim = shift;
    $this->get_stimulus_space->put ($new_stim);

    $new_stim;
}

sub get_stimulus_space {
    my $this = shift;
    $this->{stimulus_space};
}

sub set_stimulus_space {
    my $this = shift;
    my $stimuli = shift;

    unless ((ref $stimuli) and ($stimuli->isa ('StimulusSet'))) {
	confess "Argument to set_stimulus_space must be StimulusSet, not " . (ref $stimuli) . ":$!";
    }
    $this->{stimulus_space} = $stimuli;
}

sub add_constraint {
    my $this = shift;
    my $new_con = shift;

    unless ((ref $new_con) eq 'Constraint') {
	warn "Type of argument to add_constraint must be Constraint, not " . (ref $new_con) 
	    . ": $!";
	return;
    }

    my $sig = $this->get_signature ($new_con);  # Add to sig table
    $new_con->set_name (
	$this->get_signature_class_number ($sig) . ":" . $new_con->get_unique_id
	);

    $this->get_constraints->put ($new_con);
}

sub get_constraints {
    my $this = shift;
    $this->{constraints};
}

sub set_constraints {
    my $this = shift;
    my $new_constraint_set = shift;

    for ($new_constraint_set->get_all) {
	unless (ref ($_) eq 'Constraint') {
	    confess "Non-constraint ($_) found in constraint set: $!";
	}
	$_->set_name (
	    $this->get_signature_class_number ($this->get_signature ($_)) . ":" .
	    $_->get_unique_id
	    );
    }
    $this->{constraints} = $new_constraint_set;
}

sub set_meta {
    my $this = shift;
    my $meta = shift;
    
    unless (ref ($meta) and $meta->isa ('ConstraintSet')) {
	carp "Argument to set_meta must be ConstraintSet, not " . ref ($meta);
	confess;
    }
    $this->{meta} = $meta;
}

sub get_meta {
    my $this = shift;
    $this->{meta};
}
sub set_initial_population {
    my $this = shift;
    my $initcon = shift;
    
    unless (ref ($initcon) and $initcon->isa ('ConstraintSet')) {
	carp "Argument to set_initial_population must be ConstraintSet, not " . 
	    ref ($initcon);
	confess;
    }
    $this->{initcon} = $initcon;

    $this->reset_population;
}

sub get_initcon {
    my $this = shift;
    $this->{initcon};
}

# Check whether a constraint violates any of the inviolable meta-constraints.
sub violates_hard_meta {
    my $this = shift;
    my $con = shift;

    unless (ref ($con) and $con->isa ('Constraint')) {
	carp "Argument to violates_hard_meta must be Constraint, not " . ref ($con);
	confess;
    }

    my $target = $con->get_target;
    my $meta = $this->get_meta;
    my $bad_flag;

    if ($meta and keys %{$meta}) {
        for my $metaconstraint ($meta->get_all) {
	    my $polarity = $metaconstraint->get_polarity;
	    
	    # The inviolable constraints are the ones with polarity +/-Inf
	    next unless (abs ($polarity) == +"Inf");

            if (
		# Things to definitely not have:
		( ($polarity < 0) and ($metaconstraint->matches_in ($target)) )
		or
		# Things to definitely have:
		( ($polarity > 0) and !($metaconstraint->matches_in ($target)) )
		) {

#	print "DEBUG:  Constraint " . $con->get_name . "\n";
#	print "Which says \n";
#	print (join "\n", $con->to_string);
#	print "\n";
#	print "Violates the hard metaconstraint " . 
#	    $metaconstraint->get_name . "\n";
#	print "which says\n";
#	print (join "\n", $metaconstraint->to_string);
#	print "\n\n";

                $bad_flag = 1;
                last;
            }
        }
    }

#    unless ($bad_flag) {
#	print "DEBUG:  Constraint " . $con->get_name . "\n";
#	print "Which says \n";
#	print (join "\n", $con->to_string);
#	print "\n";
#	print "violates NO hard metaconstraints!\n\n";
#    }


    $bad_flag;
}

# Score a constraint for violations of the "soft" meta-constraints
sub soft_meta_score {
    my $this = shift;
    my $con = shift;

    unless (ref ($con) and $con->isa ('Constraint')) {
	carp "Argument to soft_meta_score must be Constraint, not " . ref ($con);
	confess;
    }

    my $target = $con->get_target;
    my $meta = $this->get_meta;

    my $total_score=0;
    if ($meta and keys %{$meta}) {
        for my $metaconstraint ($meta->get_all) {
	    my $polarity = $metaconstraint->get_polarity;

	    # "soft" metaconstraints have finite polarities
	    next unless (abs ($polarity) < +"Inf");
	    my $score = $metaconstraint->score ($target);
	    $total_score += $score;
        }
    }

#    if ($total_score != 0) {
#	print "DEBUG:  Constraint " . $con->get_name . "\n";
#	print "Which says \n";
#	print (join "\n", $con->to_string);
#	print "\n";
#	
#	print "DEBUG:  Has total score on metaconstraints of $total_score\n";
#    }
    $total_score;
}


sub set_zeta {
    my $this = shift;
    $this->{zeta} = shift;
}

sub get_zeta {
    my $this = shift;
    $this->{zeta} or 0;
}

sub set_N {
    my $this = shift;
    $this->{N} = shift;
}

sub get_N {
    my $this = shift;
    $this->{N} or 1;
}

sub set_eta {
    my $this = shift;
    $this->{eta} = shift;
}

sub get_eta {
    my $this = shift;
    $this->{eta};
}

sub set_mu {
    my $this = shift;
    $this->{mu} = shift;
    if ($this->{mu} < 0 or $this->{mu} > 1) {
	confess "Mutation probability mu must lie in [0,1], not " . $this->{mu};
    };
}

sub get_mu {
    my $this = shift;
    $this->{mu};
}


sub set_parth {
    my $this = shift;
    $this->{parth} = shift;
}

sub get_parth {
    my $this = shift;
    $this->{parth};
}

sub set_momentum {
    my $this = shift;
    $this->{momentum} = shift;
}

sub get_momentum {
    my $this = shift;
    $this->{momentum};
}

sub set_adj {
    my $this = shift;
    my $new_adj = shift;

    if (!defined $new_adj) {
	$this->{adj} = undef;
    } else {
	if ($new_adj eq 'random' 
	    or $new_adj eq 'fitness'
	    or $new_adj eq 'fitnessDraconian' 
	    or $new_adj eq 'fitnessBinary' 
	    or $new_adj eq 'validity'
	    or $new_adj eq 'validityDraconian'
	    or $new_adj eq 'validityBinary'
	    ) {
	    $this->{adj} = $new_adj;
	} else {
	    confess "Unknown population-adjustment method $new_adj";
	}
    }
}

sub get_adj {
    my $this = shift;
    $this->{adj};
}

sub set_clutch_size {
    my $this = shift;
    my $new_r = shift;
    unless (&Scalar::Util::looks_like_number ($new_r)) {
	confess "Clutch size (r) must be number, not $new_r: $!";
    }
    unless ($new_r > 0) {
	confess "Clutch size (r) must be positive, not $new_r: $!";
    }
    $this->{clutch_size} = $new_r;
}

sub get_clutch_size {
    my $this = shift;
    $this->{clutch_size};
}

sub increment_trial {
    my $this = shift;
    $this->{last_trial}++;
}

sub get_last_trial {
    my $this = shift;
    $this->{last_trial};
}

sub set_mutagens {
    my $this = shift;
    my $mutagens = shift;

    unless (ref $mutagens and $mutagens->isa ('MutagenSet')) {
	confess "Expected MutagenSet, got $mutagens: $!";
    }

    $this->{mutagens} = $mutagens;
}

sub get_mutagens {
    my $this = shift;

    $this->{mutagens};
}

sub get_mutagen_named {
    my (
	$this,
	$mutagen_name
	) = @_;
    
    $this->get_mutagens->get_mutagen_named ($mutagen_name);
}

# set_constraint_mutagen, set_representation_mutagen -- Even though 
# representations and constraints are made of the same stuff, we may
# want to mutate them differently.

sub set_constraint_mutagen {
    my $this = shift;
    $this->set_current_mutagen ('constraint', @_);
}

sub set_representation_mutagen {
    my $this = shift;
    $this->set_current_mutagen ('representation', @_);
}

sub set_current_mutagen {
    my (
	$this,
	$which,
	$new_mutagen_name
	) = @_;

    unless ($which eq 'constraint' or $which eq 'representation') {
	confess "First argument to set_mutagen must be either 'constraint' or 'representation', not $which: $!";
    }
    
    $this->{"mutagen for $which"} = $this->get_mutagen_named ($new_mutagen_name);
}

sub get_constraint_mutagen {
    my $this = shift;
    $this->{'mutagen for constraint'};
}

sub get_representation_mutagen {
    my $this = shift;
    $this->{'mutagen for representation'};
}

# And now that all over again for recombinators

sub set_recombinators {
    my $this = shift;
    my $recombinators = shift;

    unless (ref $recombinators and $recombinators->isa ('RecombinatorSet')) {
	confess "Expected RecombinatorSet, got $recombinators: $!";
    }

    $this->{recombinators} = $recombinators;
}

sub get_recombinators {
    my $this = shift;

    $this->{recombinators};
}

sub get_recombinator_named {
    my (
	$this,
	$recombinator_name
	) = @_;
    
    $this->get_recombinators->get_recombinator_named ($recombinator_name);
}

# set_constraint_recombinator, set_representation_recombinator -- Even though 
# representations and constraints are made of the same stuff, we may
# want to mutate them differently.

sub set_constraint_recombinator {
    my $this = shift;
    $this->set_current_recombinator ('constraint', @_);
}

sub set_representation_recombinator {
    my $this = shift;
    $this->set_current_recombinator ('representation', @_);
}

sub set_current_recombinator {
    my (
	$this,
	$which,
	$new_recombinator_name
	) = @_;

    unless ($which eq 'constraint' or $which eq 'representation') {
	confess "First argument to set_recombinator must be either 'constraint' or 'representation', not $which: $!";
    }
    
    $this->{"recombinator for $which"} = $this->get_recombinator_named ($new_recombinator_name);
}

sub get_constraint_recombinator {
    my $this = shift;
    $this->{'recombinator for constraint'};
}

sub get_representation_recombinator {
    my $this = shift;
    $this->{'recombinator for representation'};
}




sub set_constraint_number {
    my $this = shift;
    $this->{how_many_constraints} = shift;
}

sub get_constraint_number {
    my $this = shift;
    $this->{how_many_constraints};
}

sub get_tries {
    my $this = shift;
    @{$this->{tries}};
}

# This memoizes constraint scores
sub get_score {
    my $this = shift;
    my $stim = shift;
    my $con = shift;

    return 0 unless (defined $stim);

    unless (ref ($stim) and $stim->isa ('Stimulus')) {
	carp "Stims must be Stimulus, not $stim: $!";
	confess;
	die;
    }

    unless (ref ($con) and $con->isa ('Constraint')) {
	carp "Constraints must be Constraint, not $con: $!";
	confess;
    }

    my $score = $this->{score_matrix}->{$stim}->{$con->to_target_string};
    if (!defined $score) {
	$score = $con->score ($stim->get_rep);
	$this->{score_matrix}->{$stim}->{$con->to_target_string} = $score;
    }
    $score;
}
	

sub get_current_harmony {
    my $this = shift;
    my $stim = shift;

    unless (ref ($stim) and $stim->isa ('Stimulus')) {
	carp "Stims must be a Stimulus, not $stim: $!";
	confess;
	die;
    }

    $this->{current_harmony}->{$stim->get_name};
}

sub get_most_harmonic {
    my $this = shift;
    my $stim_set = shift;

    unless (ref ($stim_set) and $stim_set->isa ('StimulusSet')) {
	confess "Argument must be StimulusSet, not $stim_set";
    }

    my @stims = $stim_set->get_all;
    my $most_harmonic = shift @stims;
    while (@stims) {
	if ($this->get_current_harmony ($stims [0]) >
	    $this->get_current_harmony ($most_harmonic)) {
	    $most_harmonic = $stims [0];
	}
	shift @stims;
    }
    $most_harmonic;
}


# Remember the fitness of a constraint so that if a replica
# arises, it can be initialized to that fitness
sub store_fitness {
    my (
	$this,
	$constraint
	) = @_;

    unless (ref $constraint and $constraint->isa ('Constraint')) {
	confess "Argument must be Constraint, not $constraint";
    }

    my $as_text = $constraint->to_target_string;
    $this->{constraint_fitness}->{$as_text} = $constraint->get_fitness;
}

sub retrieve_fitness {
    my (
	$this,
	$constraint
	) = @_;

    unless (ref $constraint and $constraint->isa ('Constraint')) {
	confess "Argument must be Constraint, not $constraint";
    }

    my $as_text = $constraint->to_target_string;
    $this->{constraint_fitness}->{$as_text};
}


# Remember the validity of a constraint so that if a replica
# arises, it can be initialized to that validity
sub store_validity {
    my (
	$this,
	$constraint,
	$validity
	) = @_;

    unless (ref $constraint and $constraint->isa ('Constraint')) {
	confess "First argument must be Constraint, not $constraint";
    }

    my ($v, $total_diff, $n_trials) = @{$validity};    

    unless ($v >= -1 and $v <= 1) {
	confess "First element of second argument must be between -1 and 1, not $v in ($v, $total_diff, $n_trials)";
    }

    my $as_text = $constraint->to_target_string;
    $this->{constraint_validity}->{$as_text} = [$v, $total_diff, $n_trials];
}

sub retrieve_validity {
    my (
	$this,
	$constraint
	) = @_;

    unless (ref $constraint and $constraint->isa ('Constraint')) {
	confess "Argument must be Constraint, not $constraint";
    }

    my $as_text = $constraint->to_target_string;
    my $validity = $this->{constraint_validity}->{$as_text};
    unless ($validity) {
	[0, 0, 0];
    } else {
	$validity;
    }
}


# Record which signature classes spawned which others, and
# when it happened.
sub record_lineage {
    my (
	$this,
	$mom, $dad, $kid  # All of type Stimulus
	) = @_;
    
    my $mom_sig = $this->get_signature ($mom);
    my $mom_class = $this->get_signature_class_number ($mom_sig);
    my $kid_sig = $this->get_signature ($kid);
    my $kid_class = $this->get_signature_class_number ($kid_sig);	
    my $now = $this->get_last_trial;
    my ($dad_sig, $dad_class);
    # In case of parthenogenesis, there is no dad:
    if (defined $dad) {
	$dad_sig = $this->get_signature ($dad);
	$dad_class = $this->get_signature_class_number ($dad_sig);
	push @{$this->{lineage}->[$mom_class]->[$dad_class]->[$kid_class]}, $now;
    } else {
	push @{$this->{lineage}->[$mom_class]->[0]->[$kid_class]}, $now;
    }
}

sub get_lineage {
    my $this = shift;
    $this->{lineage};
}

########################################################################
#
# Update methods
#
########################################################################

sub update_score_matrix {
    my $this = shift;
    for my $stim ($this->get_stimulus_space->get_all) {
	for my $con ($this->get_constraints->get_all) {
	    my $score = $this->get_score ($stim, $con);
	}
    }
    $this->{score_matrix};
}



sub update_harmony {
    my $this = shift;
    
    my $zeta = $this->get_zeta;
    for my $stim ($this->get_stimulus_space->get_all) {
	my $H;
	for my $con ($this->get_constraints->get_all) {
	    $H += $zeta * $this->get_score ($stim, $con);
	}
	$this->{current_harmony}->{$stim->get_name} = $H;
    }
    $this->{current_harmony};
}


sub update_validities {
    my (
	$this,
	$emp, 
	$other_candidates
	) = @_;

    unless (ref $emp and $emp->isa ('Stimulus')) {
	confess "First argument should be Stimulus, not $emp";
    }
    unless (ref $other_candidates and $other_candidates->isa ('StimulusSet')) {
	confess "Second argument should be StimulusSet, not $other_candidates";
    }
	
    my @other_candidates = $other_candidates->get_all;
    my ($foil) = $this->guess (\@other_candidates);

#    &Report::print_to_log ("emp : " . $emp->to_flat_string);
#    &Report::print_to_log ("foil: " . $foil->to_flat_string);

    for my $constraint ($this->get_constraints->get_all) {
#        &Report::print_to_log ("Updating validity for " . 
#			       $constraint->to_flat_string);
	my $score_emp  = $this->get_score ($emp, $constraint);
	my $score_foil = $this->get_score ($foil, $constraint);

	my $diff = $score_emp - $score_foil;
	
#	&Report::print_to_log ("Score:  emp = $score_emp, foil = $score_foil, diff = $diff");

        my $old_validity = $this->retrieve_validity ($constraint);

	my ($v, $total_diff, $n_trials) = @{$old_validity};
#	&Report::print_to_log ("Validity was:  $v; n_trials was: $n_trials; total difference was: $total_diff");
	$n_trials++;
	$total_diff += $diff;
	$v = $total_diff/$n_trials;
#	&Report::print_to_log ("Validity now:  $v; n_trials now: $n_trials; total difference now: $total_diff");
	$this->store_validity ($constraint, [$v, $total_diff, $n_trials]);
    }
}


# Update the fitness of every constraint based on the outcome of a trial
sub update_fitnesses {
    my $this = shift;
    my $emp  = shift;  # The candidate that should have been chosen
    my $pred = shift;  # The candidate that was actually chosen

    my $eta      = $this->get_eta;
    my $zeta     = $this->get_zeta;
    my $momentum = $this->get_momentum;

    for my $constraint ($this->get_constraints->get_all) {
#        &Report::print_to_log ("Updating fitness for " . 
#			       $constraint->to_flat_string);

	my $signature_class_number = $this->get_signature_class_number (
	    $this->get_signature ($constraint));

	my $score_emp  = $this->get_score ($emp, $constraint);
	my $score_pred = $this->get_score ($pred, $constraint);
	my $diff = $score_emp - $score_pred;
	
#	my $yay = (abs ($diff) > 0) ? "XXXXXX" : "";
#	&Report::print_to_log ("score_emp = $score_emp; score_pred = $score_pred; diff = $diff; $yay");

	if ($diff > 0) {
	    $this->record_win ($signature_class_number)
	}
	if ($diff == 0) {
	    $this->record_tie ($signature_class_number)
	}
	if ($diff < 0) {
	    $this->record_loss ($signature_class_number)
	}

	# Metaconstraint adjustment:
	my $delta = $this->soft_meta_score ($constraint);

	# Winnow update
	my $winnow_fitness = (1+$eta)**($diff+$delta*$zeta);

#	&Report::print_to_log ("delta = $delta; winnow_fitness = $winnow_fitness");

	my $old_fitness = $this->retrieve_fitness ($constraint);
	$old_fitness = 1 unless defined $old_fitness;
	my $new_fitness = (1-$momentum) * $winnow_fitness + 
	    $momentum * $old_fitness;

#        &Report::print_to_log ("old fitness = " . 
#			       $constraint->get_fitness . 
#			       "; new fitness = $winnow_fitness");
	$constraint->set_fitness ($new_fitness);
	$this->store_fitness ($constraint);
    }
}


########################################################################
#
# Experimental trials
#
########################################################################

sub pr_corr {
    my (
	$this,
	$correct_stim,
	$other_stims,
	$except_constraint    # Constraint to omit in calculations
	) = @_;
    my @other_stims = $other_stims->get_all;
    my @stimuli = ($correct_stim, @other_stims);

    my $zeta = $this->get_zeta;
    my @harmonies = map {
	my $h = $this->get_current_harmony ($_);
	$h = 0 unless defined $h;
	if ($except_constraint) {
	    $h = $h - $zeta * $this->get_score ($_, $except_constraint)
	}
	$h
    } @stimuli;

    my $Z;
    for (@harmonies) {
	$Z += exp ($_);
    }
    
    exp ($harmonies [0])/$Z;
}

sub pr_err {
    1-&pr_corr (@_);
}

# Assign probabilities and choose a candidate using Max Ent
sub guess {
    my (
	$this,
	$stimuli
	) = @_;
    
    unless ($stimuli =~ /ARRAY/) {
	confess "Argument should be array reference, not $stimuli";
    }

    my @stimuli = @{$stimuli};

    my @harmonies = map {
	my $h = $this->get_current_harmony ($_);
	$h = 0 unless defined $h;
	$h;
    } @stimuli;
    my $Z;
    for (@harmonies) {
	$Z += exp ($_);
    }
    my @pr = map {
        exp ($_)/$Z
    } @harmonies;
    my $r = rand (1);
    my $i=0;
    my $p;
    while ($i <= $#pr) {
	$p += $pr [$i];
	last if $p > $r;
	$i++;
    }
    my $guess = $stimuli [$i];

    return (
	$guess,   # The chosen one
	\@pr      # The probabilities of the candidates
	);
}

sub prediction {
    my (
	$this,
	$correct_stim,
	$other_stims
	) = @_;
    
    my @other_stims = $other_stims->get_all;
    my @stimuli = ($correct_stim, @other_stims);

    my ($guess, $pr_ref) = $this->guess (\@stimuli);
    my @pr = @{$pr_ref};

    # Probability that learner guessed right
    my $pr_corr = $pr [0];
    unless (defined $pr_corr) {
	confess "Undefined probability $pr_corr: $!";
    }

    # Whether the learner really did guess right
    my $corr = 0 + ($guess eq $correct_stim);
    unless (defined $corr) {
	confess "Undefined correct/incorrect score $corr: $!";
    }

    return ($pr_corr, $corr, $correct_stim, $guess);
}


########################################################################
#
# Methods related to performance measures
#
########################################################################

# Total score assigned by a constraint to all of the training stims
sub total_score {
    my $this = shift;
    my $con = shift;
    
    my $score = 0;
    for ($this->get_stimulus_space->get_all) {
	my $con_text = $con->to_target_string;
	$score += $this->get_score ($_, $con) * $_->get_weight;
    }

    $score;
}

# For counting wins/ties/losses by each signature class
sub record_outcome {
    my (
	$this,
	$signature_class_number,
	$outcome
	) = @_;

    $this->{$outcome}->[$signature_class_number]->[$this->get_last_trial]++;
}

sub count_outcome {
    my (
	$this,
	$signature_class_number,
	$outcome
	) = @_;

    my $outcomes = $this->{$outcome}->[$signature_class_number];
    return 0 unless $outcomes;
    grep {
	defined $_ 
    } @{$outcomes};
}

sub count_outcomes {
    my (
	$this,
	$signature_class_number
	) = @_;
    
    my $wins = $this->count_outcome ($signature_class_number, "win");
    my $ties = $this->count_outcome ($signature_class_number, "tie");
    my $losses = $this->count_outcome ($signature_class_number, "loss");

    ($wins, $ties, $losses);
}

sub record_win {
    my $this = shift;
    my $signature_class_number = shift;

    $this->record_outcome ($signature_class_number, "win")
}

sub record_loss {
    my $this = shift;
    my $signature_class_number = shift;

    $this->record_outcome ($signature_class_number, "loss")
}

sub record_tie {
    my $this = shift;
    my $signature_class_number = shift;

    $this->record_outcome ($signature_class_number, "tie")
}

    

########################################################################
#
# Population methods
#
########################################################################


sub adjust_population {
    my $this = shift;
    my $adj = shift;

    $adj = $this->get_adj unless $adj;
    my $N = $this->get_N; 

    my $constraints = $this->get_constraints; 
    &Report::print_to_log ("We have " . $constraints->how_many . 
			   " but need $N");

    if ($constraints->how_many > $N) {
	my @deleted = ();
	my $N_to_delete = $constraints->how_many - $N;
	&Report::print_to_log ("Need to delete $N_to_delete constraints to have $N");

	if ($adj eq 'random') {
	    @deleted = @{$constraints->delete_random (
			     $N_to_delete)};
	}

	if ($adj eq 'fitness') {
	    @deleted = @{$constraints->delete_least_fit (
				$N_to_delete)};
	}
	if ($adj eq 'fitnessDraconian') {
	    # Remove every constraint that has ever helped cause an error
	    my @constraints = List::Util::shuffle ($constraints->get_all);

	    while (@deleted < $N_to_delete) {
		last unless @constraints;
		my $constraint = shift @constraints;
		if ($constraint->get_fitness < 1) {
		    $constraints->delete ($constraint);
		    push @deleted, $constraint;
		    next;
		}
	    }
	    $this->set_constraints ($constraints);
	    my $still_needed = $N_to_delete - @deleted;

	    # Any still needed, delete least fit
	    if ($still_needed > 0) {
		push @deleted, @{$constraints->delete_least_fit (
				$still_needed)};
	    }
	}

	if ($adj eq 'fitnessBinary') {
	    # Remove every constraint that has ever helped cause an error,
	    # unless that's all we've got
	    my @constraints = List::Util::shuffle ($constraints->get_all);

	    my $N_deleted=0;
	    for (@constraints) {
		if ($_->get_fitness < 1) {
		    $constraints->delete ($_);
		    push @deleted, $_;
		    $N_deleted++;
		}
	    }
	    
	    # Any still needed, delete randomly:
	    if ($N_deleted < $N_to_delete) {
		push @deleted, @{$constraints->delete_random ($N_to_delete - $N_deleted)};
	    }

	    $this->set_constraints ($constraints);
	}

	if ($adj eq 'validity') {
	    my @constraints = List::Util::shuffle ($constraints->get_all);

	    my @nonperfect = grep {
		$this->retrieve_validity ($_)->[0] < 1;
	    } @constraints;
	    my @perfect = grep {
		$this->retrieve_validity ($_)->[0] == 1;
	    } @constraints;
	    while (@deleted < $N_to_delete) {
		my $constraint = shift @nonperfect;
		my $v = $this->retrieve_validity ($constraint)->[0];

		# This would return $p=0 for the ones in @perfect:
		my $p = ($v>=0) ? 1-(1/2 + $v/2) : 1-(1/2 - (-$v)/2);

		if (rand (1) < $p) {
		    $constraints->delete ($constraint);
		    push @deleted, $constraint;
		    next;
		} else {
		    push @nonperfect, $constraint;
		}
	    }
	    while (@deleted < $N_to_delete) {
		push @deleted, pop @perfect;
	    }
	    $this->set_constraints ($constraints);
	}

	if ($adj eq 'validityDraconian') {
	    my @constraints = List::Util::shuffle ($constraints->get_all);
	    my @nonperfect = grep {
		$this->retrieve_validity ($_)->[0] < 1;
	    } @constraints;
	    my @perfect = grep {
		$this->retrieve_validity ($_)->[0] == 1;
	    } @constraints;

	    while (@deleted < $N_to_delete and @nonperfect) {
		my $constraint = shift @nonperfect;
		my $v = $this->retrieve_validity ($constraint)->[0];

		# DEBUG
		my $p;
		if ($v < -0.005) {
		    $p = 1;
		} else {
		    $p = ($v>=0) ? 1-(1/2 + $v/2) : 1-(1/2 - (-$v)/2);
		}
		if (rand (1) < $p) {
		    $constraints->delete ($constraint);
		    push @deleted, $constraint;
		    next;
		} else {
		    push @nonperfect, $constraint;
		}
	    }

	    while (@deleted < $N_to_delete) {
		push @deleted, pop @perfect;
	    }

	    $this->set_constraints ($constraints);
	}

	if ($adj eq 'validityBinary') {
	    my @constraints = List::Util::shuffle ($constraints->get_all);

	    while (@deleted < $N_to_delete) {
		last unless @constraints;
		my $constraint = shift @constraints;
		if ($this->retrieve_validity ($constraint)->[0] < -0.005) {   # DEBUG
		    $constraints->delete ($constraint);
		    push @deleted, $constraint;
		    next;
		}
	    }
	    $this->set_constraints ($constraints);
	    my $still_needed = $N_to_delete - @deleted;

	    # Any still needed, delete randomly
	    if ($still_needed > 0) {
		push @deleted, @{$constraints->delete_random (
				$still_needed)};
	    }
	}

	&Report::print_to_log ("Deleted the following constraints ($adj-based selection):");
	for (@deleted) {
	    my $fitness = $_->get_fitness;
	    my $validity = $this->retrieve_validity ($_)->[0];
	    &Report::print_to_log ("(f:$fitness, v:$validity) " . $_->to_flat_string);
	}
	&Report::print_to_log ("which left the following:");
	for ($this->get_constraints->get_all) {
	    my $fitness = $_->get_fitness;
	    my $validity = $this->retrieve_validity ($_)->[0];
	    &Report::print_to_log ("(f:$fitness, v:$validity) " . $_->to_flat_string);
	}
    }

    if ($constraints->how_many == 0) {
	&Report::print_to_log ("Deleted all the constraints; resetting to initial population");
	$this->reset_population;
    }

    $constraints = $this->get_constraints;

    if ($constraints->how_many < $N) {
	my $needed = $N - $constraints->how_many;
	my @duplicated = @{$constraints->duplicate_random ($needed)};
	for (@duplicated) {
	    &Report::print_to_log ("Added duplicate " . $_->to_flat_string);
	}
    }

    $this->set_constraints ($constraints);
}


sub make_kids {
    my $this = shift;
    my @breeders = @_;

    my $constraint_recombinator = $this->get_constraint_recombinator;
    my $zeta = $this->get_zeta;
    my $clutch_size = $this->get_clutch_size;

    # How many parthenogenetic vs. two-parent kids do we need?
    my $N_parth_moms_needed = int (@breeders * $this->get_parth + 0.5);
    my $N_nonparth_moms_needed = @breeders - $N_parth_moms_needed;

    # Make unmutated kids
    my @kids;
    # Breeders breed with each other
    for my $mom (@breeders) {
	# Decide whether this breeder breeds parthenogenetically
	my $pr_parth = $N_parth_moms_needed/($N_parth_moms_needed + 
					     $N_nonparth_moms_needed);
	my $parth = (rand (1) < $pr_parth);
	if ($parth) {
	    $N_parth_moms_needed--;
	} else {
	    $N_nonparth_moms_needed--;
	}

	for my $egg (1..$clutch_size) {
#	    &Report::print_to_log ("Egg $egg of $clutch_size");
	    my ($dad, $kid);

	    # Keep trying until we get a viable new constraint.
	    while (1) {
		if ($parth) {
		    $kid = $mom->clone;
		    $dad = undef;
		} else {
		    # Mom *must* breed up, so that advantageous mutations
		    # aren't swamped by back-crossing
		    $dad = $breeders [rand @breeders];
		    while ($mom->get_fitness < $dad->get_fitness) {
			$dad = $breeders [rand @breeders];
		    }
		    $kid = $mom->breed ($dad, $constraint_recombinator);
		    $kid->set_fitness (
			($mom->get_fitness + $dad->get_fitness)/2);
		}

		# Better not violate any of the hard constraints on
		# representations (*[+high +low], etc.)
		next if (
		    &Globals::get_representations->is_forbidden (
			 $kid->get_target)
		    );
		# Retry also if kid violates any of the hard metaconstraints
		next if ($this->violates_hard_meta ($kid));
		
		last;
	    }
	    push @kids, [$kid, $mom, $dad];
	}
    }
    &Report::print_to_log ("Made " . scalar (@kids) . " unmutated kids");

    return (@kids);

}

sub mutate_kids {
    my (
	$this, 
	@unmutated_kids
	) = @_;

    @unmutated_kids = List::Util::shuffle (@unmutated_kids);

    # Mutate the kids.  We want a particular fraction of actual mutants:
    my $mu = $this->get_mu;
    my $N_kids = @unmutated_kids;
    my $N_mutants = int ($N_kids * $mu + 0.5);
    my $N_nonmutants = $N_kids - $N_mutants;
    my $constraint_mutagen = $this->get_constraint_mutagen;

    &Report::print_to_log ("Need $N_mutants mutants, and $N_nonmutants non-mutants");

    my @mutated_kids;
    while (@mutated_kids < $N_mutants) {
	my ($unmutated_kid, $mom, $dad) = @{$unmutated_kids [0]};

	my $mutated_kid;
	while (1) {
	    $mutated_kid = $unmutated_kid->mutant ($constraint_mutagen);
	    
	    # Insure non-vacuous mutation
	    next if ($mutated_kid->to_target_string eq
		     $unmutated_kid->to_target_string);

	    # Better not violate any of the hard constraints on
	    # representations (*[+high +low], etc.)
	    next if (
		&Globals::get_representations->is_forbidden (
		     $mutated_kid->get_target)
		);
	    
	    # Retry also if kid violates any of the hard metaconstraints
	    next if ($this->violates_hard_meta ($mutated_kid));

	    &Report::print_to_log ("Got a well-formed mutant:",
		"Mutated from " . $unmutated_kid->to_flat_string,
		"Mutated to   " . $mutated_kid->to_flat_string);

	    # Have we seen this constraint before?
	    my $stored_fitness = $this->retrieve_fitness ($mutated_kid);
	    my $stored_validity = $this->retrieve_validity ($mutated_kid)->[0];
	    if (defined $stored_fitness) {
#		&Report::print_to_log ("We've seen this mutant before, the last time with fitness $stored_fitness and validity $stored_validity");
		$mutated_kid->set_fitness ($stored_fitness);
	    }
	    last;
	}
	
	push @mutated_kids, [$mutated_kid, $mom, $dad];
	shift @unmutated_kids;
    }
    
    # Get the right number of non-mutants
    my @nonmutants;
    while (@nonmutants < $N_nonmutants) {
	my ($kid, $mom, $dad) = @{$unmutated_kids [0]};
	push @unmutated_kids, (shift @unmutated_kids);
	push @nonmutants, [$kid->clone, $mom, $dad];
    }

    # Combine
    my @kids;
    for ((@mutated_kids, @nonmutants)) {
	my ($kid, $mom, $dad) = @{$_};
	$this->record_lineage ($mom, $dad, $kid);
	push @kids, $kid;
    }
    @kids;
}

sub memorize_stim {
    my (
	$this,
	$params,
	$correct_stim,
	$chosen_stim
	) = @_;

    return unless (defined $params->{p_memorize} and 
		   $params->{p_memorize} > 0);
    return unless (rand (1) < $params->{p_memorize});
    my $n = $this->get_clutch_size;
    if (defined $params->{n_memorize}) {
	unless ($params->{n_memorize} >= 0) {
	    confess "Can't memorize " . $params->{n_memorize} . " stimuli: $!";
	}
	$n = $params->{n_memorize};
    }
    return unless $n;
    if (defined $params->{mem_polarity}) {
	return unless (abs ($params->{mem_polarity}) == 1);
    }
    my $c_stim;
    if ($params->{mem_polarity} > 0) {
	$c_stim = new Constraint (
	    $correct_stim->get_rep->clone,
	    $params->{mem_polarity},
	    $correct_stim->get_name
	    );
    } else {
	$c_stim = new Constraint (
	    $chosen_stim->get_rep->clone,
	    $params->{mem_polarity},
	    $correct_stim->get_name
	    )
    }

    unless ($this->violates_hard_meta ($c_stim)) {
	$this->store_validity ($c_stim, [1,1,1]);
	$this->add_constraint ($c_stim);
	if ($n>1) {
	    for (2..$n) {
		$this->add_constraint ($c_stim->clone);
	    }
	}
        &Report::print_to_log ("Memorized $n copies of " . 
			       $c_stim->to_flat_string);
    }
}
    

########################################################################
#
# Methods relating to tracking extensional equivalence of constraints
#
########################################################################

# A constraint's "signature" is the pattern of marks that it assigns
# to all of the training data.  Constraints with the same signature are
# effectively the same constraint, even if they are formally different.

sub get_signature {
    my $this = shift;
    my $con = shift;

    unless ($con and ref $con) {
	confess "Argument to get_signature should be a Constraint, not $con: $!";
    }

    my $key = $con->to_target_string;
    my $sig = $this->{signature}->{$key};

    unless (defined $sig) {
	for my $stim ($this->get_stimulus_space->get_all) {
	    $sig .= (0+$this->get_score ($stim, $con)) . ".";
	}
	$this->{signature}->{$con->to_target_string} = $sig;
    
	unless (defined ($this->{signature_class_since}->{$sig})) {
	    $this->{signature_class_since}->{$sig} = $this->get_last_trial;
	}
	unless (defined ($this->{signature_class_number}->{$sig})) {
	    $this->{signature_class_number}->{$sig} = ++$this->{max_signature_class};
	}
    }
    unless (defined ($this->{signature_class_examples}->{$sig})) {
	$this->{signature_class_examples}->{$sig} = $con;
    }

    $sig;
}

sub get_all_signatures {
    my $this = shift;
    
    keys %{$this->{signature_class_number}}
}

sub count_signature_classes {
    my $this = shift;

    my $count = $this->{max_signature_class};
    
    unless ($count) {
	for my $con ($this->get_constraints->get_all) {
	    $this->get_signature ($con);
	}
	$count = $this->{max_signature_class};
    }
    $count;
}

sub get_signature_class_number {
    my $this = shift;
    my $sig = shift;
    $this->{signature_class_number}->{$sig};
}

sub get_signature_class_since {
    my $this = shift;
    my $sig = shift;

    $this->{signature_class_since}->{$sig};
}

sub get_signature_class {
    my $this = shift;
    my $sig = shift;

    grep {
	$sig eq $this->get_signature ($_);
    } ($this->get_constraints->get_all);
}

sub get_signature_class_example {
    my $this = shift;
    my $sig = shift;

    my @class = $this->get_signature_class ($sig);
    my $ex = ($this->{signature_class_examples}->{$sig}) or pop @class;
    my $ex_string = "(none)";
    if (ref $ex and $ex->can ('to_string')) {
        $ex_string = $ex->to_string;
    }

    # Get shortest example
    for (@class) {
        my $as_string = $_->to_string;
        if (length ($ex_string) > length ($as_string)) {
            $ex = $_;
            $ex_string = $as_string;
        }
    }
    
    $this->{signature_class_examples}->{$sig} = $ex;
    $ex;
}

sub reset_population {
    my $this = shift;
    
    my $initcon = $this->get_initcon;
    $this->{constraints} = new ConstraintSet;
    for ($initcon->get_all) {
	$this->add_constraint ($_);
    }
}

sub initialize {
    my $this = shift;
    $this->reset_population;
    unless ($this->get_constraint_number) {
	$this->set_constraint_number ($this->get_constraints->how_many);
    }
    $this->set_N ($this->get_constraint_number);
    $this->adjust_population ('random');

    $this->{initialized} = 1;
    $this->update_harmony;
}	

sub is_initialized {
    my $this = shift;
    $this->{initialized} or 0;
}

########################################################################
#
# Learning methods
#
########################################################################

sub learn {
    my $this = shift;
    my $params = shift;

    unless ($this->is_initialized) {
	$this->initialize;
    } 

    $this->increment_trial;
    &Report::print_to_log ("\nTRIAL " . $this->get_last_trial);
    
    # Set learner parameters according to what is specified in the
    # trial descriptor
    $this->set_eta ($params->{eta}) if (defined $params->{eta});
    $this->set_zeta ($params->{zeta}) if (defined $params->{zeta});
    $this->set_parth ($params->{parth}) if (defined $params->{parth});
    $this->set_mu ($params->{mu}) if (defined $params->{mu});
    $this->set_momentum ($params->{momentum}) if (defined $params->{momentum});
    $this->set_adj ($params->{adj}) if (defined $params->{adj});

    $this->set_clutch_size ($params->{clutch_size}) if defined ($params->{clutch_size});

    $this->set_constraint_mutagen ($params->{const_mutagen}) 
	if defined ($params->{const_mutagen});
    $this->set_constraint_recombinator ($params->{const_recombinator}) 
	if defined ($params->{const_recombinator});
    if (defined $params->{N}) {
	$this->set_N ($params->{N}); 
    }
    # In case population size has changed:
    $this->adjust_population;

    my $correct_stim = ($this->get_stimulus_space->get_stimulus_named (
			    $params->{correct_stim}))[0];

    &Report::print_to_log ("Correct: " . $correct_stim->to_flat_string);

    # If no alternative candidates are defined, use the entire stimulus set
    my $other_candidates = new StimulusSet;
    if (defined $params->{other_candidates} and $params->{other_candidates} ne 'NA') {
	map {
	    $other_candidates->put (($this->get_stimulus_space->get_stimulus_named ($_))[0]);
	} @{$params->{other_candidates}};
    } else {
	map {
	    unless ($_ eq $correct_stim) {
		$other_candidates->put ($_);
	    }
	} ($this->get_stimulus_space->get_all);
    }

    # Compare predicted with correct stim
    my ($pr_corr, $corr, $chosen_stim);
    ($pr_corr, $corr, $correct_stim, $chosen_stim) = 
	$this->prediction ($correct_stim, $other_candidates);

    $this->update_validities ($correct_stim, $other_candidates);
    if ($corr) {
	return ($chosen_stim->get_name, $pr_corr, $corr);
    }
    $this->update_fitnesses  ($correct_stim, $chosen_stim);

    # The fitnesses determine the number of offspring which
    # the constraint is likely to leave.
    my @breeders;
    my @survivors;   # Not used presently
    for my $constraint ($this->get_constraints->get_all) {
	my $fitness = $constraint->get_fitness;
	while ($fitness >= 1) {
#	    &Report::print_to_log ("Breeder ($fitness):  " . $constraint->to_flat_string);
	    $fitness--;
	    push @breeders, $constraint;
	} 
	if ($fitness > 0) {
	    if (rand (1) < $fitness) {
		push @breeders, $constraint;
#		&Report::print_to_log ("Breeder ($fitness):  " . $constraint->to_flat_string);
	    }
	}
    }
    &Report::print_to_log ((scalar @breeders) . " breeders");

    # Avoid mass extinction by saving a pair
    if (@breeders + @survivors < 2) {
        &Report::print_to_log ("Averting mass extinction by saving a pair");
    }
    while (@breeders < 2) {
	push (@breeders, $this->get_constraints->get_one);
    }
    
    my $NextGeneration = new ConstraintSet;
    # Breeders breed with each other
    $NextGeneration = new ConstraintSet (@survivors);
    my @kids = $this->mutate_kids ($this->make_kids (@breeders));
    for my $kid (@kids) {
	$NextGeneration->put ($kid);
    }
    $this->set_constraints ($NextGeneration);
    
    # Memorize relevant training stimulus if desired
    $this->memorize_stim ($params, $correct_stim, $chosen_stim);

    $this->adjust_population;

    return ($chosen_stim->get_name, $pr_corr, $corr);
}

########################################################################
#
# I/O methods
#
########################################################################

sub to_text {
    my (
	$this,
	$depth
	) = (@_, 0);

    my @text;
    push @text, (
	($Globals::indent x $depth) .
	"Learner" 
    );
    push @text, $this->get_stimulus_space->to_text ($depth+1);
    push @text, $this->get_constraints->to_text ($depth+1);
    push @text, $this->get_meta->to_text ($depth+1);
    push @text, $this->get_mutagens->to_text ($depth+1);
    push @text, $this->get_recombinators->to_text ($depth+1);

    @text;
}

sub from_text {
    my $this = shift;
    # Pass text in as ref to array, or array itself
    my $text = (ref $_[0] eq 'ARRAY') ? $_[0] : \@_;

    # Return empty-handed unless the text begins by describing a Learner
    unless (
        $text->[0] =~ /^((?:$Globals::indent_re)*)
                    Learner\s+$/x) {
        return;
    }
    my $depth  = length ($1)/length ($Globals::indent);
    shift @{$text};

    my $next_line = $text->[0];
    $next_line =~ /^((?:$Globals::indent_re)*)/;
    my $next_depth = length ($1)/length ($Globals::indent);
    if ($next_depth <= $depth) {
        confess "Expected StimulusSet, found $next_line: $!";
    }
    my ($stimulus_space) = &StimulusSet::from_text ($text);
    unless ($stimulus_space) {
        confess "Couldn't parse $next_line: $! (1)";
    }

    $next_line = $text->[0];
    $next_line =~ /^((?:$Globals::indent_re)*)/;
    $next_depth = length ($1)/length ($Globals::indent);
    if ($next_depth <= $depth) {
        confess "Expected constraint set, found $next_line: $!";
    }
    my ($con) = &ConstraintSet::from_text ($text);
    unless ($con) {
        confess "Couldn't parse $next_line: $! (2)";
    }

    $next_line = $text->[0];
    $next_line =~ /^((?:$Globals::indent_re)*)/;
    $next_depth = length ($1)/length ($Globals::indent);
    if ($next_depth <= $depth) {
        confess "Expected constraint set, found $next_line: $!";
    }
    my ($forbid) = &ConstraintSet::from_text ($text);
    unless ($forbid) {
        confess "Couldn't parse $next_line: $! (3)";
    }


    $next_line = $text->[0];
    $next_line =~ /^((?:$Globals::indent_re)*)/;
    $next_depth = length ($1)/length ($Globals::indent);
    if ($next_depth <= $depth) {
        confess "Expected MutagenSet, found $next_line: $!";
    }
    my ($mutagens) = &MutagenSet::from_text ($text);
    unless ($mutagens) {
	confess "Couldn't parse $next_line: $! (4)";
    }

    $next_line = $text->[0];
    $next_line =~ /^((?:$Globals::indent_re)*)/;
    $next_depth = length ($1)/length ($Globals::indent);
    if ($next_depth <= $depth) {
        confess "Expected RecombinatorSet, found $next_line: $!";
    }
    my ($recombinators) = &RecombinatorSet::from_text ($text);
    unless ($recombinators) {
	confess "Couldn't parse $next_line: $! (4)";
    }

    my $new;
    unless ($new = new Learner ($stimulus_space, $con, $forbid,
				$mutagens, $recombinators)) {
	carp "Error in $next_line: $!";
        carp "Couldn't add constraint target.";
        confess;
        die;
    }

    $new;
}

sub print {
    my $this = shift;
    print (join "\n", $this->to_text);
    print "\n";
}

# Report the distinct classes of extensionally-equivalent constraints, and
# give an example of each.
sub report_population {
    my $this = shift;
    local *FH = (shift  or *STDOUT);
    my $format = (shift or 'text');
    unless ($format eq 'text' or $format eq 'latex' or $format eq 'flat') {
	carp "Unknown format $format: $!";
	confess;
	die;
    }

    $this->update_harmony;

    my %signature_example;
    my %signature_fitness_total;
    my $none = 0;


    # Entropy of the constraint set
    my $entropy = 0;
    my $how_many_constraints = $this->get_constraints->get_all;
    for my $sig ($this->get_all_signatures) {
	my $p = $this->get_signature_class ($sig)/$how_many_constraints;
	next unless $p;
	$entropy += -$p * log ($p)/log (2);
    }
    
    printf FH "%d classes; entropy = %5.3f\n", $this->count_signature_classes, $entropy;

    for my $sig (
	sort {
	    $this->get_signature_class ($b) <=> $this->get_signature_class ($a)
	} ($this->get_all_signatures)
	) {

	# Omit macro-constraints with no micro-constraints
        my $class_size=0;
        for my $con ($this->get_signature_class ($sig)) {
            $class_size++;
        }
        next unless $class_size;


	my $signum = 	$this->get_signature_class_number ($sig);
	printf FH "CLASS %s (since %d, now %d):  ", 
	$signum,
	$this->get_signature_class_since ($sig),
        $this->get_last_trial;

	my $ex = $this->get_signature_class_example ($sig);
	unless ($ex) {
	    print "\n";
	    next;
	}

	my @signature_class = $this->get_signature_class ($sig);
	my $total_fitness;
	my $total_validity;
	for (@signature_class) {
	    $total_fitness += $_->get_fitness;
	    $total_validity += $this->retrieve_validity ($_)->[0];
	}
	my $mean_fitness = $total_fitness/@signature_class;
	my $mean_validity = $total_validity/@signature_class;

	my ($wins, $ties, $losses) = $this->count_outcomes ($signum);
	printf FH "%d micro-constraints (w:%d, t:%d, l:%d; mf:%6.3f; mv:%6.3f)\n", 
	$class_size, $wins, $ties, $losses, $mean_fitness, $mean_validity;

	if ($format eq 'text') {
	    print FH ($this->get_signature ($ex));
	    print FH "\n";
	    print FH (join ("\n", $ex->to_text));
	    print FH "\n";
	} elsif ($format eq 'latex') {
	    print FH $ex->to_tree;
	} else {
	    print FH $ex->to_flat_string;
	    print FH "\n";
	}
    }
    print FH "\n";
        
}


1;
