package MCat;
use Globals;
use Environment;
use MCatGenus;
use warnings;
use strict;
use Carp;

########################################################################
#
# MCat.pm
#
# Object implementing a morphological category.  An MCat has a single
# PrCat dependent, plus zero or more unordered dependent Features.
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2019 January 26 ($)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


our @ISA = qw( UnaryFeature );

########################################################################
#
# Constructors
#
########################################################################

# plus -- Create new [+ Foo] node.

sub plus {
    my $class = shift;
    my $MCT = shift;
    unless (ref ($MCT) and $MCT->isa ('MCatGenus')) {
	carp "Found ", ref (\$MCT), " while looking for a MCatGenus: $!";
	confess;
	die;
    }

    my $this = {};
    $this->{genus} = $MCT;
    $this->{genus_name} = $MCT->get_name;
    $this->{constituent_genus} = $MCT->get_constituent;
    
    $this->{subfeatures} = ();
    $this->{constituents} = [];
    bless ($this, $class);
    for (@_) {
	unless ($this->put ($_)) {
	    carp "Can't put $this ";
	    if (ref ($_) and $_->can ('get_genus_name')) {
		carp " of genus ", $_->get_genus_name; 
	    }
	    carp " under $this, of genus ", $this->get_genus_name, ":$!";
	    confess;
	    die;
	}
    }
    $this;
}

sub clone {
    my $this = shift;
    my $copy = plus MCat (
	$this->get_genus
	);
    my @subfeatures = $this->get_all_immediate_subfeatures;
    for (@subfeatures) {
	$copy->put ($_->clone);
    }
    $copy->set_constituent ($this->get_unique_constituent);
    $copy;
}

sub pseudo_clone {
    # Make a ``shallow copy'' of the Unary Feature $this.
    # Differs from clone in that the copy has *the same* dependents, 
    # rather than clones of those dependents.  Faster than cloning, and
    # safe as long as neither the original nor the copy is modified.

    my $this = shift;
    my $copy = plus MCat (
	$this->get_genus
	);

#    my @subfeatures = $this->get_all_immediate_subfeatures;
#    for (@subfeatures) {
#	$copy->put ($_);
#    }
#    Maybe this is faster:
#    $copy->{subfeatures} = {};
#    for (keys (%{$this->{subfeatures}})) {
#	$copy->{subfeatures}->{$_} = $this->{subfeatures}->{$_};
#    }
    %{$copy->{subfeatures}} = %{$this->{subfeatures}};
    @{$copy->{constituents}} = @{$this->{constituents}};
    $copy;
}

sub mutant {
    my $this = shift;
    my $mutagen = (shift or '(no argument)');

    unless (ref ($mutagen) and $mutagen->isa ('Mutagen')) {
        carp "Argument to mutant must be of type Mutagen, not $mutagen:$!";
        confess;
	die;
    }

    return $this->clone unless ($mutagen->mutate);

    # Mutate the MCat and its dependents
    my $copy = plus MCat ($this->get_genus);
    for (@{$this->get_genus->get_dependent_features_ref}) {

	# Consider each *potential* subfeature of $this.  Does $this 
	# actually have such a subfeature?
	my $subfeature = $this->get_immediate_subfeature ($_->get_name);
	my $subfeature_ref = ref ($subfeature);

	if ($subfeature) {
	    # Yes.  Copy it (mutated) unless we fail saving throw.

	    if ($subfeature_ref eq 'BinaryFeature' or 
		$subfeature_ref eq 'BinaryFeatureVariable') {
		unless ($mutagen->lose_binary) {
		    $copy->put ($subfeature->mutant ($mutagen));
		}
		next;
	    }

	    if ($subfeature_ref eq 'UnaryFeature' or
		$subfeature_ref eq 'UnaryFeatureVariable') {
		if ($subfeature_ref eq 'UnaryFeature' and
		    $mutagen->feature_parents_are_protected and 
		    $subfeature->get_all_immediate_subfeatures) {
		    $copy->put ($subfeature->mutant ($mutagen));
		} else {
		    unless ($mutagen->lose_unary) {
			$copy->put ($subfeature->mutant ($mutagen))
		    }
		} 
		next;
	    }
	} else {
	    # No.  Add a new one only if we fail saving throw.
	    my $potential_dependent_ref = ref ($_);

	    if ($potential_dependent_ref eq 'BinaryFeatureGenus') {
		if ($mutagen->gain_binary) {
		    if (rand () < 0.5) {
			$copy->put ($_->plus->mutant ($mutagen));
		    } else {
			$copy->put ($_->minus->mutant ($mutagen));
		    }
		}
		next;
	    }
	    if ($potential_dependent_ref eq 'UnaryFeatureGenus') {
		if ($mutagen->gain_unary) {
		    $copy->put ($_->plus->mutant ($mutagen));
		}
		next;
	    }
	    carp "$_ should be Unary or BinaryFeatureGenus: $!";
	}
    }

    # Mutate the constituent
    $copy->set_constituent ($this->get_unique_constituent->mutant ($mutagen));

    $copy;
}

sub breed {
    my (
	$this, $that,        # Parents
	$recombinator
	) = @_;

    if ((ref $that) and ($that->isa ("Recombinator"))) {
        $recombinator = $that; $that = undef
    }
    unless (defined $recombinator) {
        confess "No Recombinator in call to breed (): $!";
    }

    my $kid;

  SWITCH: {
      if (!defined ($that)) {
          $kid = (!$recombinator->null_wins) ? $this->clone : undef;
          last SWITCH;
      }

      unless ($this->get_genus_name eq $that->get_genus_name) {
          carp "Genera don't match: " . $this->get_genus_name . " != " .
              $that->get_genus_name . ":$!";
          confess;
          die;
      }

      if ($that->isa ("MCat")) {
	  $kid = ($recombinator->this_wins) ? $this->get_genus->plus :
	      $that->get_genus->plus;

	  # this and that are of same genus, so (named) dependents are
	  # drawn from same set
	  my @potential_dependents = $this->get_genus->get_dependent_features;
	  for (@potential_dependents) {
	      my $name = $_->get_name;
	      my $this_subfeature = $this->get_immediate_subfeature ($name);
	      my $that_subfeature = $that->get_immediate_subfeature ($name);

	      # If both this and that have the subfeature, breed them.

	      if ($this_subfeature and $that_subfeature) {
		  $kid->put (
		      $this_subfeature->breed ($that_subfeature, 
					       $recombinator));
		  next;
	      }

	      # If only one has the subfeature, kid may or may not get it.
	      
	      if ($this_subfeature) {
		  if (!$recombinator->null_wins) {
		      $kid->put ($this_subfeature->clone);
		  }
		  next;
	      }

	      if ($that_subfeature) {
		  if (!$recombinator->null_wins) {
		      $kid->put ($that_subfeature->clone);
		  }
		  next;
	      }

	      # The subfeature is found in neither this nor that, so
	      # the kid doesn't get it.
	  }
	  
	  # Breed the constituents (each parent has at most one):
	  my $this_constituent = $this->get_unique_constituent;
	  my $that_constituent = $that->get_unique_constituent;
	  if ($this_constituent) {
	      if ($that_constituent) {
		  $kid->set_constituent ($this_constituent->breed (
				     $that_constituent, $recombinator));
	      } else {
		  if (!$recombinator->null_wins) {
		      $kid->set_constituent ($this_constituent);
		  }
	      }
	  } elsif ($that_constituent) {
	      if (!$recombinator->null_wins) {
		  $kid->set_constituent ($that_constituent);
	      }
	  }

	  last SWITCH;
      }

      confess "Can't breed a MCat with a ". 
	  ref ($that) . ": $!";
    }

    return $kid;
}


########################################################################
#
# Access methods for dependents
#
########################################################################

# *** the subtree_size methods are from the parent class UnaryFeature,
# and so they don't count what's in the PrCat constituent.  Does it 
# matter?

sub set_constituent {
    # Set the PrCat constitutent
    my $this = shift;
    my $new_constituent = shift;

    unless (ref $this and 
	    $new_constituent->get_genus_name eq 
	    $this->get_constituent_genus->get_name) {
	carp "Can't add $new_constituent under $this: $!";
	confess;
    }

    @{$this->{constituents}} = ($new_constituent);
}

sub get_all_constituents {
    # There should only ever be one
    my $this = shift;
    
    my @constituents = @{$this->{constituents}};
    if (@constituents > 1) {
	carp "MCat $this should have at most one PrCat constituent: $!";
	confess;
    }
    @constituents;
}

sub get_unique_constituent {
    my $this = shift;
    
    my @constituents = $this->get_all_constituents;
    pop @constituents;
}

sub get_constituent_genus {
    my $this = shift;

    $this->{constituent_genus};
}

sub put {
    my ($this, $new) = @_;

    unless (ref $new) {
        carp "Can't put $new under $this: $!";
        confess;
        die;
    }
    
    my $new_ref = ref ($new);
    my $success;
    if ($new_ref eq 'BinaryFeature' or $new_ref eq 'UnaryFeature' or 
        $new_ref eq 'PrCat' or
        $new_ref eq 'BinaryFeatureVariable' or 
        $new_ref eq 'UnaryFeatureVariable' or
        $new_ref eq 'PrCatVariable' or
        $new->isa ('Variable')
        ) {
        if ($new->get_genus_name eq $this->get_constituent_genus->get_name) {
	    $this->set_constituent ($new);
	    $success = $new;
	} else {
	    unless ($success = &UnaryFeature::put ($this, $new)
		    or
		    $success = $this->get_unique_constituent->put (
			$this, $new)
		) {
		carp "Can't add $new as constituent of $this: $!";
		confess;
	    }
	}
    }
    $success;
}

sub set_head {
    my $this = shift;
    # Reach down into the PrWd and set the head
    $this->get_unique_constituent->set_head;
}


########################################################################
#
# Matching methods
#
########################################################################

# A MCat $this matches in a MCat $that iff $this and 
# $that are of the same type (genus, e.g., both [Place]), and every
# immediate dependent of $this matches in its namesake dependent of
# $that.

sub matches_in {
    my (
	$this, 
	$that,
	$env
	) = @_;
    defined ($env) or ($env = new Environment);

    unless (ref ($that) and $that->can ('matches_in')) {
	carp "Match target must be a matchable object, not $that: $!";
        confess;
        die;
    }
    unless (ref ($env) and $env->isa ('Environment')) {
        carp "Second argument to matches_in must be Environment, not $that: $!";
        confess;
        die;
    }

    if ($that->isa ('MCat')) {
	return unless ($this->get_genus_name eq $that->get_genus_name);

	# Check whether the constituents match.
	return unless ($this->get_unique_constituent->matches_in (
			   $that->get_unique_constituent, $env));
	
	# Recursively check whether the subfeatures match:

	# Break $this into any subfeature vs. the rump.
	my @this_subfeatures = $this->get_all_immediate_subfeatures;

	# Basis case: No subfeatures; match succeeds.
	return (my @foo = ($env)) unless @this_subfeatures;

	# Quick check:  Are there enough subfeatures to match?
	return unless (@this_subfeatures <= $that->get_all_immediate_subfeatures);

	# Quick check:  Do all the this_subfeature names occur in the that_subfeatures?
	my @that_subfeatures = $that->get_all_immediate_subfeatures;
	my @that_subfeature_names = map {
	    $_->get_genus_name
	} @that_subfeatures;
	my $matched_all = 1;
	for my $this_subfeature (@this_subfeatures) {
	    my $matched_this = 0;
	    my $this_genus_name = $this_subfeature->get_genus_name;
	    for my $that_subfeature (@that_subfeatures) {
		my $that_genus_name = $that_subfeature->get_genus_name;
		if ($this_genus_name eq $that_genus_name) {
		    $matched_this = 1;
		    last;
		}
	    }
	    unless ($matched_this) {
		$matched_all = 0;
		last;
	    }
	}
	return unless $matched_all;

	# Otherwise:  Make a clone, diminished by removal of first subfeature.
	my $this_0 = shift @this_subfeatures;
	my $this_prime = $this->pseudo_clone;
	$this_prime -> delete_immediate_subfeature ($this_0->get_genus_name);
	
	# Do the same to $that, breaking off the same subfeature if psbl:
	my $that_0 = $that->get_immediate_subfeature ($this_0->get_genus_name);
	return unless $that_0;
	my $that_prime = $that->pseudo_clone;
	$that_prime->delete_immediate_subfeature ($that_0->get_genus_name);

	# Get refinements of $env that allow $this_0 and $that_0 to match:
	my @initial_successes = $this_0->matches_in ($that_0, $env);
	return unless @initial_successes;
	
	# Now use those environments to match the rump $this_prime and
	# $that_prime.

	my @successes = ();
	for (@initial_successes) {
	    unless (ref ($_) and ($_->isa ('Environment'))) {
		carp "Expected environment, found $_ of type ",
		ref ($_), " instead: $!";
		confess;
		die;
	    }
	    push @successes, $this_prime->matches_in ($that_prime, 
						      $_->clone)
	}

	return @successes;
    } else {
	return ();
    # A MCat matches in nothing but another MCat, not 
    # even in a MCatVariable.
    }
}

	
########################################################################
#
# I/O methods
#
########################################################################

sub to_flat_string {
    my (
	$this,
	$is_head
	) = (@_, 0);

    my $flat_string = $this->get_sign . $this->get_genus_abbrev;
    if ($is_head) {
	$flat_string = "[*$flat_string*";
    } else {
	$flat_string = "[$flat_string";
    }
    for (sort {
	$a->get_genus_name cmp $b->get_genus_name
	 }($this->get_all_immediate_subfeatures)) {
	$flat_string .=  $_->to_flat_string ();
    }
    $flat_string .= $this->get_unique_constituent->to_flat_string;
    $flat_string .= ']';
    $flat_string;
}

sub to_text {
    my (
        $this,
        $depth,       # Depth of nesting
        $is_head    # Head of parent?
        ) = (@_, 0, 0);

    my @text;
    if ($is_head) {
        push @text, ($Globals::indent x $depth) .
            '*' . 
            $this-> get_sign . 
            $this-> get_genus_name .
            '*'
    } else {
        push @text, ($Globals::indent x $depth) .
            '[' . 
            $this-> get_sign .
            $this-> get_genus_name .
            ']';
    }

    for (sort {
	$a->get_genus_name cmp $b->get_genus_name
	 }($this->get_all_immediate_subfeatures)) {
	push @text, $_->to_text ($depth + 1);
    }

    my $constituent = $this->get_unique_constituent;
    if ($constituent) {
	push @text, $constituent->to_text ($depth+1);
    }

    @text;
}

sub from_text {
    # Pass text in as ref to array, or array itself
    my $text = (ref $_[0] eq 'ARRAY') ? $_[0] : \@_;

    # Return empty-handed unless the text begins by describing an MCat
    unless (
        $text->[0] =~ /^((?:$Globals::indent_re)*)
                    \[\+([\w\s]+\w)\]\s*$/x) {
        return;
    }
    my ($indent_string, 
        $genus_name) = ($1, $2);
    
    my $depth = length ($indent_string)/length ($Globals::indent);
    my $genus = &Globals::get_representations->get ($genus_name);

    unless (ref ($genus) and $genus->isa ('MCatGenus')) {
	warn "$genus is not an MCatGenus: $!";
        return;
    }
    
    my $new = $genus->plus;
    shift @{$text};
    
    while (@{$text}) {

	# Have we run out of dependents?
	my $next_line = $text->[0];
	$next_line =~ /^((?:$Globals::indent_re)*)/;
	my $next_depth = length ($1)/length ($Globals::indent);
	last if ($next_depth <= $depth);
	
	my ($child, $child_depth, $child_is_head) =
	    &UnaryFeature::from_text ($text);
	unless ($child) {
	    ($child, $child_depth, $child_is_head) =
		&BinaryFeature::from_text ($text);
	} 
	unless ($child) {
	    ($child, $child_depth, $child_is_head) =
		&BinaryFeatureVariable::from_text ($text);
	} 
	unless ($child) {
	    ($child, $child_depth, $child_is_head) =
		&PrCat::from_text ($text);
	}
	unless ($child) {
	    carp "Couldn't parse $next_line: $!";
	    confess;
	}
	unless ($new->put ($child)) {
	    carp "Error in $next_line: $!";
	    confess;
	}
    }
    ($new, $depth);
}


sub print {
    my (
        $this,
        $depth,     # Depth of nesting
	$is_tree
        ) = (@_, 0);
    print (join "\n", $this->to_text ($depth));
    print "\n";
    
}

sub to_tree {
    my $this = shift;
    my $width = (shift or '1in');

    my $indent_level = (shift or 0);
    my $indent_quantum = 2;
    my $indent = ' ' x ($indent_quantum * (1+$indent_level));

    my $leader_level = (shift or 0);
    my $leader_quantum = 1;
    my $leader = '.' x ($leader_quantum * $leader_level);

    my $style = (shift or '\ttfamily');

    my $string = "{$style" . "{$leader}" . '{[}' . $this->get_genus_abbrev . 
	'{]}}\\\\';;
    $string = $indent . $string . "\n";

    # First the dependent features
    for ($this->get_genus->get_dependent_feature_names) {
	my $subfeature = $this->get_immediate_subfeature ($_);
	next unless $subfeature;
	$string .= $subfeature->to_tree ($width, $indent_level + 1, 
	    $leader_level + 1, $style);
    }
    # Then the (unique) PrCat constituent
    if ($this->get_constituent) {
	$string .= $this->get_constituent->to_tree ($width, $indent_level+1,
						    $leader_level+1, $style)
    }

    $string;
}


1;
