package MCatGenus;
use MCat;
use UnaryFeatureGenus;
use strict;
use warnings;
use Carp;

our @ISA = qw ( UnaryFeatureGenus );

########################################################################
#
# MCatGenus.pm
#
# Used to define morphosyntactic features.  An MCatGenus is where you
# store the information that, e.g., a word has number and gender.  An
# MCatGenus specifies any number of unordered Feature dependents, plus
# a single PrCatGenus.
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2019 January 26 ($)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


########################################################################
#
# Constructors
#
########################################################################

sub new {
    my $class = shift;
    my (
	$name,    # Name of MCat genus ("Word", etc.)
	$abbrev   # Short name of MCat genus
	) = @_;

    my $this = new UnaryFeatureGenus ($name, $abbrev);

    bless ($this, $class);
}

########################################################################
#
# Access methods
#
########################################################################

sub add_dependent {
    my $this = shift;
    my $dependent = shift;

    return unless (ref $dependent);
    if ($dependent->isa ('PrCatGenus')) {
	$this->set_constituent ($dependent);
    } else {
	if ($dependent->isa ('UnaryFeatureGenus') or
	    $dependent->isa ('BinaryFeatureGenus')) {
	    $this->add_dependent_features ($dependent);
	} else {
	    carp "Can't add dependent $dependent under $this: $!";
	    confess;
	}
    }
}

sub set_constituent {
    my $this = shift;
    my $constituent = shift;
    unless (ref ($constituent) and
	    ($constituent->isa ('PrCatGenus')
	    )) { 
	carp "Constituent type ($constituent) must be PrCatGenus: $!";
	confess;
	die;
    }

    $constituent->set_parent ($this);
    $this->{constituent} = $constituent;
}

sub get_constituent {
    my $this = shift;
    $this->{constituent};
}

########################################################################
#
#  Methods for creating a new MCat node of this genus
#
########################################################################

sub plus {
    my $this = shift;
    
    plus MCat ($this);
}


########################################################################
#
# I/O methods
#
########################################################################

sub print {
    my $this = shift;
    my $indent = (shift or "");

    print $indent;
    printf "%s (%s) ", $this->get_name, $this->get_abbrev;
    if ($this->get_parent) {
	print ", dependent of ", $this->get_parent->get_name;
    }
    print "\n";
    for ($this->get_dependent_features) {
        $_->print ("$indent  ");
    }
    print "$indent "; 
    if ($this->get_constituent) {
	$this->get_constituent->print ("$indent  ");
    }
}

1;
