package Mutagen;
use strict;
use warnings;
use Carp;
use Globals;

########################################################################
#
# Mutagen.pm
#
# Does the die-rolling for the mutant () methods; is passed to them as
# an argument.  This is so that we can have different Mutagens, and hence
# different probabilities of different kinds of mutations, for constraints
# and representations.
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


sub new {
    my $class = shift;
    my $this = {};
    bless $this, $class;
    my %Ps = @_;
    for (keys %Ps) {
	$this->set_probability ($_, $Ps {$_});
    }
    
    $this->protect_PrCat_parents;
    $this->protect_feature_parents;

    srand;
    $this;
}

sub set_name {
    my (
	$this,
	$name
	) = @_;

    unless ($name =~ /./) {
	confess "Name must be non-null: $!";
    }
    
    $this->{name} = $name;
}

sub get_name {
    my $this = shift;

    $this->{name};
}

sub set_probability {
    my $this = shift;
    my ($name, $p) = @_;
    $p or ($p = 0);
    if ($p < 0 or $p > 1) {
	carp "Tried to set P($this) to invalid probability $p:$!";
	confess;
	die;
    }
    $this->{contents}->{$name} = $p;
}

sub protect_PrCat_parents {
    # Forbid deletion, or conversion to a variable, of UnaryFeatures
    # which have dependents.
    my $this = shift;
    $this->{contents}->{"protect_PrCat_parents"} = 1;
}

sub unprotect_PrCat_parents {
    my $this = shift;
    $this->{contents}->{"protect_PrCat_parents"} = 0;
}

sub protect_feature_parents {
    # Forbid deletion, or conversion to a variable, of UnaryFeatures
    # which have dependents.
    my $this = shift;
    $this->{contents}->{"protect_feature_parents"} = 1;
}

sub unprotect_feature_parents {
    my $this = shift;
    $this->{contents}->{"protect_feature_parents"} = 0;
}


sub get_P {
    my ($this, $which) = @_;
    my $P = $this->{contents}->{"P_$which"};
    unless (defined $P) {
	confess "P_$which is undefined";
    }
    $P;
}    

sub do_we {
    my ($this, $which) = @_;
    my $choice = (rand () < $this->get_P ($which));
    return $choice;
}

# Do we mutate at all?
sub get_P_mutate {
    my $this = shift;
    $this->get_P ("mutate");
}
sub mutate {
    my $this = shift;
    $this->do_we ("mutate");
}

########################################################################
#
# Saving throws for mutating PrCats
#
########################################################################

sub add_constituent {
    my $this = shift;
    $this->do_we ("add_constituent");
}

sub n_copies {
    my $this = shift;
    # Returns how many copies to make of a constituent: 0, 1, or 2.
    if (!$this->do_we ("delete_constituent")) {
	if ($this->do_we ("duplicate_constituent")) {
	    2;
	} else {
	    1;
	} 
    } else {
	0;
    }
}

sub gain_head {
    my $this = shift;
    $this->do_we ("gain_head");
}

sub lose_head {
    my $this = shift;
    $this->do_we ("lose_head");
}

sub head_no_head {
    my $this = shift;
    $this->do_we ("head_no_head");
}

sub move_head {
    my $this = shift;
    # Returns distance to move the head: -1, 0, or +1.
    if ($this->do_we ("move_head")) {
	return ((rand () < 0.5) ? -1 : 1);
    } else {
	return;
    }
}

sub flip_anchor {
    my $this = shift;
    $this->do_we ("flip_anchor");
}

########################################################################
#
# Saving throws for mutating Unary and BinaryFeatures and their Variables.
#
########################################################################

sub invert_polarity {
    my $this = shift;
    $this->do_we ("invert_polarity");
}

sub gain_unary {
    my $this = shift;
    $this->do_we ("gain_unary");
}

sub lose_unary {
    my $this = shift;
    $this->do_we ("lose_unary");
}

sub gain_binary {
    my $this = shift;
    $this->do_we ("gain_binary");
}

sub lose_binary {
    my $this = shift;
    $this->do_we ("lose_binary");
}

sub invert_unary_coefficient {
    # Only applies to UnaryFeatureVariables like -alpha Place
    my $this = shift;
    $this->do_we ("invert_unary_coefficient");    
}

sub invert_binary_coefficient {
    my $this = shift;
    $this->do_we ("invert_binary_coefficient");    
}

sub invert_PrCat_coefficient {
    # Only applies to PrCatVariables like -alpha Syllable
    my $this = shift;
    $this->do_we ("invert_PrCat_coefficient");
}

sub unary_feature_to_variable {
    my $this = shift;
    $this->do_we ("unary_feature_to_variable");
}

sub unary_variable_to_feature {
    my $this = shift;
    $this->do_we ("unary_variable_to_feature");
}

sub binary_feature_to_variable {
    my $this = shift;
    $this->do_we ("binary_feature_to_variable");
}

sub binary_variable_to_feature {
    my $this = shift;
    $this->do_we ("binary_variable_to_feature");
}

sub PrCat_to_PrCat_variable {
    my $this = shift;
    $this->do_we ("PrCat_to_PrCat_variable");
}

sub PrCat_variable_to_PrCat {
    my $this = shift;
    $this->do_we ("PrCat_variable_to_PrCat");
}

sub PrCat_parents_are_protected {
    my $this = shift;
    return ($this->{contents}->{"protect_PrCat_parents"} or 0);
}

sub feature_parents_are_protected {
    my $this = shift;
    return ($this->{contents}->{"protect_feature_parents"} or 0);
}

########################################################################
#
# I/O routines
#
########################################################################

sub to_text {
    my (
	$this,
	$depth
	) = (@_, 0);
    
    my $indent_string = $Globals::indent x $depth;

    my @text;
    my $name = $this->get_name;
    push @text, $indent_string . "Mutagen $name";
    for my $pname (sort (keys %{$this->{contents}})) {
	my $pval = $this->{contents}->{$pname};
	push @text, "$indent_string$Globals::indent$pname\t$pval";
    }
    @text;
}

sub from_text {
    my $text = (ref $_[0] eq 'ARRAY') ? $_[0] : \@_;
    
    my $indent_re = $Globals::indent_re;

    unless (defined $text->[0]) {
	return;
    }

    unless ($text->[0] =~ /^(($indent_re)*)Mutagen\s+([\w.:\-+]+)/) {
	return;
    }
    my $indent_string = $1;
    my $name = $3;

    shift @{$text};

    my $mutagen = new Mutagen;
    $mutagen->set_name ($name);
    while ($text->[0] =~ /^$indent_string$indent_re(\w+)\s+(\d*\.?\d+)/) {
	my ($pname, $pval) = ($1, $2);
	
	unless ($pname =~ /./) {
	    confess "Ill-formed mutation probability name $pname: $!";
	}
	unless ($pval =~ /./) {
	    confess "Ill-formed probability $pname = $pval: $!";
	}
	$mutagen->set_probability ($pname, $pval);
	shift (@{$text});
    }

    $mutagen;
}



sub print {
    my $this = shift;
    for (keys %{$this->{contents}}) {
	printf "%30s %4.3f\n", $_, $this->{contents}->{$_};
    }
    if ($this->PrCat_parents_are_protected ()) {
	print "PrCat parents are protected\n";
    } else {
	print "PrCat arents are not protected\n";
    }
    if ($this->feature_parents_are_protected ()) {
	print "Feature parents are protected\n";
    } else {
	print "Feature parents are not protected\n";
    }
}

sub to_latex {
    my $this = shift;
    my $latex = "
\\begin{center}
  \\begin{tabular}{ll}
";
    for (keys %{$this->{contents}}) {
	$latex .= sprintf "%s & %4.3f\\\\\n", $_, $this->{contents}->{$_};
    }

    $latex .= "
  \\end{tabular}

";
    if ($this->PrCat_parents_are_protected ()) {
	$latex .= "PrCat parents are protected\n\n";
    } else {
	$latex .= "PrCat arents are not protected\n\n";
    }
    if ($this->feature_parents_are_protected ()) {
	$latex .= "Feature parents are protected\n\n";
    } else {
	$latex .= "Feature parents are not protected\n\n";
    }
    $latex .= "
\\end{center}
";
    $latex =~ s/_/ /g;

    $latex;
}

1;
