package MutagenSet;
use Globals;
use Mutagen;
use warnings;
use strict;
use Carp;

########################################################################
#
#  MutagenSet.pm
#
#  A set of named Mutagens, which can be passed to a Learner, so that
#  the Learner can use different Mutagens on different trials.
#
#  E. Moreton  *  UNC-Chapel Hill Linguistics  *  2019 March 8 (F)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


sub new {
    my $class =shift;

    my $this = {};
    bless $this, $class;

    $this;
}

########################################################################
#
#  Access methods
#
########################################################################

sub add {
    my (
	$this,
	$mutagen
	) = @_;

    unless (ref $mutagen and ($mutagen->isa ('Mutagen'))) {
	confess "First argument to add must be Mutagen, not $mutagen: $!";
    }
    my $name = $mutagen->get_name;
    unless ($name =~ /./) {
	confess "Non-null name required for Mutagen: $!";
    }
    if ($this->{contents}->{$name}) {
	confess "Mutagen $name already exists: $!";
    }

    $this->{contents}->{$name} = $mutagen;
}

sub get_mutagen_named {
    my (
	$this,
	$name
	) = @_;

    unless (defined $name) {
	confess "Mutagen name must be non-null: $!";
    }

    my $mutagen = $this->{contents}->{$name};
    unless (defined $mutagen) {
	confess "Unknown mutagen $name: $!";
    }
    $mutagen;
}

sub get_names {
    my $this = shift;

    keys %{$this->{contents}};
}

sub get_all {
    my $this = shift;

    values %{$this->{contents}}
}

########################################################################
#
#  I/O methods
#
########################################################################

sub to_text {
    my (
	$this, 
	$depth
	) = (@_, 0);

    my $indent_string = $Globals::indent x $depth;
    my @text;
    push @text, $indent_string . "MutagenSet";
    for my $mutagen ($this->get_all){
	push @text, $mutagen->to_text ($depth + 1);
    }
    @text;
}

sub from_text {
    my $text = (ref $_[0] eq 'ARRAY') ? $_[0] : \@_;
    
    my $indent_re = $Globals::indent_re;
    
    unless ($text->[0] =~ /^(($indent_re)*)MutagenSet/) {
	return;
    }
    my $indent_string = $1;
    my $name = $3;
    shift @{$text};

    my $ms = new MutagenSet;
    while (my $mutagen = &Mutagen::from_text ($text)) {
	last unless (ref $mutagen and $mutagen->isa ("Mutagen"));
	$ms->add ($mutagen);
    }

    $ms;
}

    
1;

