package PrCat;
use Globals;
use PrCatVariable;
use Environment;
use warnings;
use strict; 
use Carp;
use List::Util;

########################################################################
#
# PrCat.pm
#
# Object implementing a prosodic node.  
#
# Has left and right boundary anchors, either of which may be set or not,
# and an unlimited number of anonymous, ordered dependents.  
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2009 March 8 ($)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


########################################################################
#
# Constructors
#
########################################################################

sub new {
    my $class = shift;
    my $PCT = shift;
    unless (ref ($PCT) and $PCT->isa ('PrCatGenus')) {
	carp "Found ", ref (\$PCT), " while looking for a PrCatGenus: $!";
	confess;
	die;
    }

    my $this = {};

    $this->{genus} = $PCT;
    $this->{genus_name} = $PCT->get_name;
    $this->{constituent} = $PCT->get_constituent;
    $this->{constituents} = ();

    bless $this, $class;

    $this->set_both_anchors;

    $this;
}

sub clone {
    my $this = shift;
    my $copy = new PrCat ($this->get_genus);
    $copy->set_both_anchors;
    $copy->unset_left_anchor unless $this->get_left_anchor;
    $copy->unset_right_anchor unless $this->get_right_anchor;

    for ($this->get_all_constituents) {
	if ($_ eq $this->get_head_constituent) {
	    $copy->put_head ($_->clone);
	} else {
	    $copy->put ($_->clone);
	}
    }
    $copy;
}

sub pseudo_clone {
    # Make shallow copy of $this.
    # Makes a new PrCat that has *the same dependents* as this one,
    # rather than *clones* of these dependents.  Faster than real
    # cloning, and perfectly safe as long as you're sure that no
    # changes will be made to the original or the clone.  Used to
    # speed up matching.

    my $this = shift;
    my $copy = new PrCat ($this->get_genus);
    $copy->set_both_anchors;
    $copy->unset_left_anchor unless $this->get_left_anchor;
    $copy->unset_right_anchor unless $this->get_right_anchor;

    for ($this->get_all_constituents) {
	if ($_ eq $this->get_head_constituent) {
	    $copy->put_head ($_);
	} else {
	    $copy->put ($_);
	}
    }
    $copy;
}

sub mutant {
    # Create a mutated version.  There are lots of ways to change a 
    # PrCat.

    my $this = shift;
    my $mutagen = (shift or '(no argument)');
    unless (ref ($mutagen) and $mutagen->isa ('Mutagen')) {
        carp "Argument to mutant must be of type Mutagen, not $mutagen:$!";
        confess;
        die;
    }
	
    return ($this->clone) unless ($mutagen->mutate);

    # Replace this PrCat with a PrCatVariable?
    unless ($mutagen->PrCat_parents_are_protected and
            $this->has_dependents) {
        if ($mutagen->PrCat_to_PrCat_variable) {
            return plus PrCatVariable ($this->get_genus, 'alpha');
        }
    }

    my $copy = new PrCat ($this->get_genus);

    # Set the mutant's anchors to match $this's; then perhaps flip them.
    $copy->set_both_anchors;
    $copy->unset_left_anchor unless $this->get_left_anchor;
    $copy->unset_right_anchor unless $this->get_right_anchor;
    if ($mutagen->flip_anchor) {
	if ($copy->get_left_anchor) {
	    $copy->unset_left_anchor;
	} else {
	    $copy->set_left_anchor;
	}
    }
    if ($mutagen->flip_anchor) {
	if ($copy->get_right_anchor) {
	    $copy->unset_right_anchor;
	} else {
	    $copy->set_right_anchor;
	}
    }

    # The next several paragraphs of code are concerned with copying the
    # constituents.  We copy them from left to right, perhaps messing
    # with them.  Since some of the messing is L-R asymmetric, we have to
    # half the time reverse the order of the constituents before messing,
    # then restore it when we're through.

    my $LtoR = (rand () > 0.5);
    my @old_constituents = $this->get_all_constituents;
    unless ($LtoR) {
	@old_constituents = (reverse @old_constituents);
    }
    my $old_head = $this->get_head_constituent;

    # Decide whether or not to add a new constituent.
    if ($mutagen->add_constituent) {
	my $constituent_type = $this->contains_genus;
	my $constituent_type_ref = ref ($constituent_type);
	my $new_constituent;
#	if ($constituent_type->isa ('PrCatGenus')) {
	if ($constituent_type_ref eq 'PrCatGenus') {
	    $new_constituent = new PrCat (
		$this->contains_genus
		)->mutant ($mutagen);
	}
#	if ($constituent_type->isa ('UnaryFeatureGenus')) {
	if ($constituent_type_ref eq 'UnaryFeatureGenus') {
	    $new_constituent = $this->contains_genus->plus->mutant ($mutagen);
	}
	unless ($new_constituent) {
	    carp "Constituent type must be PrCatGenus or UnaryFeatureGenus, not",
	    $constituent_type->get_name, ": $!";
	    confess;
	    die;
	}
	
	# Decide where to add the new constitutent
	my $where = rand (@old_constituents);
	splice (@old_constituents, 
		$where, 
		0,
		$new_constituent
	    );
    }

    # Copy (or not) the constituents, mutating each one recursively:
    my @new_constituents;
    my $new_head = '';
    while (@old_constituents) {
	my $current = shift @old_constituents;
	my $n_copies = $mutagen->n_copies;

	# If parental protection is turned on, don't delete the node
	# if it has dependents.
	if ($mutagen->PrCat_parents_are_protected and
	    $current->has_dependents) {
	    $n_copies = 1 unless $n_copies;
	}

	for (1..$n_copies) {
	    push @new_constituents, $current->mutant ($mutagen);
	}

	# If the old head was copied twice, the last one copied becomes
	# the new head.  That's why we have to go L->R half of the time
	# and R->L the other half -- otherwise, heads would slowly
	# migrate rightwards.
	if ($current eq $old_head and @new_constituents) {
	    $new_head = $new_constituents [-1];
	}
    }
    
    # If there's no head, we may want to designate one; if there is one,
    # we may want to get rid of it.
    if ($new_head) {
	if ($mutagen->lose_head) {
	    $new_head = '';
	}
    } else {
	if ($mutagen->gain_head) {
	    $new_head = $new_constituents [rand $#new_constituents];
	}
    }

    # We could also designate a different head, but I'm leaving that for
    # later.

    # Make new PrCat
    unless ($LtoR) {
	@new_constituents = reverse (@new_constituents);
    }
    for (@new_constituents) {
	if ($_ eq $new_head) {
	    $copy->put_head ($_);
	} else {
	    $copy->put ($_);
	}
    }

    $copy;
}

sub breed {
    my (
	$this, $that,      # Parents
	$recombinator
	) = @_;

    if ((ref $that) and ($that->isa ("Recombinator"))) {
        $recombinator = $that; $that = undef
    }
    unless (defined $recombinator) {
        confess "No Recombinator in call to breed (): $!";
    }

    my $kid;

  SWITCH: {
      if (!defined ($that)) {
          $kid = (!$recombinator->null_wins) ? $this->clone : undef;
          last SWITCH;
      }

      unless ($this->get_genus_name eq $that->get_genus_name) {
          carp "Genera don't match: " . $this->get_genus_name . " != " .
              $that->get_genus_name . ":$!";
          confess;
      }

      if ($that->isa ('PrCat')) {
	  $kid = $this->get_genus->unanchored;
	  
	  if ($this->get_left_anchor) {
	      if ($that->get_left_anchor) {
		  $kid->set_left_anchor;
	      } else {
		  if (!$recombinator->null_wins) {
		      $kid->set_left_anchor;
		  }
	      }
          }
	  if ($this->get_right_anchor) {
	      if ($that->get_right_anchor) {
		  $kid->set_right_anchor;
	      } else {
		  if (!$recombinator->null_wins) {
		      $kid->set_right_anchor;
		  }
	      }
          }

	  
	  # Align the constituents.  The ones that get lined up with 
	  # nothing get either inherited whole or deleted; those that
	  # get lined up with other constituents get bred with them.

	  my @this_deps = $this->get_all_constituents;
	  my @that_deps = $that->get_all_constituents;

	  # Make aligned pairs of constituents and '(none)'
	  my ($this_mate, $that_mate) = &align (0+@this_deps, 0+@that_deps);

	  for my $i (0..(@this_deps + @that_deps -1)) {
	      my $this_dep = ($this_mate->[$i] eq '(none)') ?
		  '(none)' : $this_deps [$this_mate->[$i]];
	      my $that_dep = ($that_mate->[$i] eq '(none)') ?
		  '(none)' : $that_deps [$that_mate->[$i]];

	      next if ($this_dep eq '(none)' 
		       and $that_dep eq '(none)');

	      my ($this_head, $that_head);  # keep track of parental heads
	      $this_head = $this->is_head_constituent ($this_dep);
	      $that_head =  $that->is_head_constituent ($that_dep);

	      my $new_dep;	      
	      if ($this_dep eq '(none)') {
		  if ($recombinator->null_wins) {
		      next;
		  } else {
		      $new_dep = $that_dep->clone;
		  }
	      } else {
		  if ($that_dep eq '(none)') {
		      if ($recombinator->null_wins) {
			  next;
		      } else {
			  $new_dep = $this_dep->clone;
		      }
		  } else {
		      $new_dep = $this_dep->breed (
			  $that_dep, $recombinator);
		  }
	      }
	  
	      # Figure out about the head
	      if ($this_head) {
		  # new_dep is head in this
		  if ($that_head) {
		      # new_dep is head in both parents, so head in kid
		      $kid->put_head ($new_dep);
		  } else {
		      # new_dep head only in this
		      if ($this_dep eq '(none)') {
			  # No competition---this wins
			  $kid->put_head ($new_dep);
		      } else {
			  # head in this, non-head in that
			  if ($kid->get_head_constituent) {
			      # kid already has head; must be from that
			      if ($recombinator->this_wins) {
				  # this steals headship
				  $kid->put_head ($new_dep);
			      } else {
				  $kid->put ($new_dep);
			      }
			  } else {
			      $kid->put ($new_dep);
			  }
		      }
		  }
	      } else {
		  # new_dep isn't head in this
		  if ($that_head) {
		     # new_dep is head in that, but not this
		      if ($this_dep eq '(none)') {
			  # this is empty, so that wins by defaut
			  $kid->put_head ($new_dep);
		      } else {
			  # new_dep is head in that, but not this
			  if (!$recombinator->this_wins) {
			      # that steals headship
			      $kid->put_head ($new_dep);
			  } else {
			      $kid->put ($new_dep);
			  }
		      }
		  } else {
		      # new_dep is head in neither this nor that
		      $kid->put ($new_dep);
		  }
	      }

	  }

	  last SWITCH;
      }

      if ($that->isa ("PrCatVariable")) {
	  $kid = ($recombinator->this_wins) ? $this->clone : $that->clone;
	  last SWITCH;
      }

      confess "Can't breed a PrCat with a " . ref ($that) . ": $!";
    }

    return $kid;
}


sub align {
    # Given two sequences 0...N1-1 and 0...N2-1, randomly align them,
    # preserving order, so that some elements are paired with other
    # elements, and some elements are paired with "(none)".

    my ($N1, $N2) = @_;

    my @which1 = List::Util::shuffle (split '', (1 x $N1 . 0 x $N2));
    my @which2 = List::Util::shuffle (split '', (0 x $N1 . 1 x $N2));

    my ($i1, $i2);
    my (@mate1, @mate2);
    while (@which1 and @which2) {
	if ($which1[0]) {
	    push @mate1, $i1++;
	} else {
	    push @mate1, '(none)';
	}
	if ($which2[0]) {
	    push @mate2, $i2++;
	} else {
	    push @mate2, '(none)';
	}
	shift @which1; shift @which2;
    }
    return (\@mate1, \@mate2);
}

########################################################################
#
# Access methods for permanent properties
#
########################################################################

sub get_genus_name {
    my $this = shift;
#    $this->get_genus->get_name or '';
    $this->{genus_name};
}

sub get_genus_abbrev {
    my $this = shift;
    $this->get_genus->get_abbrev or '';
}

sub get_genus {
    my $this = shift;
    $this->{genus} or '';
}

sub contains_genus {
    my $this = shift;
    $this->{constituent};
}

########################################################################
#
# Access methods for L/R anchors
#
########################################################################

sub set_left_anchor {
    my $this = shift;
    $this->{left_anchor} = 1;
    $this;
}

sub unset_left_anchor {
    my $this = shift;
    delete $this->{left_anchor};
    $this;
}

sub get_left_anchor {
    my $this = shift;
    $this->{left_anchor};
}

sub set_right_anchor {
    my $this = shift;
    $this->{right_anchor} = 1;
    $this;
}

sub unset_right_anchor {
    my $this = shift;
    delete $this->{right_anchor};
    $this;
}

sub get_right_anchor {
    my $this = shift;
    $this->{right_anchor};
}

sub set_both_anchors {
    my $this = shift;
    $this->set_left_anchor;
    $this->set_right_anchor;
}

sub unset_anchors {
    my $this = shift;
    $this->unset_left_anchor;
    $this->unset_right_anchor;
}

########################################################################
#
# Access methods for constituents
#
########################################################################

# put, put_head -- Add a new node at the right edge.  If it's the wrong
#     type to be an immediate dependent, try to put it inside the last 
#     of the immediate dependents, and so on down recursively.  put_head
#     is similar, but when it succeeds, it makes the just-added node the
#     head in its parent.  
#
#     This only works because there is never any ambiguity about which
#     of the rightmost categories is the appropriate parent.  If we allow
#     (e.g.) a Segment to be a dependent of a Syllable *or* a Mora, then
#     put-ting a Segment becomes ambiguous:  Do you want it added to the
#     last Mora of the last Syllable, or directly to the last Syllable?
#
#     BUG:  There is no error if never finds where to put the new thing!.

sub put {
    my $this = shift;
    grep {
	$this->put_guts ($_);
    } @_;
}

sub put_head {
    my $this = shift;
    grep {
	$this->put_guts ($_, 1);
    } @_;
}

sub put_guts {
    my $this = shift;
    my $new = shift;
    my $make_head = (shift or 0);

    my $success;

    # Add $new at right edge, possibly making it the head.

    unless (ref $new) {
	carp "Can't put $new under $this: $!";
	confess;
	die;
    }

    my $new_ref = ref ($new);
    if ($new_ref eq 'BinaryFeature' or $new_ref eq 'UnaryFeature' or 
	$new_ref eq 'PrCat' or
	$new_ref eq 'BinaryFeatureVariable' or 
	$new_ref eq 'UnaryFeatureVariable' or
	$new_ref eq 'PrCatVariable' or
	$new->isa ('Variable')
	) {

#    if ($new->isa ('PrCat') or ($new->isa ('PrCatVariable')) or
#	$new->isa ('UnaryFeature') or $new->isa ('UnaryFeatureVariable') or
#	$new->isa ('BinaryFeature') or $new->isa ('BinaryFeatureVariable') or
#	$new->isa ('Variable')
#	) {

	if ($new->get_genus_name eq $this->contains_genus->get_name) {

	    # Add *as* last constituent if possible:
	    push @{$this->{constituents}}, $new;
	    if ($make_head) {
		$this->set_head;
	    }
	    $success = $new;
	} else {
	    # Failing that, try to add *in* rightmost constituent
	    my $last_constituent = $this->get_last_constituent;
	    if ($last_constituent) {
		if ($make_head and ref ($last_constituent) and
		    $last_constituent->can ('put_head')) {
		    $success = $last_constituent->put_head ($new);
		} else {
		    $success = $last_constituent->put ($new);
		}
	    } else {
		carp "Can't add $new as constituent of $this: $!";
		confess;
		die;
	    }
	}
    }

    $success;
}

sub set_head {
    my $this = shift;
    $this->{head} = $this->get_last_constituent;
    $this->{head};
}

sub unset_head {
    my $this = shift;
    $this->{head} = undef;
}

sub has_dependents {
    my $this = shift;
    if ($this->get_all_constituents) {
	1;
    }
}

sub get_depth {
    # How deep is the subtree rooted at this node?
    my $this = shift;
    my $max_dependent_depth = 0;
    for ($this->get_all_constituents) {
	my $dependent_depth = $_->get_depth;
	$max_dependent_depth = $dependent_depth 
	    if $max_dependent_depth < $dependent_depth;
    }
    $max_dependent_depth + 1;
}


##### INTERNAL USE ONLY: ######

sub shift_constituent {
    my $this = shift;
    my $shifted = shift (@{$this->{constituents}});
    return unless $shifted;
    if ($shifted eq $this->get_head_constituent) {
	delete ($this->{head});
    }
    $shifted;
}

sub pop_constituent {
    my $this = shift;
    my $popped = pop (@{$this->{constituents}});
    return unless $popped;
    if ($popped eq $this->get_head_constituent) {
	delete ($this->{head});
    }
    $popped;
}

sub get_all_constituents {
    my $this = shift;
    if ($this->{constituents}) {
	return @{$this->{constituents}};
    } else {
	return;
    }
}

sub get_head_constituent {
    my $this = shift;
    $this->{head} or '';
}

sub is_head_constituent {
    my $this = shift;
    my $constituent = shift;
    ($this->get_head_constituent) and 
	($constituent eq $this->get_head_constituent);
}

sub get_first_constituent {
    my $this = shift;
    ${$this->{constituents}}[0];
}

sub get_last_constituent {
    my $this = shift;
    ${$this->{constituents}}[-1];
}

########################################################################
#
# Matching methods
#
########################################################################

# A successful match of a PrCat $this in another PrCat $that occurs when 
# the two have the same type, AND every L/R anchor set in $this is also
# set in $that, AND each constituent of $this is matched in a constituent
# of $that, respecting left-to-right order and boundary anchors. For each
# successful match, matches_in returns an Environment with the variable
# assignments needed to make the match succeed.

# Uses a lot of cloning to guard against clobbering, but that's slow.
# We may be able to dispense with it in future.

sub matches_in {
    my (
	$this,  # Thing to match
	$that,  # Thing to match it in
	$env    # Environment with the current variable assignments
	) = @_;
    defined ($env) or ($env = new Environment);

    unless (ref ($that) and $that->can ('matches_in')) {
	carp "Match target must be a matchable object, not $that: $!";
	confess;
	die;
    }
    unless (ref ($env) and $env->isa ('Environment')) {
	carp "Second argument to matches_in must be Environment, not $that: $!";
	confess;
	die;
    }

    if ($that->isa ('PrCat')) {

	# Match fails if $this and $that are of different genus
	return unless ($this->get_genus_name eq $that->get_genus_name);

	# Match fails if $this has anchors which $that lacks.
	return if ($this->get_left_anchor and !$that->get_left_anchor);
	return if ($this->get_right_anchor and !$that->get_right_anchor);

	# OK, the proper parts of $this match in those of $that.  Now,
	# recursively match the constituents of $this in those of $that.

	# Break $this into zeroth constituent vs. rump PrCat.
	my $this_prime = $this->pseudo_clone;
	my $this_prime_head = $this_prime->get_head_constituent;
	my $this_0 = $this_prime->shift_constituent;

	# Now try to match the constituents of $this in a consecutive 
	# sequence of constituents of $that.  If $this is left-anchored,
	# we only try one way, viz, i-th constituent to i-th constituent.
	# If $this is right-anchored, we only try the reverse.  If 
	# $this is unanchored, we try every possible way.

	my $this_count = ($this->get_all_constituents or 0);
	return (my @there_must_be_an_easier_way_to_return_a_singleton_array = 
		($env)) unless $this_count;
	my $that_count = ($that->get_all_constituents or 0);
	return if ($this_count > $that_count);

	my $first_place = 0;
	my $last_place = $that_count - $this_count;
	$last_place  =                         0 if ($this->get_left_anchor);
	$first_place = $that_count - $this_count if ($this->get_right_anchor);

	my @successes = ();

	for my $i ($first_place..$last_place) {
	    # Create a diminished version of $that which has the same
	    # number of constituents as $a, starting at the i-th one.

	    my $that_prime = $that->pseudo_clone;
	    for my $j (0..($i-1)) {
		$that_prime->shift_constituent;
	    }
	    for my $k (0..($that_count - $this_count - $i -1)) {
		$that_prime->pop_constituent;
	    }

	    # Match $this_0 to $that_0 (the i-th ordered dependent of 
	    # $that) to get environments which allow that match.  If
	    # $this_0 is the head of $this, it can't match in $that_0
	    # unless $that_0 is the head of $that.

	    my $that_prime_head = $that_prime->get_head_constituent;
	    my $that_0 = $that_prime->shift_constituent;
	    next if ($this_prime_head eq $this_0 and 
		     $that_prime_head ne $that_0);

	    next if ($this_prime_head eq $this_0 and 
		     $that_prime_head ne $that_0);
	    my @initial_matches = $this_0->matches_in (
		$that_0, $env);

	    # Now, use those environments to match the remaining 
	    # constituents.
	    for (@initial_matches) {
		unless (ref ($_) and ($_->isa ('Environment'))) {
		    carp "Expected environment, found $_ of type ",
		    ref ($_), " instead, while trying to match ";
		    $this->print;
		    print "in ";
		    $that->print;
		    carp "At the time, initial_matches had ", 0+@initial_matches, " elements\n";
		    confess;
		    die;
		}
		push @successes, $this_prime->matches_in ($that_prime, 
		    $_->clone)
	    }
	}
	return @successes;
    }
    
    if ($that->isa ('PrCatVariable')) {
	return;
    }
}

sub equals {
    my ($this, $that, $env) = @_;
    $this->matches_in ($that, $env) and $that->matches_in ($this, $env);
}


########################################################################
#
# I/O methods
#
########################################################################

sub to_flat_string {
    my (
	$this,
	$is_head
	) = (@_, 0);
    my $flat_string;
    if ($is_head) {
	$flat_string .=
	    '[*' .
	    ($this->get_left_anchor ? "L" : "-") .
	    " " .
	    $this->get_genus_abbrev .
	    " " .
	    ($this->get_right_anchor ? "R" : "-") .
	    '*';
    } else {
	$flat_string .=
	    '[(' .
	    ($this->get_left_anchor ? "L" : "-") .
	    " " .
	    $this->get_genus_abbrev .
	    " " .
	    ($this->get_right_anchor ? "R" : "-") .
	    ')';	
    }

    for ($this->get_all_constituents) {
	if ($this->is_head_constituent ($_)) {
	    $flat_string .= $_->to_flat_string (1);
	} else {
	    $flat_string .= $_->to_flat_string (0);
	}
    }

    $flat_string .= ']';
    $flat_string;
}
    

sub to_text {
    my (
        $this,
        $depth,     # Depth of nesting
	$is_head    # Head of parent?
        ) = (@_, 0, 0);

    my @text;
    if ($is_head) {
	push @text,  ($Globals::indent x $depth) .
	    '*' .
	    ($this->get_left_anchor ? "L" : "-") .
	    " " .
	    $this->get_genus_name .
	    " " .
	    ($this->get_right_anchor ? "R" : "-") .
	    '*';
    } else {
	push @text,  ($Globals::indent x $depth) .
	    '(' .
	    ($this->get_left_anchor ? "L" : "-") .
	    " " .
	    $this->get_genus_name .
	    " " .
	    ($this->get_right_anchor ? "R" : "-") .
	    ')';	
    }

    for ($this->get_all_constituents) {
	if ($this->is_head_constituent ($_)) {
	    push @text, $_->to_text ($depth+1, 1);
	} else {
	    push @text, $_->to_text ($depth+1);
	}
    }

    @text;
}

sub from_text {
    # Pass text in as ref to array, or array itself
    my $text = (ref $_[0] eq 'ARRAY') ? $_[0] : \@_;

    # Return empty-handed unless the text begins by describing a BinFeat
    unless (
        $text->[0] =~ /^((?:$Globals::indent_re)*)
                    ([\(\*])    \s*
                    ([L\-])     \s+
                    ([\w\s]+\w) \s+
                    ([R\-])     \s*
                    [\)\*]      \s*$/x) {
        return;
    }
    my ($indent_string, 
	$head_string,
	$left_anchor_string,
	$genus_name,
	$right_anchor_string) = ($1, $2, $3, $4, $5);

    my $depth = length ($indent_string)/length ($Globals::indent);
    my $is_head = ($head_string eq '*') ? 1 : 0;
    
    my $genus = &Globals::get_representations->get ($genus_name);

    unless (ref ($genus) and $genus->isa ('PrCatGenus')) {
        return;
    }

    my $new = $genus->unanchored;
    shift @{$text};

    $new->set_left_anchor if ($left_anchor_string eq 'L');
    $new->set_right_anchor if ($right_anchor_string eq 'R');

    while (@{$text}) {

        # Have we run out of dependents?
        my $next_line = $text->[0];
        $next_line =~ /^((?:$Globals::indent_re)*)/;
        my $next_depth = length ($1)/length ($Globals::indent);
        last if ($next_depth <= $depth);

        my ($child, $child_depth, $child_is_head) =
            &PrCat::from_text ($text);
        unless ($child) {
            ($child, $child_depth, $child_is_head) =
                &UnaryFeature::from_text ($text);
        } 
        unless ($child) {
            ($child, $child_depth, $child_is_head) =
                &BinaryFeature::from_text ($text);
        }  
        unless ($child) {
            ($child, $child_depth, $child_is_head) =
		&PrCatVariable::from_text ($text);
        }
        unless ($child) {
            ($child, $child_depth, $child_is_head) =
		&UnaryFeatureVariable::from_text ($text);
        }
        unless ($child) {
            ($child, $child_depth, $child_is_head) =
		&BinaryFeatureVariable::from_text ($text);
        }
        unless ($child) {
            carp "Couldn't parse $next_line: $!";
            confess;
            die;
        }
	
	unless ($new->put ($child)) {
	    carp "Error in $next_line: $!";
	    carp "Can't add $child of genus " . $child->get_genus_name;
	    confess;
	    die;
	}
        if ($child_is_head) {
	    $new->set_head ($child);
	}
    }

    ($new, $depth, $is_head);
}

sub print {
    my (
        $this,
        $depth,     # Depth of nesting
	$is_head    # Head of parent?
        ) = (@_, 0, 0);

    print (join "\n", $this->to_text ($depth));
    print "\n";
}

sub to_tree {
    my $this = shift;
    my $width = (shift or '1in');

    my $indent_level = (shift or 0);
    my $indent_quantum = 2;
    my $indent = ' ' x ($indent_quantum * (1+$indent_level));

    my $leader_level = shift;   #Ignored for PrCats;
    
    my $style = (shift or '\ttfamily');
    my $is_head = shift;
    
    my $string = "{$style " .
	($this->get_left_anchor ? "L " : "- ") .
	($this->get_genus_abbrev) .
	($this->get_right_anchor ? " R" : " -") .
	'}';
    if ($is_head) {
	$string = "\\fbox{$string}";
    }
    $string = $indent . "[." . $string . "\n";
    
    for ($this->get_all_constituents) {
	my $tree;
	my ($fbox_left, $fbox_right) = ("", "");
	if ($this->is_head_constituent ($_)) {
	    $tree = $_->to_tree ($width, $indent_level+1, 
				 0, $style, 1);
	    $fbox_left = '\fbox{'; $fbox_right = '}';
	} else {
	    $tree = $_->to_tree ($width, $indent_level+1, 
				 0, $style, 0);
	}
	if ($_->isa ('UnaryFeature') or $_->isa ('BinaryFeature')) {
	    $string .= $indent . (' ' x $indent_quantum) .
		$fbox_left . '\parbox{' .
		$width ."}{\n" . $tree .
		$fbox_right;;
	    $string .= $indent . (' ' x $indent_quantum) . "}\n";
	} else {	    
	    $string .= $tree;
	}
    }

    $string .= $indent . ']' . "\n";
    $string;
}



1;

