package PrCatGenus;
use PrCat;
use strict;
use warnings;
use Carp;

########################################################################
#
# PrCatGenus.pm
#
# Used for defining a prosodic hierarchy.  A PrCatGenus is where you 
# store the general information that, e.g., a Syllable consists of 
# Segments, and is itself found in a Foot.
#
# PrCats differ from Features in that PrCats are nameless, have left
# and right boundary anchors, and have an unlimited number of nameless
# ordered dependents.
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2009 March 8 ($)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


########################################################################
#
# Constructors
#
########################################################################

sub new {
    my $class = shift;
    my (
	$name,    # Name of PrCat genus ("Syllable", etc.)
	$abbrev   # Short name of PrCat genus
	) = @_;

    my $this = {};
    $this->{name} = $name;
    $this->{abbrev} = $abbrev;

    bless ($this, $class);
}

########################################################################
#
# Access methods
#
########################################################################

sub get_name {
    my $this = shift;
    $this->{name};
}

sub get_abbrev {
    my $this = shift;
    $this->{abbrev};
}

# What PrCatGenus or UnaryFeatureGenus can be a constituent?  At present,
# all constitutents must be of same type (i.e., only one constituent type
# is allowed).  This will have to change if we want to do extrametricality,
# ambisyllabicity, etc., since we'll need to violate Strict Layering.

sub set_constituent {
    my $this = shift;
    my $constituent = shift;
    unless (ref ($constituent) and
	    ($constituent->isa ('PrCatGenus') or
	     $constituent->isa ('UnaryFeatureGenus')
	    )) { 
	carp "Constituent type ($constituent) must be PrCatGenus or UnaryFeatureGenus: $!";
	confess;
	die;
    }
    $constituent->set_parent ($this);
    $this->{constituent} = $constituent;
}

sub get_constituent {
    my $this = shift;
    $this->{constituent};
}

sub is_dependent_of {
    my ($this, $that) = @_;
    $this->get_parent eq $that;
}

# set_parent -- Create link to parent type.
sub set_parent {
    my $this = shift;
    my $new_parent = shift;
    unless ($new_parent->isa ('PrCatGenus') or $new_parent->isa ('MCatGenus')) {
	carp "Parent ($new_parent) must be a PrCatGenus: $!";
	confess;
	die;
    }
    $this->{parent} = $new_parent;
}

sub get_parent {
    my $this = shift;
    $this->{parent};
}

########################################################################
#
# Methods for creating a PrCat node from this PrCatGenus.
#
########################################################################

# unanchored, left_anchored, right_anchored, double_anchored -- create
#     PrCat of this type with specified settings of boundary anchors.

sub unanchored {
    my $this = shift;
    my $new = new PrCat ($this);
    $new->unset_left_anchor;
    $new->unset_right_anchor;
    $new;
}

sub left_anchored {
    my $this = shift;
    my $new = new PrCat ($this);
    $new->set_left_anchor;
    $new->unset_right_anchor;
    $new;
}

sub right_anchored {
    my $this = shift;
    my $new = new PrCat ($this);
    $new->unset_left_anchor;
    $new->set_right_anchor;
    $new;
}

sub double_anchored {
    my $this = shift;
    my $new = new PrCat ($this);
    $new->set_left_anchor;
    $new->set_right_anchor;
    $new;
}

# plus, minus -- Create [+/-alpha Foo] Variables matching PrCats of this type.

sub plus {
    my $this = shift;
    my $species = shift;
    unless ($species) {
	carp "PrCat can be neither + nor -";
	confess;
	die;
    }
    plus PrCatVariable ($this, $species);
}

sub minus {
    my $this = shift;
    my $species = shift;
    unless ($species) {
	carp "PrCat can be neither + nor -";
	confess;
	die;
    }
    minus PrCatVariable ($this, $species);
}

########################################################################
#
# I/O methods
#
########################################################################

sub print {
    my $this = shift;
    my $indent = (shift or "");

    print $indent;
    printf "%s (%s) ", $this->get_name, $this->get_abbrev;
    if ($this->get_parent) {
	print ", dependent of ", $this->get_parent->get_name;
    }
    print "\n";
    print "$indent "; 
    if ($this->get_constituent) {
	$this->get_constituent->print ("$indent  ");
    }
}

1;
