package Recombinator;
use strict;
use warnings;
use Carp;
use Globals;

########################################################################
#
# Recombinator.pm
#
# Does the die-rolling for the breed () methods; is passed to them as
# an argument.  This is so that we can have different Recombinators, and
# so fiddle with the operations that breeding carries out.  Based on much 
# older Mutagen.pm
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2019 April 29 (M)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


sub new {
    my $class = shift;
    my $this = {};
    bless $this, $class;
    my %Ps = @_;
    for (keys %Ps) {
	$this->set_probability ($_, $Ps {$_});
    }
    
    srand;
    $this;
}

sub set_name {
    my (
	$this,
	$name
	) = @_;

    unless ($name =~ /./) {
	confess "Name must be non-null: $!";
    }
    
    $this->{name} = $name;
}

sub get_name {
    my $this = shift;

    $this->{name};
}

sub set_probability {
    my $this = shift;
    my ($name, $p) = @_;
    $p or ($p = 0);
    if ($p < 0 or $p > 1) {
	carp "Tried to set P($this) to invalid probability $p:$!";
	confess;
	die;
    }
    $this->{contents}->{$name} = $p;
}


########################################################################
#
# Saving throws for breeding
#
########################################################################

# Probability that when breed ($this, $that) is called, $this wins.
sub this_wins {
    my $this = shift;   # Not the same $this!
    my $P_this_wins = $this->{contents}->{"P_this_wins"};
    unless (defined $P_this_wins) {
	confess "P_this_wins undefined";
    }
    rand () < $P_this_wins;
}

# Probability that when one of the two breedees is undefined, the
# undefined one wins (i.e., probability of intersecting the trees)
sub null_wins {
    my $this = shift;

    my $P_null_wins = $this->{contents}->{"P_null_wins"};
    unless (defined $P_null_wins) {
	confess "P_null_wins undefined";
    }
    rand () < $P_null_wins;
}


########################################################################
#
# I/O routines
#
########################################################################

sub to_text {
    my (
	$this,
	$depth
	) = (@_, 0);
    
    my $indent_string = $Globals::indent x $depth;

    my @text;
    my $name = $this->get_name;
    push @text, $indent_string . "Recombinator $name";
    for my $pname (sort (keys %{$this->{contents}})) {
	my $pval = $this->{contents}->{$pname};
	push @text, "$indent_string$Globals::indent$pname\t$pval";
    }
    @text;
}

sub from_text {
    my $text = (ref $_[0] eq 'ARRAY') ? $_[0] : \@_;
    
    my $indent_re = $Globals::indent_re;

    unless (defined $text->[0]) {
	return;
    }

    unless ($text->[0] =~ /^(($indent_re)*)Recombinator\s+([\w.:\-+]+)/) {
	return;
    }
    my $indent_string = $1;
    my $name = $3;

    shift @{$text};

    my $recombinator = new Recombinator;
    $recombinator->set_name ($name);
    while ($text->[0] =~ /^$indent_string$indent_re(\w+)\s+(\d*\.?\d+)/) {
	my ($pname, $pval) = ($1, $2);
	
	unless ($pname =~ /./) {
	    confess "Ill-formed mutation probability name $pname: $!";
	}
	unless ($pval =~ /./) {
	    confess "Ill-formed probability $pname = $pval: $!";
	}
	$recombinator->set_probability ($pname, $pval);
	shift (@{$text});
    }

    $recombinator;
}



sub print {
    my $this = shift;
    for (keys %{$this->{contents}}) {
	printf "%30s %4.3f\n", $_, $this->{contents}->{$_};
    }
}

sub to_latex {
    my $this = shift;
    my $latex = "
\\begin{center}
  \\begin{tabular}{ll}
";
    for (keys %{$this->{contents}}) {
	$latex .= sprintf "%s & %4.3f\\\\\n", $_, $this->{contents}->{$_};
    }

    $latex .= "
  \\end{tabular}

";
    $latex .= "
\\end{center}
";
    $latex =~ s/_/ /g;

    $latex;
}

1;
