package RecombinatorSet;
use Globals;
use Recombinator;
use warnings;
use strict;
use Carp;

########################################################################
#
#  RecombinatorSet.pm
#
#  A set of named Recombinators, which can be passed to a Learner, so that
#  the Learner can use different Recombinators on different trials.
#
#  E. Moreton  *  UNC-Chapel Hill Linguistics  *  2019 April 29 (M)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


sub new {
    my $class =shift;

    my $this = {};
    bless $this, $class;

    $this;
}

########################################################################
#
#  Access methods
#
########################################################################

sub add {
    my (
	$this,
	$recombinator
	) = @_;

    unless (ref $recombinator and ($recombinator->isa ('Recombinator'))) {
	confess "First argument to add must be Recombinator, not $recombinator: $!";
    }
    my $name = $recombinator->get_name;
    unless ($name =~ /./) {
	confess "Non-null name required for Recombinator: $!";
    }
    if ($this->{contents}->{$name}) {
	confess "Recombinator $name already exists: $!";
    }

    $this->{contents}->{$name} = $recombinator;
}

sub get_recombinator_named {
    my (
	$this,
	$name
	) = @_;

    unless (defined $name) {
	confess "Recombinator name must be non-null: $!";
    }

    my $recombinator = $this->{contents}->{$name};
    unless (defined $recombinator) {
	confess "Unknown recombinator $name: $!";
    }
    $recombinator;
}

sub get_names {
    my $this = shift;

    keys %{$this->{contents}};
}

sub get_all {
    my $this = shift;

    values %{$this->{contents}}
}

########################################################################
#
#  I/O methods
#
########################################################################

sub to_text {
    my (
	$this, 
	$depth
	) = (@_, 0);

    my $indent_string = $Globals::indent x $depth;
    my @text;
    push @text, $indent_string . "RecombinatorSet";
    for my $recombinator ($this->get_all){
	push @text, $recombinator->to_text ($depth + 1);
    }
    @text;
}

sub from_text {
    my $text = (ref $_[0] eq 'ARRAY') ? $_[0] : \@_;
    
    my $indent_re = $Globals::indent_re;
    
    unless ($text->[0] =~ /^(($indent_re)*)RecombinatorSet/) {
	return;
    }
    my $indent_string = $1;
    my $name = $3;
    shift @{$text};

    my $ms = new RecombinatorSet;
    while (my $recombinator = &Recombinator::from_text ($text)) {
	last unless (ref $recombinator and $recombinator->isa ("Recombinator"));
	$ms->add ($recombinator);
    }

    $ms;
}

    
1;

