package Report;
use strict;
use warnings;
use Carp;
use IO::Handle;  # For flushing filehandles

########################################################################
#
#  Report.pm
#
#  Write report of simulation.  For a simulation named Foo, the report
#  goes into a directory with the following structure:
#
#  Foo (dir)
#    Invocation.txt      Command-line invocation
#    Realtime.txt        Wall-clock seconds elapsed from start of run
#    Data.csv            For each trial (row), the following (columns):
#                          (Indep. variables copied from the .exp file)
#                          simname
#                          N
#                          zeta
#                          eta
#                          parth
#                          mutagen
#                          meta (= metaconstraint file name)
#                          trial_type
#                          pr_corr
#                          corr
#    Log.txt               Macro-constraint population after each error,
#                            human-readable and very wordy.  Each macro-
#                            constraint is represented by its shortest
#                            micro-constraint.
#    Report.csv          For each trial (column), the following (rows):
#                          pos_stim
#                          neg_stim
#                          pr_corr
#                          correct
#                          harmony of Stimulus 1
#                          :
#                          harmony of Stimulus m
#                          weight of (macro-)Constraint 1
#                          :
#                          weight of (macro-)Constraint n
#    Stimuli.txt         Stimulus-number decoder
#    Constraints.txt     Constraint-number decoder
#    Lineage.txt         Which macro-constraints spawned which micro
#
#  E. Moreton  *  UNC-Chapel Hill Linguistics  *  2018 December 8 (S)  
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################



our $fmt = "%04.3g";    # For sprintf
our $fh;

sub new {
    my (
	$class,
	$learner,
	$name,       # Unique identifying name for this simulation
	$path,       # Directory to put the report subdirectory in
	$log_format  # Format to print constraints in in Log.txt
	) = @_;

    my $this = {};

    $this->{learner}   = $learner;
    $this->{name}      = $name;
    $this->{pos_stim}  = [];    # pos stimnum [trialnum]
    $this->{neg_stim}  = [];    # neg stimnum [trialnum]
    $this->{pr_corr}   = [];    # pr(corr) [trialnum]
    $this->{correct}   = [];    # correct? [trialnum]
    $this->{harmonies} = [];    # (harmony [trialnum])[stimnum]
    $this->{weights}   = [];    # (weight [trialnum])[constraintnum]
    $this->{log_fh}    = undef;    # Filehandle for log file        
    $this->{log_format} = ($log_format or 'text'); # Constraint format for log file
    $this->{start_time} = time;
    
    $this->{dir} = "$path/$name";
    mkdir "$this->{dir}" or carp "Couldn't create directory $path/$name: $!";

    bless $this, $class;

    $this;
}

sub get_learner {
    my $this = shift;

    $this->{learner};
}

sub get_dir {
    my $this = shift;
    
    $this->{dir};
}

sub get_simulation_name {
    my $this = shift;

    $this->{name};
}

sub start {
    my $this = shift;
    
    $this->write_invocation;
    $this->write_stimuli;
    $this->open_log;
    $this->add_trial (undef, undef, undef, undef);
    $this->add_to_log ("NA", "NA", "NA", "NA");
}

sub update {
    my $this = shift;

    $this->write_report;
    $this->write_constraints;
    $this->write_lineage;
    $this->write_harmony;
    $this->write_weight;
    $this->write_learner;
}

sub finish {
    my $this = shift;

    $this->close_log;
    $this->write_report;
    $this->write_constraints;
    $this->write_lineage;
    $this->write_harmony;
    $this->write_weight;
    $this->write_learner;
    $this->write_realtime;
}


########################################################################
#
#  Methods for Invocation.txt 
#
########################################################################

sub write_invocation {
    my $this = shift;
    
    my $invocation = "perl $0 ".(join ' ', @ARGV). "\n";
    my $invocation_txt = $this->get_dir . "/Invocation.txt";

    open (INVOCATION, "> $invocation_txt") or
	carp "Couldn't open $invocation_txt: $!";
    print INVOCATION $invocation;
    close INVOCATION;
}

########################################################################
#
#  Methods for Realtime.txt
#
########################################################################

sub write_realtime {
    my $this = shift;

    my $realtime_csv = $this->get_dir . "/Realtime.txt";
    open (REALTIME, "> $realtime_csv") or
	carp "Couldn't open $realtime_csv: $!";
    print REALTIME (time - $this->{start_time});
    print REALTIME "\n";
}
    

########################################################################
#
#  Methods for Stimuli.csv
#
#  The comma-separated fields are 
#    <line number>
#    <stimulus name>:   Normally a serial number
#    <teaching output>: 1 if the stim is positive, -1 if it is negative
#    <struc>:           The flat-string representation of the stimulus
#
########################################################################

sub write_stimuli {
    my $this = shift;
    
    my $stimuli_csv = $this->get_dir . "/Stimuli.csv";
    open (STIMULI, "> $stimuli_csv") or
	carp "Couldn't open $stimuli_csv: $!";
    print STIMULI "line,stimname,teach,weight,struc\n";

    my @stimulus_space = $this->get_learner->get_stimulus_space->get_all;
    for my $i (0..($#stimulus_space)) {
	my $stim = $stimulus_space [$i];
	print STIMULI "$i," . $stim->get_name . ",". 
	    $stim->get_teach . "," . $stim->get_weight . "," . 
	    $stim->get_rep->to_flat_string . "\n";
    }
    close STIMULI;
}

########################################################################
#
#  Methods for Constraints.csv
#
#  The comma-separated fields are
#    <line number>
#    <class number>
#    <flat-string representation of a representative>
#    <signature>
#
########################################################################

sub write_constraints {
    my $this = shift;
    
    my $constraints_csv = $this->get_dir . "/Constraints.csv";
    open (CONSTRAINTS, "> $constraints_csv") or
	carp "Couldn't open $constraints_csv: $!";
    print CONSTRAINTS "line,conname,struc,signature\n";

    my $line = 1;
    my $learner = $this->get_learner;
    for my $sig ($learner->get_all_signatures) {
        my $sig_class_num = $learner->get_signature_class_number ($sig);
	my $sig_class_ex  = $learner->get_signature_class_example ($sig);
	print CONSTRAINTS $line++ . ",$sig_class_num,";
	if (ref $sig_class_ex and $sig_class_ex->can ("to_flat_string")) {
	    print CONSTRAINTS $sig_class_ex->to_flat_string . ",";
	} else {
	    print CONSTRAINTS "(none),";
	}
	print CONSTRAINTS "$sig\n";
    }

    close CONSTRAINTS;
}

########################################################################
#
#  Methods for Lineage.csv
#
#  This file keeps track of the macro-constraint classes that go with
#  each mom-dad-kid triad.  The structure is:
#
#  <line>,<mom class>,<dad class>,<kid class>,<number of times>
#
########################################################################

sub write_lineage {
    my $this = shift;
   
    my $lineage_csv = $this->get_dir . "/Lineage.csv";
    open (LINEAGE, "> $lineage_csv") or
	carp "Couldn't open $lineage_csv: $!";
    print LINEAGE "line,mom,dad,kid,count\n";

    my $line = 1;
    my $lineage = $this->get_learner->get_lineage;

    foreach my $mom_class (0..@{$lineage}) {
	next unless $lineage->[$mom_class];
	foreach my $dad_class (0..@{$lineage->[$mom_class]}) {
	    next unless $lineage->[$mom_class]->[$dad_class];
	    foreach my $kid_class (0..@{$lineage->[$mom_class]->[$dad_class]}) {
		next unless $lineage->[$mom_class]->[$dad_class]->[$kid_class];
		my $count = @{$lineage->[$mom_class]->[$dad_class]->[$kid_class]};
		print LINEAGE $line++ . ",$mom_class,$dad_class,$kid_class,$count\n";
	    }
	}
    }

    close (LINEAGE);
}

########################################################################
#
#  Methods for Log.txt
#
#  This is the more-or-less human-readable progress log that is updated
#  while the simulation is running.  Each update to the log prints the
#  one representative example of each macro-constraint, together with
#  the number of micro-constraints.
#
########################################################################

sub open_log {
    my $this = shift;
    
    my $log_txt = $this->get_dir . "/Log.txt";
    open $this->{log_fh}, ">", $log_txt or
	confess "Couldn't open $log_txt: $!";
    $fh = $this->{log_fh};
}

sub add_to_log {
    my (
	$this,
	$correct_stim,
	$chosen_stim,
	$pr_corr
	) = @_;

    my $trialnum = $this->get_trial_number;
    print {$fh} "\n"."vvvv $trialnum vvvv\n";
    print {$fh} "correct:  $correct_stim; chosen:  $chosen_stim; pr_corr:  $pr_corr\n";
    $this->get_learner->report_population ($this->{log_fh}, $this->{log_format});
    # MORE GUTS GO HERE
    flush $fh;
}

sub print_to_log {
    for (@_) {
	print {$fh} "$_\n";
    }
    flush $fh;
}

sub close_log {
    my $this = shift;

    # GUTS GO HERE
    close ($this->{log_fh});
}

########################################################################
#
#  Methods for Data.csv
#
#  The structure is as follows:
#  
#  Columns:
#    (Indep. variables copied from the .exp file)
#    simname
#    N
#    zeta
#    eta
#    parth
#    mutagen
#    meta (= metaconstraint file name)
#    trial_type
#    pr_corr
#    corr
#
#  Rows:  One for each row of the .exp file.
#
########################################################################


########################################################################
#
#  Methods for Report.csv
#
#  The report is not printed until the end of the simulation.  The 
#  structure is as follows:
#
#  Columns:
#    <line number>
#    <simulation name>
#    <data type>       : "pos_stim_name", "neg_stim_name", "pr_corr", "corr", 
#                        "stim_name", or "constraint_name"
#    <data value>      : Name of stimulus or name of constraint if <data_type>
#                        is "stim_name" or "constraint_name", else NA
#    <trial 1>
#    :
#    <trial ntrials>
#
#  Rows:
#     1,"pos_stim_name",NA,<Trial 1 pos stim>,...,<Trial N pos stim>
#     2,"neg_stim_name",NA,<Trial 1 neg stim>,...,<Trial N neg stim>
#     3,"pr_corr",NA,<Trial 1 Pr(corr)>,...,<Trial N Pr(corr)>
#     4,"corr",NA,<Trial 1 T/F>,...<Trial N T/F>
#     5,"harmony",<name of Stim 1>,<Trial 1 harmony>,...<Trial N harmony>
#     :
#   m+6,"harmony",<name of Stim m>,<Trial 1 harmony>,...<Trial N harmony>
#   m+7,"weight",<name of Constraint 1>,<Trial 1 weight>,...,
#       <Trial N weight>
#     :
# m+n+8,"weight",<name of Constraint n>,<Trial 1 weight>,...,
#       <Trial N weight>
#
########################################################################

sub add_trial {
    my (
	$this,
	$pos_stim,
	$neg_stim,
	$pr_corr,
	$correct
	) = @_;

    unless (defined $pos_stim) {
	$pos_stim = 'NA'
    }
    unless (defined $neg_stim) {
	$neg_stim = 'NA';
    }
    unless (defined $pr_corr) {
	$pr_corr = 'NA';
    }
    unless (defined $correct) {
	$correct = 'NA';
    }

    push @{$this->{pos_stim}}, $pos_stim;
    push @{$this->{neg_stim}}, $neg_stim;
    push @{$this->{pr_corr}}, $pr_corr;
    push @{$this->{correct}}, $correct;

    my $learner = $this->get_learner;

    # Record harmony of each stimulus
    for my $stim ($learner->get_stimulus_space->get_all) {
	push @{$this->{harmony}->{$stim}}, 
	  $learner->get_current_harmony ($stim);
    }

    # Record weight of each constraint
    my $zeta = $learner->get_zeta;
    for my $sig ($learner->get_all_signatures) {
        my $sig_class_num = $learner->get_signature_class_number ($sig);
	my $sig_class_size = $learner->get_signature_class ($sig);
        # If this is a new constraint (new signature class), record 0 for
	# all previous trials
	my $trial = $this->get_trial_number;
	if (!exists $this->{weight}->{$sig_class_num} and $trial>=1) {
	    for (0..($trial-1)) {
		push @{$this->{weight}->{$sig_class_num}}, 0;
	    }
	} 
	push @{$this->{weight}->{$sig_class_num}}, $sig_class_size * $zeta;

    }	
}

sub get_pos_stims {
    my $this = shift;
    @{$this->{pos_stim}};
}
sub get_neg_stims {
    my $this = shift;
    @{$this->{neg_stim}};
}
sub get_pr_corr {
    my $this = shift;
    @{$this->{pr_corr}};
}
sub get_correct {
    my $this = shift;
    @{$this->{correct}};
}
sub get_harmony {
    my $this = shift;
    %{$this->{harmony}};
}
sub get_weight {
    my $this = shift;
    %{$this->{weight}};
}
sub get_trial_number {
    my $this = shift;
    0+$this->get_learner->get_last_trial;
}

sub write_report {
    my $this = shift;

    my $report_csv = $this->get_dir . "/Report.csv";
    open ($this->{report_fh}, "> $report_csv") or
	carp "Couldn't open $report_csv: $!";
    local *FH = $this->{report_fh};
    print FH "row,simulation_name,data_type,data_name,";
    print FH (join ",", (0..$this->get_trial_number));
    print FH "\n";

    my $name = $this->get_simulation_name;
    local *FH = $this->{report_fh};
    my $line = 1;

    print FH $line++ . ",$name,pos_stim_name,NA,";
    print FH (join ",", $this->get_pos_stims);
    print FH "\n";

    print FH $line++ . ",$name,neg_stim_name,NA,";
    print FH (join ",", $this->get_neg_stims);
    print FH "\n";

    print FH $line++ . ",$name,pr_corr,NA,";
    my @pr_corr = map {
	if (defined $_ and ($_ ne 'NA')) {
	    my $formatted = sprintf $fmt, $_;
	    $formatted =~ s/^-?0*([^.])/$1/;
	    $formatted;
	} else {
	    "NA"
	}
    } ($this->get_pr_corr);
    print FH (join ",", @pr_corr);
    print FH "\n";

    print FH $line++ . ",$name,correct,NA,";
    my @correct = map {
	if (defined $_) {
	    $_ ? "T" : "F"
	} else {
	    "NA"
        } 
    } ($this->get_correct);
    print FH (join ",", @correct);
    print FH "\n";

    close FH;
}




########################################################################
#
#  Methods for Harmony.csv
#
#     <line>,<simname>,<name of Stim 1>,<Trial 1 harmony>,...<Trial N harmony>
#
########################################################################

sub write_harmony {
    my $this = shift;

    my $harmony_csv = $this->get_dir . "/Harmony.csv";
    open ($this->{harmony_fh}, "> $harmony_csv") or
	carp "Couldn't open $harmony_csv: $!";
    local *FH = $this->{harmony_fh};
    print FH "row,simulation_name,stimulus_name,";
    print FH (join ",", (0..$this->get_trial_number));
    print FH "\n";

    my $name = $this->get_simulation_name;
    my $line = 1;

    # For each stimulus, print its name and its harmony on each trial
    my @stims = $this->get_learner->get_stimulus_space->get_all;
    my %harmony = $this->get_harmony;
    for my $stim (@stims) {
	print FH $line++ . ",$name," . $stim->get_name . ",";
	my @harmonies = map {
	    my $raw = (defined $_) ? $_ : 0;
	    my $formatted = sprintf $fmt, $raw;
	    $formatted =~ s/^-?0*([^.])/$1/;
	    $formatted;
	} @{$harmony {$stim}};
	print FH (join ",", @harmonies);
	print FH "\n";
    }

    close (FH);
}


########################################################################
#
#  Methods for Weight.csv
#
#     <line>,<simname>,<name of Con 1>,<Trial 1 weight>,...<Trial N weight>
#
########################################################################

sub write_weight {
    my $this = shift;

    my $weight_csv = $this->get_dir . "/Weight.csv";
    open ($this->{weight_fh}, "> $weight_csv") or
	carp "Couldn't open $weight_csv: $!";
    local *FH = $this->{weight_fh};
    print FH "row,simulation_name,constraint_name,";
    print FH (join ",", (0..$this->get_trial_number));
    print FH "\n";

    my $name = $this->get_simulation_name;
    my $line = 1;

    # For each constraint class (macro-constraint), print its name and
    # its weight (number of micro-constraints times zeta) on each trial
    my %weight = $this->get_weight;
    for my $sig_num (keys %weight) {
	print FH $line++ . ",$name,$sig_num,";
	my @weights = map {
	    my $formatted = sprintf $fmt, $_;
	    $formatted =~ s/^-?0*([^.])/$1/;
	    $formatted;
	} @{$weight {$sig_num}};
	print FH (join ",", @weights);
	print FH "\n";
    }

    close (FH);
}

########################################################################
#
#  Methods for Learner.txt
#
#  Save the state of the learner so that we can read it in again.
#
########################################################################

sub write_learner {
    my $this = shift;
    
    my $learner_txt = $this->get_dir . "/Learner.txt";
    open (LEARNER, "> $learner_txt") or 
	carp "Couldn't open $learner_txt";
    print LEARNER (join "\n", $this->get_learner->to_text);
    close (LEARNER);
}


1;

