package Representations;
use Globals;
use ConstraintSet;
use UnaryFeatureGenus;
use BinaryFeatureGenus;
use PrCatGenus;
use MCatGenus;
use Carp;
use strict;
use warnings;

########################################################################
#
# Representations.pm
#
# Used to define a representational system.  This consists of two 
# components:
#
# 1.  A tree of FooGenera, rooted at some ``base'' (e.g., PrWd).  This
#     is obligatory.
#
# 2.  A ConstraintSet called ``forbidden''.  Each constraint in that
#     set specifies a configuration which is absolutely forbidden in
#     the representational system, e.g., [+high +low].  This can be
#     empty.
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


########################################################################
#
# Constructors
#
########################################################################

sub new {
    my $class = shift;
    my $this = {};

    bless ($this, $class);
    $this;
}


# from_text:  Read a text description and build the necessary tree.
sub from_text {
    my $this = shift;
    my $raw_text = $_[0];
    unless (ref ($raw_text) eq 'ARRAY') {
	$raw_text = \@_;
    }

    # Purge comment lines
    my $text = [grep (/^[^#]/, @{$raw_text})];

    $this->tree_from_text ($text);
    
    shift @{$text};
    ($this->{forbidden}) = &ConstraintSet::from_text ($text) or (new ConstraintSet);
}

sub tree_from_text {
    my (
	$this,
	$text
	) = @_;
    
    my $line = shift @{$text};
    return if ($line =~ /^\s*$/);

    unless ($line =~ /^((?:$Globals::indent)*)(\w+)\s+\[([\w\s]+)\]\s+\[([\w\s]+)\]/) {
	carp "Couldn't parse $line: $!";
	confess;
	die;
    }
    my ($indent, $type, $name, $abbrev) = (
	length ($1)/length ($Globals::indent), $2, $3, $4);

    my $genus;
  SWITCH: {
      if ($type eq 'BinaryFeature') {
	  $genus = new BinaryFeatureGenus ($name, $abbrev);
	  last SWITCH;
      }
      if ($type eq 'UnaryFeature') {
	  $genus = new UnaryFeatureGenus ($name, $abbrev);
	  last SWITCH;
      }
      if ($type eq 'PrCat') {
	  $genus = new PrCatGenus ($name, $abbrev);
	  last SWITCH;
      }  
      if ($type eq 'MCat') {
	  $genus = new MCatGenus ($name, $abbrev);
	  last SWITCH;
      }  
      carp "$type ($name, $abbrev) doesn't belong in a representation: $!";
      confess;
      die;
    }

    # If there are dependents (things indented further), add them.  We're
    # currently parsing every line twice, which is inefficient.
    while (@{$text}) {
	my $next_line = $text->[0];
	last if ($next_line =~ /^\s*$/);
	unless ($next_line =~ /^((?:$Globals::indent)*)(\w+)\s+\[([\w\s]+)\]\s+\[([\w\s]+)\]/) {
	    carp "Can't parse $next_line: $!";
	    confess;
	    die;
	}
	my $next_line_indent  = length ($1)/length ($Globals::indent);

	last if ($next_line_indent <= $indent);

	my $dependent = $this->tree_from_text ($text);

	if ($type eq 'MCat') {
	    $genus->add_dependent ($dependent);
	}
	if ($type eq 'UnaryFeature' or $type eq 'BinaryFeature') {
	    $genus->add_dependent_features ($dependent);
	}
	if ($type eq 'PrCat') {
	    $genus->set_constituent ($dependent);
	}
    }
    
    $this->{by_name}->{$name} = $genus;
    $this->{by_abbrev}->{$abbrev} = $genus;
    $this->{base} = $genus;
}

########################################################################
#
# Access methods
#
########################################################################

sub get_base {
    my $this = shift;
    $this->{base};
}

sub get {
    my ($this,
	$target) = @_;

    $this->get_genus_named ($target) or $this->get_genus_abbreviated ($target);
}

sub get_genus_named {
    my ($this,
	$name) = @_;
    $this->{by_name}->{$name};
}

sub get_genus_abbreviated {
    my ($this,
	$abbrev) = @_;
    $this->{by_abbrev}->{$abbrev};
}

sub get_all_PrCats {
    my $this = shift;
    grep {
	$_->isa ('PrCatGenus')
    } (values (%{$this->{by_name}}));
}

sub get_all_features {
    my $this = shift;
    grep {
	$_->isa ('UnaryFeatureGenus') or $_->isa ('BinaryFeatureGenus')
    } (values (%{$this->{by_name}}));
}

sub set_forbidden {
    my ($this, $forbidden) = @_;
    $this->{forbidden} = $forbidden;
}

sub get_forbidden {
    my $this = shift;
    $this->{forbidden};
}

sub is_forbidden {
    my (
	$this,
	$rep      # a representational structure
	) = @_;

    my $bad_flag;

    my $forbidden = $this->get_forbidden;

    if ($forbidden and keys %{$forbidden}) {
        for ($forbidden->get_all) {
            my $polarity = $_->get_polarity;
            if (
                # Things to definitely not have:                                                        
                ( ($polarity < 0) and ($_->matches_in ($rep)) )
                or
                # Things to definitely have:                                                            
                ( ($polarity > 0) and !($_->matches_in ($rep)) )
                ) {
                $bad_flag = 1;
                last;
            }
        }
    }
    $bad_flag;
}

########################################################################
#
# I/O methods
#
########################################################################

sub print {
    my $this = shift;
    $this->get_base->print;
    print "\n";
    print (join "\n", $this->get_forbidden->to_text);
    print "\n";
}
	
1;
