package SHJ1961;
use Representations;
use UnaryFeatureGenus;
use BinaryFeatureGenus;
use PrCatGenus;
use Carp;
use strict;
use warnings;
use Exporter;
our @ISA = 'Exporter';
our @EXPORT = qw($SHJ1961, $text);

########################################################################
#
# SHJ1961.pm
#
# Makes a representation based on one of the SHJ 1961 color/shape/size
# stimuli.
#
# This package isn't part of the Subtree Constraint Schema; none of those
# packages depend on it, and scripts using the SCS packages are free to
# substitute something else for this. 
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2018 June 4 (M)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


our $SHJ1961; # Three-bit representations
our $text;   # Feature tree (as text) => ASCII IPA symbol

my $RepSys = new Representations;
my @text = &Globals::read_text ($Globals::home_dir . "/SHJ1961.rep");
&Globals::set_representations ($RepSys);
$RepSys->from_text (@text);

#_______________________________________________________________________
my @Root = ($RepSys->get ('Root'), '+');
my @A  = ($RepSys->get ('big'), '+');
my @a  = ($RepSys->get ('big'), '-');
my @B  = ($RepSys->get ('black'), '+');
my @b  = ($RepSys->get ('black'), '-');
my @C  = ($RepSys->get ('triangle'), '+');
my @c  = ($RepSys->get ('triangle'), '-');



#_______________________________________________________________________
$SHJ1961->{"+++"} = [(@Root, @A, @B, @C)];
$SHJ1961->{"++-"} = [(@Root, @A, @B, @c)];
$SHJ1961->{"+-+"} = [(@Root, @A, @b, @C)];
$SHJ1961->{"+--"} = [(@Root, @A, @b, @c)];
$SHJ1961->{"-++"} = [(@Root, @a, @B, @C)];
$SHJ1961->{"-+-"} = [(@Root, @a, @B, @c)];
$SHJ1961->{"--+"} = [(@Root, @a, @b, @C)];
$SHJ1961->{"---"} = [(@Root, @a, @b, @c)];

# Make hash for reverse lookup
for my $figure_code (keys %{$SHJ1961}) {
    my $tree = join ("; ", &to_feature_tree ($figure_code)->to_text);
    if (exists ($text->{$tree})) {
        $text->{$tree} .= " $figure_code";
    } else {
        $text->{$tree} = "$figure_code";
    }
}

#_______________________________________________________________________

sub get {
    my $target = shift;
    $RepSys->get ("$target");
}


sub to_representation {
    my $string = shift;
    my $rep = $RepSys->get ('Figure')->double_anchored;
    my @symbols = (split /\s/, $string);
    for my $s (@symbols) {
	if (is_symbol ($s)) {
	    $rep->put (to_feature_tree ($s));
	    next;
	}
	carp "Unrecognized symbol $s: $!";
	confess;
	die;
    }
    $rep;
}

### The methods below are the SHJ equivalents of methods which,
### when we're using CELEX rather than SHJ1961, are found in IPA.pm

sub to_features {
    my $symbol = shift;
    my $features = $SHJ1961->{$symbol};
    if ($features) {
	@{$features};
    } else {
	carp "Unknown symbol [$symbol]: $!";
	confess;
	die;
    }
}

sub to_feature_tree {
    my $symbol = shift;
    my $tree;
    my @feature_rep = to_features ($symbol);

    while (@feature_rep) {
	my $feature = shift @feature_rep ;
	my $value = shift @feature_rep;

	if (defined $tree) {
	    $tree->put ($feature->plus) if ($value eq '+');
	    $tree->put ($feature->minus) if ($value eq '-');
	} else {
	    $tree = $feature->plus if ($value eq '+');
	    $tree = $feature->minus if ($value eq '-');
	}
    }
    $tree;
}

# Convert a representation back to binary
sub to_IPA {
    my $rep = shift;
    unless ($rep->get_genus_name eq 'Prosodic Word') {
        carp ("Missing PrWd level in representation\n" . $rep->to_text);
    }
    my @features = ($rep->get_all_constituents)[0]->get_all_constituents;
    my @feature_codes;
    @feature_codes = map {
	$SHJ1961::text->{$_->to_text}
    }
    join (", ", @feature_codes);
}

sub all_features {
    &Representations::get_all_features
}

sub all_symbols {
    keys %{$SHJ1961};
}

sub is_symbol {
    my $symbol = shift;
    grep {$_ eq $symbol} (all_symbols ());
}

1;
