package Stimulus;
use PrCat;
use warnings;
use strict; 
use Carp;
########################################################################
#
# Stimulus.pm
#
# Object implementing a stimulus.
# 
# A Stimulus consists of a representation, a teaching output, and an 
# optional weight (affecting the presentation probability of the 
# Stimulus on a given trial).
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2010 March 21 ($)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


########################################################################
#
# Constructors
#
########################################################################

sub new {
    my ($class,
        $rep,
        $num,
	$name,
        $teach,
        $weight
	) = @_;

    my $base = &Globals::get_representations->get_base;
    if ($base->isa ('MCatGenus')) {
	unless (ref ($rep) and $rep->isa ('MCat')) {
	    carp "Stimulus item must be a MCat, not " . 
		ref (\$rep) . ": $!";
	    confess;
	}
    }
    if ($base->isa ('PrCatGenus')) {
	unless (ref ($rep) and $rep->isa ('PrCat')) {
	    carp "Stimulus item must be a PrCat, not " . 
		ref (\$rep) . ": $!";
	    confess;
	}
    }
    defined ($teach) or ($teach = 0);
    defined ($weight) or ($weight = 1);
    
    my $this = {};

    $this->{rep} = $rep;
    $this->{num} = $num;
    $this->{name} = $name;
    $this->{teach} = $teach;
    $this->{weight} = $weight;

    bless ($this, $class);
    $this;
}

########################################################################
#
# Access methods for intrinsic properties
#
########################################################################

sub get_rep {
    my $this = shift;
    $this->{rep};
}

sub get_teach {
    my $this = shift;
    $this->{teach};
}

sub get_weight {
    my $this = shift;
    $this->{weight};
}

sub set_num {
    my $this = shift;
    my $new_num = shift;
    $this->{num} = $new_num;
}

sub get_num {
    my $this = shift;
    $this->{num};
}

sub set_name {
    my $this = shift;
    my $new_name = shift;
    $this->{name} = $new_name;
}

sub get_name {
    my $this = shift;
    $this->{name};
}

########################################################################
#
# Access methods for extrinsic constraint properties (the ones used by
# Learner.pm).
#
########################################################################

sub set_outcome {
    my (
	$this,
	$con,            # Constraint
	$outcome
	) = @_;

    unless (ref ($con) and $con->isa ('Constraint')) {
	carp "$con is not a Constraint: $!";
	confess;
	die;
    }

    $this->delete_outcome ($con) if (exists $this->{outcome}->{$con});
    $this->{outcome}->{$con} = $outcome;
}

sub get_outcome {
    my (
	$this,
	$con
	) = @_;

    $this->{outcome}->{$con};
}

sub delete_outcome {
    my (
	$this,
	$con
	) = @_;

    unless (exists ($this->{outcome}->{$con})) {
	carp "Unknown constraint $con: $!";
	confess;
	die;
    }

    delete ($this->{outcome}->{$con});

}

# For detecting stim pairs that are treated alike by the constraints.
sub get_signature {
    my $this = shift;

    my $signature;
    my @constraints = (keys (%{$this->{outcome}}));
    for (@constraints) {
	my $outcome = $this->get_outcome ($_);
	$signature .= "." . $outcome;
    }
    $signature;
}

########################################################################
#
# I/O methods
#
########################################################################

sub to_flat_string {
    my $this = shift;
    "[Stimulus " . $this->get_num . " " . $this->get_name . " " . 
	$this->get_teach . " " . $this->get_weight .
	" " . $this->get_rep->to_flat_string;
}

sub to_text {
    my (
	$this,
	$depth    # of nesting
	) = (@_, 0);

    my @text;
    push @text, (
	($Globals::indent x $depth) .
	"Stimulus " . 
	$this->get_num . " " .
	$this->get_name . " " .
	$this->get_teach . " " .
	$this->get_weight
    );
    push @text, $this->get_rep->to_text ($depth+1);

    @text;
}

sub from_text {
    # Pass text in as an array ref, or as an array
    my $text = (ref $_[0] eq 'ARRAY') ? $_[0] : \@_;

    unless (
	$text->[0] =~ /^((?:$Globals::indent_re)*)
                    Stimulus \s+
                    (\d+) \s+
                    (\S+) \s+
                    ([+\-]?\d+) \s+
                    (\d*) \s*$/x) {
	return;
    }

    my ($indent_string,
	$stim_num,
	$stim_name,
	$teach_string,
	$weight_string) = ($1, $2, $3, $4, $5);

#    print "DEBUG:  $indent_string|$stim_num|$stim_name|$teach_string|$weight_string\n";
#    print "DEBUG:  indent_string|stim_num|stim_name|teach_string|weight_string\n";


    my $depth  = length ($indent_string)/length ($Globals::indent);
    my $teach  = 0+$teach_string;
    my $weight = 0+$weight_string;

    shift @{$text};

    my $next_line = $text->[0];
    $next_line =~ /^((?:$Globals::indent_re)*)/;
    my $next_depth = length ($1)/length ($Globals::indent);
    if ($next_depth <= $depth) {
        carp "Expected representation, found $next_line: $!";
        confess;
        die;
    }

    my $base = &Globals::get_representations->get_base;
    my $rep;
    if ($base->isa ('MCatGenus')) {
	($rep) = &MCat::from_text ($text);
    }
    if ($base->isa ('PrCatGenus')) {
	($rep) = &PrCat::from_text ($text);
    }
    unless ($rep) {
        carp "Couldn't parse $next_line: $!";
        confess;
        die;
    }

    my $new;
    unless ($new = new Stimulus ($rep, $stim_num, $stim_name, $teach, $weight)) {
        carp "Error in $next_line: $!";
        carp "Couldn't add Stimulus.";
        confess;
        die;
    }

    $new;
}

sub print {
    my $this = shift;
    print join ("\n", $this->to_text), "\n";
}


1;
