package StimulusPair;
use PrCat;
use warnings;
use strict; 
use Carp;
########################################################################
#
# StimulusPair.pm
#
# Object implementing a stimulus.
#
# A StimulusPair consists of a positive and a negative stimulus, with
# facilities for keeping track of which and how many constraints get
# it right or wrong.
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2010 February 28 ($)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


########################################################################
#
# Constructors
#
########################################################################

sub new {
    my ($class,
	$pos,
	$neg
	) = @_;

    unless (ref ($pos) and $pos->isa ('PrCat')) {
	carp "Positive item must be a PrCat, not " . 
	    ref (\$pos) . ": $!";
	confess;
	die;
    }
    unless (ref ($neg) and $neg->isa ('PrCat')) {
	carp "Positive item must be a PrCat, not " . 
	    ref (\$neg) . ": $!";
	confess;
	die;
    }

    my $this = {};

    # Positive and negative instance in the pair
    $this->{pos} = $pos;
    $this->{neg} = $neg;

    # Count constraints that get it right/wrong/don't care.
    $this->{right_total}    = 0;
    $this->{neither_total}  = 0;
    $this->{wrong_total}    = 0;

    bless ($this, $class);
    $this;
}

########################################################################
#
# Access methods for intrinsic properties
#
########################################################################

sub get_pos {
    my $this = shift;
    $this->{pos};
}

sub get_neg {
    my $this = shift;
    $this->{neg};
}

########################################################################
#
# Access methods for extrinsic constraint properties (the ones used by
# Learner.pm).
#
########################################################################

sub set_outcome {
    my (
	$this,
	$con,            # Constraint
	$outcome
	) = @_;

    unless (ref ($con) and $con->isa ('Constraint')) {
	carp "$con is not a Constraint: $!";
	confess;
	die;
    }

    $this->delete_outcome ($con) if (exists $this->{outcome}->{$con});

    $this->{outcome}->{$con} = $outcome;

    $this->{right_total}++   if ($outcome >  0);
    $this->{neither_total}++ if ($outcome == 0);
    $this->{wrong_total}++   if ($outcome <  0);

    $this->get_grand_total;
}

sub get_outcome {
    my (
	$this,
	$con
	) = @_;

    $this->{outcome}->{$con};
}

sub delete_outcome {
    my (
	$this,
	$con
	) = @_;

    unless (exists ($this->{outcome}->{$con})) {
	carp "Unknown constraint $con: $!";
	confess;
	die;
    }

    my $outcome = $this->get_outcome ($con);
    $this->{right_total}--   if ($outcome > 0);
    $this->{neither_total}-- if ($outcome == 0);
    $this->{wrong_total}--   if ($outcome <  0);

    delete ($this->{outcome}->{$con});

    $this->get_grand_total;

}

sub get_right_total {
    my $this = shift;
    $this->{right_total};
}
sub get_neither_total {
    my $this = shift;
    $this->{neither_total};
}
sub get_wrong_total {
    my $this = shift;
    $this->{wrong_total};
}
sub get_grand_total {
    my $this = shift;
    $this->{right_total} + $this->{neither_total} + $this->{wrong_total};
}

# Simulates human using Luce Choice Rule
sub get_prob_correct {
    my $this = shift;
    ($this->{right_total}-$this->{wrong_total} + $this->get_grand_total/2)
	/$this->get_grand_total;
}

# For detecting stim pairs that are treated alike by the constraints.
sub get_signature {
    my $this = shift;

    my ($right, $neither, $wrong);
    my @constraints = (keys (%{$this->{outcome}}));
    for (@constraints) {
	my $outcome = $this->get_outcome ($_);
	$right   .= $_ if ($outcome >  0);
	$neither .= $_ if ($outcome == 0);
	$wrong   .= $_ if ($outcome <  0);
    }

    "$right:$neither:$wrong";
}

########################################################################
#
# I/O methods
#
########################################################################


1;
