package StimulusSet;
use Globals;
use Stimulus;
use warnings;
use strict;
use Carp;

########################################################################
#
# StimulusSet.pm
#
# A set of Stimuli, with methods for picking one randomly.
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2010 March 21 ($)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


########################################################################
#
# Constructors
#
########################################################################

sub new {
    my $class = shift;

    my $this = {};
    bless $this, $class;
    my @contents = ();
    $this->{contents} = \@contents;
    my @weights = ();
    $this->{weights} =  \@weights;   # For weighted random selection
    $this->{weight_total} = 0;

    $this->put (@_);

    $this;
}

########################################################################
#
# Access methods
#
########################################################################

sub put {
    my $this = shift;
    grep {
	unless (ref ($_) and $_->isa ('Stimulus')) {
	    carp "Can't add $_ to StimulusSet: $!";
	    confess;
	    die;
	}
	push @{$this->{contents}}, $_;
        push @{$this->{weights}}, $_->get_weight;
	$this->{weight_total} += $_->get_weight;
    } @_;
}

sub get_all {
    my $this = shift;
    (@{$this->{contents}});
}

sub get_all_ref {
    my $this = shift;
    $this->{contents};
}

sub get_one {
    my $this = shift;
    my $n = rand ($this->{weight_total});
    my $cumulative_weight = 0;
    for (@{$this->{contents}}) {
	$cumulative_weight += $_->get_weight;
	return $_ if ($cumulative_weight > $n);
    }
    carp "Failure choosing random Stimulus: $!";
    confess;
    die;
}

sub get_stimulus_named {
    my ($this, $name) = @_;

    grep {
	$_->get_name eq $name
    } ($this->get_all);
}

sub how_many {
    my $this = shift;
    exists ($this->{contents}) ? 
        0+@{$this->{contents}} : 0;
}

sub weighted_how_many {
    my $this = shift;
    $this->{weight_total};
}

########################################################################
#
# I/O methods
#
########################################################################

sub to_text {
    my (
        $this,
        $depth      # Depth of nesting
        ) = (@_, 0);

    my @text;
    push @text, (
	($Globals::indent x $depth) .
	"StimulusSet"
    );

    for ($this->get_all) {
	push @text, $_->to_text ($depth + 1);
    }

    @text;
}

sub from_text {
    # Pass text in as ref to array, or array itself
    my $text = (ref $_[0] eq 'ARRAY') ? $_[0] : \@_;

    # Return empty-handed unless the text begins by describing a StimSet
    unless (
        $text->[0] =~ /^((?:$Globals::indent_re)*)
                    StimulusSet\s*$/x) {
	return;
    }
    my $depth = length ($1)/length ($Globals::indent);

    my $new = new StimulusSet;
    shift @{$text};

    while (@{$text}) {
        # Have we run out of dependents?
        my $next_line = $text->[0];
        $next_line =~ /^((?:$Globals::indent_re)*)/;
        my $next_depth = length ($1)/length ($Globals::indent);
        last if ($next_depth <= $depth);

        my ($child, $child_depth, $child_is_head) =
            &Stimulus::from_text ($text);

	unless ($child) {
            carp "Couldn't parse $next_line: $!";
            confess;
            die;
        }
	
        unless ($new->put ($child)) {
            carp "Error in $next_line: $!";
            carp "Can't add $child of genus " . $child->get_genus_name;
            confess;
            die;
        }
    }
	
    ($new, $depth);
}

sub print {
    my $this = shift;
    print (join "\n", $this->to_text);
    print "\n";
}

1;
    
	   


