package UnaryFeature;
use Globals;
use Environment;
use UnaryFeatureGenus;
use warnings;
use strict;
use Carp;

########################################################################
#
# UnaryFeature.pm
#
# Object implementing a unary feature node in a constraint or
# representation.
#
# UnaryFeatures can have named dependents, either UnaryFeatures or 
# BinaryFeatures, of the genera allowed by their UnaryFeatureGenus.
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2009 March 8 ($)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


########################################################################
#
# Constructors
#
########################################################################

# plus -- Create new [+ Foo] node.

sub plus {
    my $class = shift;
    my $UFT = shift;
    unless (ref ($UFT) and $UFT->isa ('UnaryFeatureGenus')) {
	carp "Found ", ref (\$UFT), " while looking for a UnaryFeatureGenus: $!";
	confess;
	die;
    }

    my $this = {};
    $this->{genus} = $UFT;
    $this->{genus_name} = $UFT->get_name;
    $this->{subfeatures} = ();
    bless ($this, $class);
    for (@_) {
	unless ($this->put ($_)) {
	    carp "Can't put $this ";
	    if (ref ($_) and $_->can ('get_genus_name')) {
		carp " of genus ", $_->get_genus_name; 
	    }
	    carp " under $this, of genus ", $this->get_genus_name, ":$!";
	    confess;
	    die;
	}
    }
    $this;
}

sub clone {
    my $this = shift;
    my $copy = plus UnaryFeature (
	$this->get_genus
	);
    my @subfeatures = $this->get_all_immediate_subfeatures;
    for (@subfeatures) {
	$copy->put ($_->clone);
    }
    $copy;
}

sub pseudo_clone {
    # Make a ``shallow copy'' of the Unary Feature $this.
    # Differs from clone in that the copy has *the same* dependents, 
    # rather than clones of those dependents.  Faster than cloning, and
    # safe as long as neither the original nor the copy is modified.

    my $this = shift;
    my $copy = plus UnaryFeature (
	$this->get_genus
	);

#    my @subfeatures = $this->get_all_immediate_subfeatures;
#    for (@subfeatures) {
#	$copy->put ($_);
#    }
#    Maybe this is faster:
#    $copy->{subfeatures} = {};
#    for (keys (%{$this->{subfeatures}})) {
#	$copy->{subfeatures}->{$_} = $this->{subfeatures}->{$_};
#    }
    %{$copy->{subfeatures}} = %{$this->{subfeatures}};

    $copy;
}

sub mutant {
    my $this = shift;
    my $mutagen = (shift or '(no argument)');

    unless (ref ($mutagen) and $mutagen->isa ('Mutagen')) {
        carp "Argument to mutant must be of type Mutagen, not $mutagen:$!";
        confess;
	die;
    }

    return $this->clone unless ($mutagen->mutate);

    # Replace this UnaryFeature with a UnaryFeatureVariable?
    unless ($mutagen->feature_parents_are_protected and 
	    $this->get_all_immediate_subfeatures) {
	if ($mutagen->unary_feature_to_variable) {
	    return plus UnaryFeatureVariable ($this->get_genus, 'alpha');
	}
    }

    # No -- mutate the feature and its dependents.
    my $copy = plus UnaryFeature ($this->get_genus);
    for (@{$this->get_genus->get_dependent_features_ref}) {

	# Consider each *potential* subfeature of $this.  Does $this 
	# actually have such a subfeature?
	my $subfeature = $this->get_immediate_subfeature ($_->get_name);
	my $subfeature_ref = ref ($subfeature);

	if ($subfeature) {
	    # Yes.  Copy it (mutated) unless we fail saving throw.

	    if ($subfeature_ref eq 'BinaryFeature' or 
		$subfeature_ref eq 'BinaryFeatureVariable') {
		unless ($mutagen->lose_binary) {
		    $copy->put ($subfeature->mutant ($mutagen));
		}
		next;
	    }

	    if ($subfeature_ref eq 'UnaryFeature' or
		$subfeature_ref eq 'UnaryFeatureVariable') {
		if ($subfeature_ref eq 'UnaryFeature' and
		    $mutagen->feature_parents_are_protected and 
		    $subfeature->get_all_immediate_subfeatures) {
		    $copy->put ($subfeature->mutant ($mutagen));
		} else {
		    unless ($mutagen->lose_unary) {
			$copy->put ($subfeature->mutant ($mutagen))
		    }
		} 
		next;
	    }
	} else {
	    # No.  Add a new one only if we fail saving throw.
	    my $potential_dependent_ref = ref ($_);

	    if ($potential_dependent_ref eq 'BinaryFeatureGenus') {
		if ($mutagen->gain_binary) {
		    if (rand () < 0.5) {
			$copy->put ($_->plus->mutant ($mutagen));
		    } else {
			$copy->put ($_->minus->mutant ($mutagen));
		    }
		}
		next;
	    }
	    if ($potential_dependent_ref eq 'UnaryFeatureGenus') {
		if ($mutagen->gain_unary) {
		    $copy->put ($_->plus->mutant ($mutagen));
		}
		next;
	    }
	    carp "$_ should be Unary or BinaryFeatureGenus: $!";
	}
    }
    $copy;
}

sub breed {
    my (
	$this, $that,        # Parents
	$recombinator
	) = @_;

    if ((ref $that) and ($that->isa ("Recombinator"))) {
        $recombinator = $that; $that = undef
    }
    unless (defined $recombinator) {
        confess "No Recombinator in call to breed (): $!";
    }


    my $kid;

  SWITCH: {
    if (!defined ($that)) {
	$kid = (!$recombinator->null_wins) ? $this->clone : undef;
	last SWITCH;
    }

    unless ($this->get_genus_name eq $that->get_genus_name) {
	carp "Genera don't match: " . $this->get_genus_name . " != " .
	    $that->get_genus_name . ":$!";
	confess;
	die;
    }
    
    if ($that->isa ("UnaryFeature")) {
	$kid = ($recombinator->this_wins) ? $this->get_genus->plus :
	    $that->get_genus->plus;
	
	# this and that are of same genus, so (named) dependents are
	# drawn from same set
	my @potential_dependents = $this->get_genus->get_dependent_features;
	for (@potential_dependents) {
	    my $name = $_->get_name;
	    my $this_subfeature = $this->get_immediate_subfeature ($name);
	    my $that_subfeature = $that->get_immediate_subfeature ($name);
	    
	    # If both this and that have the subfeature, breed them.
	    
	    if ($this_subfeature and $that_subfeature) {
		$kid->put (
		    $this_subfeature->breed ($that_subfeature, $recombinator));
		next;
	    }
	    
	    # If only one has the subfeature, kid may or may not get it.
	    
	    if ($this_subfeature) {
		if (!$recombinator->null_wins) {
		    $kid->put ($this_subfeature->clone);
		}
		next;
	    }
	    
	    if ($that_subfeature) {
		if (!$recombinator->null_wins) {
		    $kid->put ($that_subfeature->clone);
		}
		next;
	    }
	    
	    # The subfeature is found in neither this nor that, so
	    # the kid doesn't get it.
	}  
	last SWITCH;
    }
    
    if ($that->isa ("UnaryFeatureVariable")) {
	$kid = ($recombinator->this_wins) ? $this->clone : $that->clone;
	last SWITCH;
    }
    
    carp "Can't breed a UnaryFeature with a ". 
	ref ($that) . ": $!";
    confess;
    die;
    }
    
    return $kid;
}



########################################################################
#
# Access methods for permanent attributes
#
########################################################################

sub get_genus {
    my $this = shift;
    $this->{genus};
}

sub get_genus_name {
    my $this = shift;
#    $this->get_genus->get_name;
    $this->{genus_name};
}

sub get_genus_abbrev {
    my $this = shift;
    $this->get_genus->get_abbrev;
}

########################################################################
#
# Access methods for dependents
#
########################################################################

# put -- Tries to add a dependent feature under $this.  First it tries
#     adding it as an immediate dependent of $this; failing that, it
#     tries adding it lower down in the feature tree hanging off of $this.

sub put {
    my $this = shift;
    my $new = shift;

    # Try to put the given Unary/Binary Feature/FeatureVariable under
    # $this.  Returns undef on failure.

    unless (ref $new) {
	carp "Can't put $new under $this: $!";
	confess;
	die;
    }
	
    my $new_ref = ref ($new);

#    if ($new->isa ('UnaryFeature') or 
#	$new->isa ('UnaryFeatureVariable') or
#	$new->isa ('BinaryFeature') or 
#	$new->isa ('BinaryFeatureVariable')) {
	
    if ($new_ref eq 'UnaryFeature' or 
	$new_ref eq 'UnaryFeatureVariable' or
	$new_ref eq 'BinaryFeature' or
	$new_ref eq 'BinaryFeatureVariable') {
	
	if ($new->get_genus->is_dependent_of ($this->get_genus)){
	    # New feature is immediate dependent of $this; set at once:
	    $this->{subfeatures}->{$new->get_genus_name} = $new;
	    return $new;
	} else {
	    # Try setting new feature in $this's dependents:
	    for ($this->get_all_immediate_subfeatures) {
		if ($_->can ('put')) {
		    if ($_->put ($new)) {
			return $new;
		    }
		}
	    }
	}
    }
}

# get_all_immediate_subfeatures -- Get the actual feature objects which
#     are dependents of this one.
sub get_all_immediate_subfeatures {
    my $this = shift;
    values (%{$this->{subfeatures}});
}

# get_immediate_subfeature -- Given the name of a subfeature, return the
#     object.
sub get_immediate_subfeature {
    my $this = shift;
    my $genus_name = shift;
    $this->{subfeatures}->{$genus_name};
}

# get_subfeature -- Search the tree rooted at $this for the feature 
#     named $genus_name.
sub get_subfeature {
    my $this = shift;
    my $genus_name = shift;

    # If $genus names one of $this's immediate subfeatures, we're in luck:
    for ($this->get_immediate_subfeature ($genus_name)) {
	return $this if $_;
    }
    
    # Else, search under $this's subfeatures:
    for ($this->get_all_immediate_subfeatures) {
	return $this if $_->get_subfeature ($genus_name)
    }
}

sub delete_immediate_subfeature {
    my $this = shift;
    my $genus_name = shift;
    delete ($this->{subfeatures}->{$genus_name});
}

sub get_sign {
    my $this = shift;
    '+';
}


sub has_dependents {
    my $this = shift;
    if ($this->get_all_immediate_subfeatures) {
	1;
    }
}

sub get_depth {
    # How deep is the subtree rooted at this node?
    my $this = shift;
    my $max_subfeature_depth = 0;
    for ($this->get_all_immediate_subfeatures) {
	my $subfeature_depth = $_->get_depth;
	$max_subfeature_depth = $subfeature_depth 
	    if $max_subfeature_depth < $subfeature_depth;
    }
    $max_subfeature_depth + 1;
}

sub get_subtree_size {
    # How many nodes are in the subtree rooted at this node?
    my $this = shift;
    my $node_count = 1;
    for ($this->get_all_immediate_subfeatures) {
        $node_count += $_->get_subtree_size;
    }
    $node_count;
}


########################################################################
#
# Matching methods
#
########################################################################

# A UnaryFeature $this matches in a UnaryFeature $that iff $this and 
# $that are of the same type (genus, e.g., both [Place]), and every
# immediate dependent of $this matches in its namesake dependent of
# $that.

sub matches_in {
    my (
	$this, 
	$that,
	$env
	) = @_;
    defined ($env) or ($env = new Environment);

    unless (ref ($that) and $that->can ('matches_in')) {
	carp "Match target must be a matchable object, not $that: $!";
        confess;
        die;
    }
    unless (ref ($env) and $env->isa ('Environment')) {
        carp "Second argument to matches_in must be Environment, not $that: $!";
        confess;
        die;
    }

    if ($that->isa ('UnaryFeature')) {
	return unless ($this->get_genus_name eq $that->get_genus_name);
	
	# Recursively check whether the subfeatures match.

	# Break $this into any subfeature vs. the rump.
	my @this_subfeatures = $this->get_all_immediate_subfeatures;

	# Basis case: No subfeatures; match succeeds.
	return (my @foo = ($env)) unless @this_subfeatures;

	# Quick check:  Are there enough subfeatures to match?
	return unless (@this_subfeatures <= $that->get_all_immediate_subfeatures);

	# Quick check:  Do all the this_subfeature names occur in the that_subfeatures?
	my @that_subfeatures = $that->get_all_immediate_subfeatures;
	my @that_subfeature_names = map {
	    $_->get_genus_name
	} @that_subfeatures;
	my $matched_all = 1;
	for my $this_subfeature (@this_subfeatures) {
	    my $matched_this = 0;
	    my $this_genus_name = $this_subfeature->get_genus_name;
	    for my $that_subfeature (@that_subfeatures) {
		my $that_genus_name = $that_subfeature->get_genus_name;
		if ($this_genus_name eq $that_genus_name) {
		    $matched_this = 1;
		    last;
		}
	    }
	    unless ($matched_this) {
		$matched_all = 0;
		last;
	    }
	}
	return unless $matched_all;

	# Otherwise:  Make a clone, diminished by removal of first subfeature.
	my $this_0 = shift @this_subfeatures;
	my $this_prime = $this->pseudo_clone;
	$this_prime -> delete_immediate_subfeature ($this_0->get_genus_name);
	
	# Do the same to $that, breaking off the same subfeature if psbl:
	my $that_0 = $that->get_immediate_subfeature ($this_0->get_genus_name);
	return unless $that_0;
	my $that_prime = $that->pseudo_clone;
	$that_prime->delete_immediate_subfeature ($that_0->get_genus_name);

	# Get refinements of $env that allow $this_0 and $that_0 to match:
	my @initial_successes = $this_0->matches_in ($that_0, $env);
	return unless @initial_successes;
	
	# Now use those environments to match the rump $this_prime and
	# $that_prime.

	my @successes = ();
	for (@initial_successes) {
	    unless (ref ($_) and ($_->isa ('Environment'))) {
		carp "Expected environment, found $_ of type ",
		ref ($_), " instead: $!";
		confess;
		die;
	    }
	    push @successes, $this_prime->matches_in ($that_prime, 
						      $_->clone)
	}

	return @successes;
    } else {
	return ();
    # A UnaryFeature matches in nothing but another UnaryFeature, not 
    # even in a UnaryFeatureVariable.
    }
}

sub equals {
    my ($this, $that, $env) = @_;
    $this->matches_in ($that, $env) and $that->matches_in ($this, $env);
}
	
########################################################################
#
# I/O methods
#
########################################################################

sub to_flat_string {
    my (
	$this,
	$is_head
	) = (@_, 0);

    my $flat_string = $this->get_sign . $this->get_genus_abbrev;
    if ($is_head) {
	$flat_string = "[*$flat_string*";
    } else {
	$flat_string = "[$flat_string";
    }
    for (sort {
	$a->get_genus_name cmp $b->get_genus_name
	 }($this->get_all_immediate_subfeatures)) {
	$flat_string .=  $_->to_flat_string ();
    }
    $flat_string .= ']';
    $flat_string;
}

sub to_text {
    my (
        $this,
        $depth,       # Depth of nesting
        $is_head    # Head of parent?
        ) = (@_, 0, 0);

    my @text;
    if ($is_head) {
        push @text, ($Globals::indent x $depth) .
            '*' . 
            $this-> get_sign . 
            $this-> get_genus_name .
            '*'
    } else {
        push @text, ($Globals::indent x $depth) .
            '[' . 
            $this-> get_sign .
            $this-> get_genus_name .
            ']';
    }

    for (sort {
	$a->get_genus_name cmp $b->get_genus_name
	 }($this->get_all_immediate_subfeatures)) {
	push @text, $_->to_text ($depth + 1);
    }
    @text;
}

sub from_text {
    # Pass text in as ref to array, or array itself
    my $text = (ref $_[0] eq 'ARRAY') ? $_[0] : \@_;

    # Return empty-handed unless the text begins by describing a BinFeat
    unless (
        $text->[0] =~ /^((?:$Globals::indent_re)*)
                    ([\[\*])    \s*
                    ([+\-])     \s*
                    ([\w\s]+\w) \s*
                    [\]\*]      \s*$/x) {
        return;
    }
    my ($indent_string, 
        $head_string,
        $sign,
        $genus_name) = ($1, $2, $3, $4);
    
    my $depth = length ($indent_string)/length ($Globals::indent);
    my $is_head = ($head_string eq '*') ? 1 : 0;
    my $coefficient = ($sign eq '+') ? 1 : -1;
    my $genus = &Globals::get_representations->get ($genus_name);

    unless (ref ($genus) and $genus->isa ('UnaryFeatureGenus')) {
        return;
    }

    my $new = $genus->plus;
    shift @{$text};

    while (@{$text}) {

	# Have we run out of dependents?
	my $next_line = $text->[0];
	$next_line =~ /^((?:$Globals::indent_re)*)/;
	my $next_depth = length ($1)/length ($Globals::indent);
	last if ($next_depth <= $depth);

	my ($child, $child_depth, $child_is_head) =
	    &UnaryFeature::from_text ($text);
	unless ($child) {
	    ($child, $child_depth, $child_is_head) =
		&BinaryFeature::from_text ($text);
	} 
	unless ($child) {
	    ($child, $child_depth, $child_is_head) =
            &UnaryFeatureVariable::from_text ($text);
	}
	unless ($child) {
	    ($child, $child_depth, $child_is_head) =
            &BinaryFeatureVariable::from_text ($text);
	}
	unless ($child) {
	    carp "Couldn't parse $next_line: $!";
	    confess;
	    die;
	}
	unless ($new->put ($child)) {
	    carp "Error in $next_line: $!";
	    confess;
	    die;
	}
    }

    ($new, $depth, $is_head);
}


sub print {
    my (
        $this,
        $depth,     # Depth of nesting
	$is_tree
        ) = (@_, 0);
    print (join "\n", $this->to_text ($depth));
    print "\n";
    
}

sub to_tree {
    my $this = shift;
    my $width = (shift or '1in');

    my $indent_level = (shift or 0);
    my $indent_quantum = 2;
    my $indent = ' ' x ($indent_quantum * (1+$indent_level));

    my $leader_level = (shift or 0);
    my $leader_quantum = 1;
    my $leader = '.' x ($leader_quantum * $leader_level);

    my $style = (shift or '\ttfamily');

    my $is_head = shift;


    my $string = "{$style" . "{$leader}" . '{[}' . $this->get_genus_abbrev . 
	'{]}}\\\\';;
    $string = $indent . $string . "\n";

    for ($this->get_genus->get_dependent_feature_names) {
	my $subfeature = $this->get_immediate_subfeature ($_);
	next unless $subfeature;
	$string .= $subfeature->to_tree ($width, $indent_level + 1, 
	    $leader_level + 1, $style);
    }
    $string;
}


1;
