package Variable;
use Environment;
use warnings;
use strict;
use Carp;

########################################################################
#
# Variable.pm
#
# A node employing a Greek-letter variable.  The Variable nodes appear
# in a subtree as [+alpha Foo], [-alpha Foo], etc., where Foo is the genus
# ("high", "Place", "PrWd", etc.) and alpha (beta, etc.)  is the species 
# (the name used to distinguish it from other variables of the same 
# genus).  
#
# In a given constraint, all [+/- alpha Foo] nodes refer to the same alpha.  
# For example, if the constraint contains the Variables [+alpha Place] and
# [-alpha Place], a successful match in a representational tree pairs 
# the first one with the subtree rooted at some Place node, and the second
# with a *non-identical* subtree rooted at some other Place node.
#
# During a match attempt, memory for what has thus far been matched to 
# [+alpha Foo] and what to [-alpha Foo] is maintained by the Environment
# in an EnvironmentVariable devoted to [alpha Foo].  See those objects
# for details.
#
# E. Moreton  *  UNC-Chapel Hill Linguistics  *  2009 March 8 ($)
#
########################################################################
#
#   Copyright 2019 by Elliott Moreton.
#
#   This file is part of the Winnow-MaxEnt-Subtree Learner, Version of
#   2019 December 17.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   free software: you can redistribute it and/or modify it under the
#   terms of the GNU General Public License as published by the Free
#   Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   The Winnow-MaxEnt-Subtree Learner, Version of 2019 December 17, is
#   distributed in the hope that it will be useful, but WITHOUT ANY
#   WARRANTY; without even the implied warranty of MERCHANTABILITY or
#   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
#   License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with the Winnow-MaxEnt-Subtree Learner, Version of 2019
#   December 17.  If not, see <https://www.gnu.org/licenses/>.
#
########################################################################


########################################################################
#
# Constructors
#
########################################################################

# plus, minus -- returns [+alpha Foo] or [-alpha Foo], given the 
#     PrCatGenus, UnaryFeatureGenus, or BinaryFeatureGenus named Foo.

sub plus {
    my ($class, $genus, $species) = @_;
    new Variable ($genus, $species);
}

sub minus {
    my ($class, $genus, $species) = @_;
    new Variable ($genus, $species, -1);
}

# new -- For internal use only.
sub new {
    my (
	$class,
	$genus,
	$species,
	$negated
	) = @_;

    unless (ref ($genus) and (
		$genus->isa ('UnaryFeatureGenus') or
		$genus->isa ('BinaryFeatureGenus') or
		$genus->isa ('PrCatGenus')
	    )) {
	carp "Found ", ref (\$genus), " while looking for a genus: $!";
	confess;
	die;
    }

    my $this = {};

    $this->{genus} = $genus;
    $this->{species} = $species;
    $this->{negated} = ($negated) ? -1 : 0;

    bless $this, $class;
    return $this;
}

sub clone {
    my $this = shift;
    new Variable (
	$this->get_genus,
	$this->get_species
	);
}

sub pseudo_clone {
    my $this = shift;
    $this;
}

########################################################################
# 
# Access methods
#
########################################################################

sub get_name {
    my $this = shift;
    $this->get_species . "::" . 
	$this->get_genus_name;
}

sub is_negated {
    my $this = shift;
    $this->{negated};
}

sub get_sign {
    my $this = shift;
    ($this->is_negated) ? "-" : "+";
}

sub get_coefficient {
    my $this = shift;
    ($this->is_negated) ? -1 : +1;
}

sub get_genus {
    my $this = shift;
    $this->{genus};
}

sub get_genus_name {
    my $this = shift;
    $this->get_genus->get_name;
}

sub get_genus_abbrev {
    my $this = shift;
    $this->get_genus->get_abbrev;
}

sub get_species {
    my $this = shift;
    $this->{species};
}

# Variables never have dependents:

sub has_dependents {
    return;
}

sub get_depth {
    1;
}

sub get_subtree_size {
    1;
}

########################################################################
# 
# Match methods
#
########################################################################

# A Variable $this matches successfully in a node $that given an
# environment $env if $this and $that are of the same genus (type), 
# say [+/-alpha Foo], and one of the following is true
#
# (1)  $env has never seen [+/-alpha Foo].  The match succeeds, and $that
#      is stored in $env as a case of [+alpha Foo] or [-alpha Foo] 
#      depending on whether $this is + or -.
#
# (2)  $env has seen [+/-alpha Foo], and $this is [+alpha Foo], and 
#      $that doesn't differ from any known case of [+alpha Foo] and is
#      not identical to any known case of [-alpha Foo].  $that is added 
#      to $env as a case of [+alpha Foo].
#
# (3)  $env has seen [+/-alpha Foo], and $this is [-alpha Foo], and
#      $that is not identical to any known case of [+alpha Foo].  $that
#      is added to $env as a case of [-alpha Foo].
#
# For each successful match, matches_in returns an Environment with the
# variable assignments required to make the match succeed (just like all
# other matches_in methods).

sub matches_in {
    my (
	$this,     # Thing to match
	$that,     # Thing to match it in
	$env       # Environment
	) = @_;

    # If the match attempt begins here, create new Environment, else
    # use current one.
    defined ($env) or ($env = new Environment);

    unless (ref ($that) and $that->can ('matches_in')) {
        carp "Match target must be a matchable object, not $that: $!";
        confess;
        die;
    }
    unless (ref ($env) and $env->isa ('Environment')) {
        carp "Second argument to matches_in must be Environment, not $env: $!";
        confess;
        die;
    }

    # Match fails if $this and $that aren't of the same genus:
    unless ($this->can ("get_genus_name")) {
	carp "Tried to match $this, of no genus: $!";
	carp $this->to_text;
	confess;
	die;
    }
    unless ($that->can ("get_genus_name")) {
	carp "Tried to match in $that, of no genus: $!";
	carp $that->to_text;
	confess;
	die;
    }
    return unless ($this->get_genus_name eq $that->get_genus_name);

    my $success = 1;

    # A negative Variable ("-alpha Place", etc.) matches in $that iff $that
    # is different from all of the known positive cases.  If the match is
    # successful, $that is added to the list of known negative cases, and,
    # if $that is a binary feature, its inverse is added to the positive
    # cases.

    my $new_env = $env->clone;
    if ($this->is_negated) {
	for ($env->get_all_positive_cases_of ($this)) {
	    $success = ($success and !$that->equals ($_, $env))
	}
	if ($success) {
	    $new_env->add_negative_case_of ($this, $that);
	    if ($that->can ("inverted_clone")) {
		$new_env->add_positive_case_of ($this, $that->inverted_clone);
	    }
	}
    } else {

    # A positive Variable ("alpha Place", "alpha voice", "alpha PrWd", etc.)
    # matches in $that iff $that is different from all of the negative
    # cases of $this, and is identical to all of the positive cases.
    # A successful match adds $that to the list of known positive cases.
    # (It is actually redundant to accumulate more than one positive case,
    # since they're all alike.  Should fix that some time.)
    
	for ($env->get_all_negative_cases_of ($this)) {
	    $success = ($success and !$that->equals ($_, $env));
	}
	for ($env->get_all_positive_cases_of ($this)) {
	    $success = ($success and $that->equals ($_, $env));
	}
	if ($success) {
	    $new_env->add_positive_case_of ($this, $that);
	}
    }

    if ($success) {
	return ($new_env);
    } else {
	return;
    }
}

sub equals {
    my ($this, $that, $env) = @_;
    if ($this->isa ('Variable') and $that->isa ('Variable')) {
	# Two Variables are equal if and only if they have the same 
	# genus, species, and sign
	($this->get_genus_name eq $that->get_genus_name) and
	    ($this->is_negated eq $that->is_negated) and 
	    ($this->get_species eq $that->get_species)
    } else {
	$this->matches_in ($that, $env) and $that->matches_in ($this, $env);
    }
}


########################################################################
#
# I/O methods
#
########################################################################

sub to_flat_text {
    my $this = shift;
    sprintf $Globals::variable_format, $this->get_sign, $this->get_species, $this->get_genus_name;
}

sub to_text {
    my (
        $this,
        $depth     # Depth of nesting                                           
        ) = @_;
    $depth = 0 unless ($depth);
    if ($depth < 0) {
        carp "Negative nesting depth $depth: $!";
        die;
    }

    ($Globals::indent x $depth) .
        (sprintf $Globals::variable_format,
         $this->get_sign,
	 $this->get_species,
         $this->get_genus_name);
}

sub print {
    my (
        $this,
        $depth     # Depth of nesting                                           
        ) = @_;
    print $this->to_text ($depth), "\n";
}

    
1;
